package gov.usgs.earthquake.nshmp;

import static com.google.common.base.Preconditions.checkArgument;

import java.util.AbstractCollection;
import java.util.Arrays;
import java.util.Collection;
import java.util.List;
import java.util.function.Function;
import java.util.stream.Collectors;
import java.util.stream.DoubleStream;

import com.google.common.base.CharMatcher;
import com.google.common.base.Enums;
import com.google.common.base.Joiner;
import com.google.common.base.Splitter;
import com.google.common.base.StandardSystemProperty;
import com.google.common.base.Strings;
import com.google.common.collect.FluentIterable;
import com.google.common.collect.Iterables;
import com.google.common.collect.Lists;
import com.google.common.primitives.Doubles;

/**
 * Miscellaneous {@code String} utilities.
 *
 * @author U.S. Geological Survey
 */
public class Text {

  private Text() {}

  /** System specific newline string. */
  public static final String NEWLINE = StandardSystemProperty.LINE_SEPARATOR.value();

  /** Null string ("null"). */
  public static final String NULL = "null";

  /** The column on which to align values in a log entry. */
  public static final int LOG_VALUE_COLUMN = 32;

  private static final int LOG_INDENT_SIZE = 8;

  /** A newline plus the number of spaces to indent multiline log entries. */
  public static final String LOG_INDENT = NEWLINE + Strings.repeat(" ", LOG_INDENT_SIZE);

  /**
   * Whitespace {@link Splitter splitter} that omits empty strings. Shortcut
   * for: {@code Splitter.on(CharMatcher.whitespace()).omitEmptyStrings();}
   */
  public static final Splitter WHITESPACE_SPLITTER = Delimiter.SPACE.splitter();

  /**
   * Verify that the supplied {@code name} is neither {@code null} nor empty.
   *
   * @param name to verify
   * @param label for message
   * @return the supplied {@code name} for inline use
   * @throws IllegalArgumentException if name or label is {@code null} or empty
   */
  public static String checkName(String name, String label) {
    checkArgument(
        !Strings.nullToEmpty(name).trim().isEmpty(),
        "%s name may not be empty or null",
        label);
    return name;
  }

  /**
   * Returns a string containing the string representation of each of
   * {@code parts} joined with {@code delimiter}.
   *
   * @param parts the objects to join
   * @param delimiter the {@link Delimiter} to join on
   * @see Joiner
   */
  public static String join(Iterable<?> parts, Delimiter delimiter) {
    return delimiter.joiner().join(parts);
  }

  /**
   * Split a {@code sequence} into string components and make them available
   * through a (possibly-lazy) {@code Iterator}.
   *
   * @param sequence the sequence of characters to split
   * @param delimiter the {@link Delimiter} to split on
   * @see Splitter
   */
  public static Iterable<String> split(CharSequence sequence, Delimiter delimiter) {
    return delimiter.splitter().split(sequence);
  }

  /**
   * Split a {@code sequence} into string components and make them available
   * through an immutable {@code List}.
   *
   * @param sequence the sequence of characters to split
   * @param delimiter the {@link Delimiter} to split on
   */
  public static List<String> splitToList(CharSequence sequence, Delimiter delimiter) {
    return delimiter.splitter().splitToList(sequence);
  }

  /**
   * Split {@code sequence} into {@code Double} components and make them
   * available through an immutable {@code List}.
   *
   * @param sequence the sequence of characters to split
   * @param delimiter the {@link Delimiter} to split on
   */
  public static List<Double> splitToDoubleList(CharSequence sequence, Delimiter delimiter) {
    return FluentIterable
        .from(split(sequence, delimiter))
        .transform(Doubles.stringConverter())
        .toList();
  }

  /**
   * Delimiter identifiers, each of which can provide a {@link Joiner} and
   * {@link Splitter}.
   */
  public enum Delimiter {

    /** Colon (':') delimiter. */
    COLON(':'),

    /** Comma (',') delimiter. */
    COMMA(','),

    /** Dash ('-') delimiter. */
    DASH('-'),

    /** Period ('.') delimiter. */
    PERIOD('.'),

    /** Forward-slash ('/') delimiter. */
    SLASH('/'),

    /**
     * Whitespace (' ') delimiter.
     * @see CharMatcher#whitespace()
     */
    SPACE(' ', CharMatcher.whitespace()),

    /** Underscore ('_') delimiter. */
    UNDERSCORE('_');

    private Joiner joiner;
    private Splitter splitter;

    private Delimiter(char separator) {
      joiner = Joiner.on(separator).skipNulls();
      splitter = Splitter.on(separator).omitEmptyStrings().trimResults();
    }

    private Delimiter(char joinSeparator, CharMatcher splitMatcher) {
      joiner = Joiner.on(joinSeparator).skipNulls();
      splitter = Splitter.on(splitMatcher).omitEmptyStrings().trimResults();
    }

    /**
     * Returns a null-skipping {@link Joiner} on this {@code Delimiter}.
     * @see Joiner#skipNulls()
     */
    public Joiner joiner() {
      return joiner;
    }

    /**
     * Returns an empty-string-omitting and result-trimming {@link Splitter} on
     * this {@code Delimiter}.
     *
     * @see Splitter#omitEmptyStrings()
     * @see Splitter#trimResults()
     */
    public Splitter splitter() {
      return splitter;
    }
  }

  /**
   * Return a {@code String} representation of an {@code Iterable<Enum>} where
   * {@code Enum.name()} is used instead of {@code Enum.toString()}.
   *
   * @param iterable to process
   * @param enumClass
   */
  public static <E extends Enum<E>> String enumsToString(Iterable<E> iterable,
      Class<E> enumClass) {
    return addBrackets(FluentIterable
        .from(iterable)
        .transform(Enums.stringConverter(enumClass).reverse())
        .join(Delimiter.COMMA.joiner()));
  }

  /**
   * Convert an {@code Enum.name()} to a space-delimited presentation-friendly
   * string.
   *
   * @param e the {@code Enum} to generate label for
   * @param capitalize true if first letter of each word should be capitalized;
   *        false if letters should all be lowercase
   */
  public static String enumLabelWithSpaces(Enum<? extends Enum<?>> e, boolean capitalize) {
    return join(splitEnum(e, capitalize), Delimiter.SPACE);
  }

  /**
   * Convert an {@code Enum.name()} to a dash-delimited presentation-friendly
   * string.
   *
   * @param e the {@code Enum} to generate label for
   * @param capitalize true if fist letter of each word should be capitalized;
   *        false if letters should all be lowercase
   */
  public static String enumLabelWithDashes(Enum<? extends Enum<?>> e, boolean capitalize) {
    return join(splitEnum(e, capitalize), Delimiter.DASH);
  }

  /* Splits and possibly capitalizes an enum.name() */
  private static List<String> splitEnum(Enum<? extends Enum<?>> e, boolean capitalize) {
    Iterable<String> sources = split(e.name(), Delimiter.UNDERSCORE);
    List<String> results = Lists.newArrayList();
    for (String s : sources) {
      results.add(capitalize ? capitalize(s) : s.toLowerCase());
    }
    return results;
  }

  /**
   * Capitalize supplied string by converting the first {@code char} to
   * uppercase and all subsequent {@code char}s to lowercase.
   *
   * @param s the string to capitalize
   */
  public static String capitalize(String s) {
    return Character.toUpperCase(s.charAt(0)) + s.substring(1).toLowerCase();
  }

  /**
   * Convert a {@code Collection<Double>} to a string of the same format
   * returned by {@link Arrays#toString(double[])} and
   * {@link AbstractCollection#toString()} , but will format the values using
   * the supplied format string. The supplied {@code format} should match that
   * expected by {@code String.format(String, Object...)}
   *
   * @param values the values to convert
   * @param format a format string
   * @param delimiter the delimiter to use
   * @param brackets {@code true} if hard brackets should be added to start and
   *        end, {@code false} otherwise
   */
  public static String toString(
      Collection<Double> values,
      String format,
      String delimiter,
      boolean brackets) {

    String base = values.stream()
        .map(Text.formatDoubleFunction(format))
        .collect(Collectors.joining(delimiter));
    // String base = Joiner.on(delimiter).join(
    // Iterables.transform(
    // values,
    // Text.formatDoubleFunction(format)));
    return brackets ? addBrackets(base) : base;
    // TODO replace with prefix suffix
  }

  /**
   * Convert a {@code DoubleStream} to a string of the same format returned by
   * {@link Arrays#toString(double[])} and {@link AbstractCollection#toString()}
   * , but will format the values using the supplied format string. The supplied
   * {@code format} should match that expected by
   * {@code String.format(String, Object...)}
   *
   * @param stream the stream to convert
   * @param format a format string
   * @param delimiter the delimiter to use
   * @param brackets {@code true} if hard brackets should be added to start and
   *        end, {@code false} otherwise
   */
  public static String toString(
      DoubleStream stream,
      String format,
      String delimiter,
      boolean brackets) {
    List<Double> values = stream.boxed().collect(Collectors.toList());
    return toString(values, format, delimiter, brackets);
  }

  /*
   * Adds brackets to the supplied string.
   */
  static String addBrackets(String s) {
    return '[' + s + ']';
  }

  /**
   * Reformats string representation of point arrays onto single line and
   * appends newline character
   */
  public static String cleanCoordinates(String s) {
    return s.replaceAll("\\[\\s+([-\\d])", "[$1")
        .replaceAll(",\\s+([-\\d])", ", $1")
        .replaceAll("(\\d)\\s+\\]", "$1]")
        .replaceAll("}\\Z", "}" + NEWLINE);
  }

  /**
   * Returns a {@link Function} for converting {@code double}s to formatted
   * strings. If a value to format is 0.0, the format string is ignored in favor
   * of always printing the often more compact string: "0.0".
   *
   * @param format string expected by
   * @see String#format(String, Object...)
   */
  public static Function<Double, String> formatDoubleFunction(String format) {
    return d -> (d == 0.0) ? "0.0" : String.format(format, d);
  }

  /**
   * Read and return the {@code double} at {@code position} in a space-delimited
   * string.
   *
   * @param s the string to read from
   * @param position the position to read
   */
  @Deprecated
  public static double readDouble(String s, int position) {
    return Double.valueOf(Iterables.get(split(s, Delimiter.SPACE), position));
  }

}
