package gov.usgs.earthquake.nshmp.calc;

import static com.google.common.base.Preconditions.checkState;
import static gov.usgs.earthquake.nshmp.data.XySequence.copyOf;

import java.util.EnumMap;
import java.util.Map;

import gov.usgs.earthquake.nshmp.data.XySequence;
import gov.usgs.earthquake.nshmp.gmm.Gmm;
import gov.usgs.earthquake.nshmp.gmm.Imt;

/**
 * Container class for the combined hazard curves derived from the individual
 * {@code Source}s in a {@code ClusterSource}, one for each
 * {@code GroundMotionModel} used. The curves will have been scaled by source
 * and rupture weights, but not by {@code GroundMotionModel} weights.
 *
 * @author U.S. Geological Survey
 */
final class ClusterCurves {

  final ClusterGroundMotions clusterGroundMotions;
  final Map<Imt, Map<Gmm, XySequence>> curveMap;

  private ClusterCurves(
      ClusterGroundMotions clusterGroundMotions,
      Map<Imt, Map<Gmm, XySequence>> curveMap) {

    this.clusterGroundMotions = clusterGroundMotions;
    this.curveMap = curveMap;
  }

  static Builder builder(ClusterGroundMotions clusterGroundMotions) {
    return new Builder(clusterGroundMotions);
  }

  static class Builder {

    private static final String ID = "ClusterCurves.Builder";
    private boolean built = false;

    private final ClusterGroundMotions clusterGroundMotions;
    private final Map<Imt, Map<Gmm, XySequence>> curveMap;

    private Builder(ClusterGroundMotions clusterGroundMotions) {
      this.clusterGroundMotions = clusterGroundMotions;
      // look at first HazardGM to determine curve table dimensions
      GroundMotions model = clusterGroundMotions.get(0);
      curveMap = new EnumMap<>(Imt.class);
      for (Imt imt : model.gmMap.keySet()) {
        Map<Gmm, XySequence> gmmMap = new EnumMap<>(Gmm.class);
        curveMap.put(imt, gmmMap);
      }
    }

    /* Makes an immutable copy of the supplied curve. */
    Builder addCurve(Imt imt, Gmm gmm, XySequence curve) {
      curveMap.get(imt).put(gmm, copyOf(curve));
      return this;
    }

    ClusterCurves build() {
      checkState(!built, "This %s instance has already been used", ID);
      built = true;
      return new ClusterCurves(clusterGroundMotions, curveMap);
    }

  }

}
