package gov.usgs.earthquake.nshmp.calc;

import static com.google.common.base.Preconditions.checkArgument;
import static java.lang.Math.PI;
import static java.lang.Math.cos;
import static java.lang.Math.exp;
import static java.lang.Math.log;
import static java.lang.Math.max;
import static java.lang.Math.min;
import static java.lang.Math.sqrt;
import static java.util.stream.Collectors.toMap;

import java.util.EnumMap;
import java.util.Map;
import java.util.Set;

import gov.usgs.earthquake.nshmp.gmm.Imt;

/**
 * Conditional spectrum implementation.
 *
 * @author U.S. Geological Survey
 */
class ConditionalSpectrum {

  // interface ImtCorrelation extends ToDoubleBiFunction<Imt, Imt> {}

  /*
   * Method 4: "Exact" CS with disaggregation weights.
   *
   * Result: Map<Imt, XySequence> (x=T, y=μ) Map<Imt, XySequence> (x=T, y=σ)
   *
   * Given target IMT T* and conditional Imt Tj
   *
   * For each source/gmm, we need epsilon at the target IMT, Maths.epsilon(...)
   * which we can obtain from the disaggregation
   *
   * Primary loop is target, iterate over all other SAs including Tj = T*
   * (should just yield μCond = μ + ε and σ = 0.0
   */

  // static final Table<Imt, Imt, Double> ρTable = initCorrelations();

  /*
   * Create a table of SA IMT correlations. Table contains duplicate values as
   * all possible ordered pairs (e.g. Imt1:Imt2 and Imt2:Imt1) are included.
   */
  // static Table<Imt, Imt, Double> initCorrelations() {
  //
  // Set<Imt> saImts = EnumSet.copyOf(Imt.mprsImts());
  // saImts.remove(Imt.PGA);
  // Table<Imt, Imt, Double> table = ArrayTable.create(saImts, saImts);
  // for (Imt imt1 : saImts) {
  // saImts.stream().forEach(imt2 -> table.put(
  // imt1,
  // imt2,
  // Maths.round(CorrelationModel.BAKER_JAYARAM_08.value(imt1, imt2), 8)));
  // }
  // return ImmutableTable.copyOf(table);
  // }

  static double totalMean() {

    // Summation over all μ * disaggWt
    // double disaggWt =
    return 0;
  }

  /**
   * @param μt conditional mean ground motion over all ruptures at a target
   *        period
   * @param μc conditional mean ground motion of a rupture at a target period
   * @param cσ conditional sigma of ground motion of a rupture at a target
   *        period
   * @param p
   * @return
   */
  static double totalSigma(double μt, double μc, double σc, double p) {
    double sigma = p * ((σc * σc) + (μc - μt) * (μc - μt));
    // TODO summation of scaled sigmas over all ruptures
    return sqrt(sigma);
  }

  /**
   * Compute the conditional mean ground motion of a source at some target
   * period.
   *
   * @param μ mean ground motion of a rupture at a target period
   * @param σ sigma of ground motion of a rupture at a target period
   * @param ρ correlation coefficient between reference and target period
   *        epsilons
   * @param ε epsilon of a ground motion of a rupture at the reference period
   */
  private static double mean(double μ, double σ, double ρ, double ε) {
    return μ + ρ * ε * σ;
  }

  /**
   * @param σ sigma of ground motion of a rupture at a target period
   * @param ρ correlation coefficient between reference and target period
   *        epsilons
   * @return
   */
  static double sigma(double σ, double ρ) {
    return σ * sqrt(1 - ρ * ρ);
  }

  public static enum CorrelationModel {

    BAKER_JAYARAM_08() {

      @Override
      public double value(Imt imt1, Imt imt2) {
        checkArgument(imt1.isSA());
        checkArgument(imt2.isSA());

        double t1 = imt1.period();
        double t2 = imt2.period();

        double tMin = min(t1, t2);
        double tMax = max(t1, t2);

        double c1 = (1.0 - cos(PI / 2.0 - log(tMax / max(tMin, 0.109)) * 0.366));
        double c2 = (tMax < 0.2)
            ? 1.0 - 0.105 * (1.0 - 1.0 / (1.0 + exp(100.0 * tMax - 5.0))) *
                (tMax - tMin) / (tMax - 0.0099)
            : Double.NaN;
        double c3 = (tMax < 0.109) ? c2 : c1;
        double c4 = c1 + 0.5 * (sqrt(c3) - c3) * (1.0 + cos(PI * tMin / 0.109));

        if (tMax <= 0.109) {
          return c2;
        } else if (tMin > 0.109) {
          return c1;
        } else if (tMax < 0.2) {
          return min(c2, c4);
        } else {
          return c4;
        }
      }
    };

    /**
     * Compute the correlation coefficient between two IMTs. The order of the
     * two IMTs does not matter. See individual models for supported IMTs.
     *
     * @param imt1 the first IMT
     * @param imt2 the second IMT
     * @return Pearson correlation coefficient between {@code imt1} and
     *         {@code imt2}
     */
    public abstract double value(Imt imt1, Imt imt2);

    /**
     * Create a map of correlation coefficients between the supplied {@code imt}
     * and {@code imts}.
     */
    public Map<Imt, Double> valueMap(Set<Imt> imts, Imt targetImt) {
      Map<Imt, Double> coeffs = imts.stream().collect(toMap(
          imt -> imt,
          imt -> value(imt, targetImt)));
      return new EnumMap<>(coeffs);
    }
  }

  public static void main(String[] args) {
    System.out.println(CorrelationModel.BAKER_JAYARAM_08.value(Imt.SA0P02, Imt.SA0P02));
  }

}
