package gov.usgs.earthquake.nshmp.calc;

import static gov.usgs.earthquake.nshmp.calc.ValueFormat.POISSON_PROBABILITY;
import static gov.usgs.earthquake.nshmp.data.MutableXySequence.emptyCopyOf;
import static java.nio.charset.StandardCharsets.UTF_8;
import static java.nio.file.StandardOpenOption.APPEND;
import static java.util.stream.Collectors.joining;
import static java.util.stream.Collectors.toCollection;
import static java.util.stream.Collectors.toList;

import java.io.IOException;
import java.nio.file.Files;
import java.nio.file.OpenOption;
import java.nio.file.Path;
import java.util.ArrayList;
import java.util.Collection;
import java.util.EnumMap;
import java.util.EnumSet;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.OptionalDouble;
import java.util.Set;
import java.util.function.Function;
import java.util.stream.IntStream;

import com.google.common.collect.FluentIterable;
import com.google.common.collect.Iterables;
import com.google.common.collect.Maps;
import com.google.common.collect.Sets;

import gov.usgs.earthquake.nshmp.Maths;
import gov.usgs.earthquake.nshmp.Text;
import gov.usgs.earthquake.nshmp.Text.Delimiter;
import gov.usgs.earthquake.nshmp.data.MutableXySequence;
import gov.usgs.earthquake.nshmp.data.XySequence;
import gov.usgs.earthquake.nshmp.geo.Location;
import gov.usgs.earthquake.nshmp.gmm.Gmm;
import gov.usgs.earthquake.nshmp.gmm.Imt;
import gov.usgs.earthquake.nshmp.model.HazardModel;
import gov.usgs.earthquake.nshmp.model.RuptureSet;
import gov.usgs.earthquake.nshmp.model.Source;
import gov.usgs.earthquake.nshmp.model.SourceTree;
import gov.usgs.earthquake.nshmp.model.SourceType;

/**
 * Hazard calculation result exporter.
 *
 * @author U.S. Geological Survey
 */
public final class HazardExport {

  private static final String GMM_DIR = "gmm";
  static final String TYPE_DIR = "source";
  private static final String CURVE_FILE_ASCII = "curves.csv";
  private static final String CURVE_FILE_ASCII_TRUNCATED = "curves-truncated.csv";
  private static final String VALUE_FMT = "%.8e";

  private final Path out;
  private final HazardModel model;
  private final CalcConfig config;
  private final boolean exportGmm;
  private final boolean exportSource;

  private final Function<Double, String> valueFormatter;

  private final boolean namedSites;

  private HazardExport(
      HazardModel model,
      CalcConfig config,
      boolean namedSites,
      Path out) throws IOException {

    this.out = out;
    this.model = model;
    this.config = config;
    this.exportGmm = config.output.dataTypes.contains(DataType.GMM);
    this.exportSource = config.output.dataTypes.contains(DataType.SOURCE);

    Function<Double, String> formatter = Text.formatDoubleFunction(VALUE_FMT);
    this.valueFormatter = (config.hazard.valueFormat == POISSON_PROBABILITY)
        ? Maths.annualRateToProbabilityConverter().andThen(formatter)
        : formatter;

    this.namedSites = namedSites;
    init();
  }

  /**
   * Create a new results handler.
   *
   * @param model being run
   * @param config that specifies output options and formats
   * @param namedSites whether the sites being processed contain a 'name' column
   *        that should be preserved in output files
   * @param out results directory; this may have been been modified from
   *        {@code config.output.directory} by calling program
   */
  public static HazardExport create(
      HazardModel model,
      CalcConfig config,
      boolean namedSites,
      Path out) throws IOException {

    return new HazardExport(model, config, namedSites, out);
  }

  /* Initialize output directories. */
  private void init() throws IOException {

    for (Imt imt : config.hazard.imts) {

      Path imtDir = out.resolve(imt.name());
      Files.createDirectory(imtDir);
      Path totalFile = imtDir.resolve(CURVE_FILE_ASCII);
      Path totalFileTruncated = imtDir.resolve(CURVE_FILE_ASCII_TRUNCATED);

      List<Object> headerValues = new ArrayList<>();
      if (namedSites) {
        headerValues.add(Site.Key.NAME);
      }
      headerValues.add(Site.Key.LON);
      headerValues.add(Site.Key.LAT);
      XySequence modelCurve = config.hazard.modelCurves().get(imt);
      headerValues.addAll(modelCurve.xValues().boxed().collect(toList()));

      String header = Text.join(headerValues, Delimiter.COMMA);
      writeLine(totalFile, header);
      writeLine(totalFileTruncated, header);

      if (exportSource) {
        Path typeParent = imtDir.resolve(TYPE_DIR);
        Files.createDirectory(typeParent);
        for (SourceType type : model.types()) {
          Path typeDir = typeParent.resolve(type.name());
          Files.createDirectory(typeDir);
          Path typeFile = typeDir.resolve(CURVE_FILE_ASCII);
          writeLine(typeFile, header);
        }
      }

      if (exportGmm) {
        Path gmmParent = imtDir.resolve(GMM_DIR);
        Files.createDirectory(gmmParent);
        for (Gmm gmm : model.gmms()) {
          Path gmmDir = gmmParent.resolve(gmm.name());
          Files.createDirectory(gmmDir);
          Path gmmFile = gmmDir.resolve(CURVE_FILE_ASCII);
          writeLine(gmmFile, header);
        }
      }
    }
  }

  /* Avoid clobbering exsting result directories via incrementing. */
  static Path createOutputDir(Path dir, OptionalDouble vs30) throws IOException {
    int i = 1;
    Path outDir = dir;
    while (Files.exists(outDir)) {
      outDir = outDir.resolveSibling(dir.getFileName() + "-" + i);
      i++;
    }
    if (vs30.isPresent()) {
      String vs30dir = "vs30-" + ((int) vs30.getAsDouble());

      outDir = outDir.resolve(vs30dir);
    }
    Files.createDirectories(outDir);
    return outDir;
  }

  /**
   * Write a hazard result.
   *
   * @param hazard to write
   */
  public void write(Hazard hazard) throws IOException {

    Set<Gmm> gmms = hazard.model.gmms();
    Set<SourceType> types = hazard.model.types();

    Location location = hazard.site.location();
    String locStr = (namedSites ? hazard.site.name() + "," : "") +
        location.longitude + "," + location.latitude + ",";

    Map<Imt, Map<SourceType, ? extends XySequence>> curvesBySource = exportSource
        ? curvesBySource(hazard)
        : null;

    Map<Imt, Map<Gmm, ? extends XySequence>> curvesByGmm = exportGmm
        ? curvesByGmm(hazard)
        : null;

    for (Entry<Imt, XySequence> imtEntry : hazard.totalCurves.entrySet()) {
      XySequence totalCurve = imtEntry.getValue();

      /* Utility empty line to write if source or gmm is missing at site */
      String emptyLine = locStr + "0.0" + ",0.0".repeat(totalCurve.size() - 1);

      Imt imt = imtEntry.getKey();
      Path imtDir = out.resolve(imt.name());

      String totalLine = locStr + valueStr(totalCurve, valueFormatter);
      String totalLineTrunc = locStr + truncValueStr(totalCurve, valueFormatter, TRUNCATION_LIMIT);

      Path totalFile = imtDir.resolve(CURVE_FILE_ASCII);
      writeLine(totalFile, totalLine, APPEND);
      Path totalFileTrunc = imtDir.resolve(CURVE_FILE_ASCII_TRUNCATED);
      writeLine(totalFileTrunc, totalLineTrunc, APPEND);

      if (exportSource) {
        Map<SourceType, ? extends XySequence> sourceCurveMap = curvesBySource.get(imt);
        Path typeParent = imtDir.resolve(TYPE_DIR);
        for (SourceType type : types) {
          Path typeDir = typeParent.resolve(type.name());
          Path typeFile = typeDir.resolve(CURVE_FILE_ASCII);
          String typeLine = emptyLine;
          if (sourceCurveMap.containsKey(type)) {
            XySequence typeCurve = sourceCurveMap.get(type);
            typeLine = locStr + valueStr(typeCurve, valueFormatter);
          }
          writeLine(typeFile, typeLine, APPEND);
        }
      }

      if (exportGmm) {
        Map<Gmm, ? extends XySequence> gmmCurveMap = curvesByGmm.get(imt);
        Path gmmParent = imtDir.resolve(GMM_DIR);
        for (Gmm gmm : gmms) {
          Path gmmDir = gmmParent.resolve(gmm.name());
          Path gmmFile = gmmDir.resolve(CURVE_FILE_ASCII);
          String gmmLine = emptyLine;
          if (gmmCurveMap.containsKey(gmm)) {
            XySequence gmmCurve = gmmCurveMap.get(gmm);
            gmmLine = locStr + valueStr(gmmCurve, valueFormatter);
          }
          writeLine(gmmFile, gmmLine, APPEND);
        }
      }
    }
  }

  private static String valueStr(
      XySequence curve,
      Function<Double, String> formatter) {

    return curve.yValues()
        .mapToObj(formatter::apply)
        .collect(joining(","));
  }

  private static String truncValueStr(
      XySequence curve,
      Function<Double, String> formatter,
      double cutoff) {

    double[] values = curve.yValues().toArray();
    int limit = truncationIndex(values, cutoff);
    return IntStream.range(0, values.length)
        .mapToObj(i -> (i <= limit) ? formatter.apply(values[i]) : "")
        .collect(joining(","));
  }

  /* No options will write a new file, supply APPEND to append. */
  static void writeLine(Path path, String line, OpenOption... options)
      throws IOException {

    Files.write(
        path,
        (line + System.lineSeparator()).getBytes(UTF_8),
        options);
  }

  /**
   * Derive maps of curves by {@code SourceType} for each {@code Imt}.
   */
  public static Map<Imt, Map<SourceType, ? extends XySequence>> curvesBySource(Hazard hazard) {

    EnumMap<Imt, Map<SourceType, MutableXySequence>> imtMap = Maps.newEnumMap(Imt.class);

    // initialize receiver
    Set<SourceType> types = hazard.ruptureSetCurves.keySet();
    for (Entry<Imt, XySequence> entry : hazard.curves().entrySet()) {
      imtMap.put(entry.getKey(), initCurves(types, MutableXySequence.copyOf(entry.getValue())));
    }

    for (Entry<SourceType, HazardCurveSet> curveSet : hazard.ruptureSetCurves.entries()) {
      for (Entry<Imt, XySequence> typeTotals : curveSet.getValue().totalCurves.entrySet()) {
        imtMap.get(typeTotals.getKey())
            .get(curveSet.getKey())
            .add(typeTotals.getValue());
      }
    }
    return Maps.immutableEnumMap(imtMap);
  }

  // get from config?
  private static final double TRUNCATION_LIMIT = 1e-4;

  /**
   * Returns the index of the first value that is less than {@code cutoff}.
   * Method is used to truncate hazard curves below the supplied {@code cutoff}
   * and assumes {@code values} is a monotonically decreasing array.
   *
   * @param values for which to compute truncation index
   */
  public static int truncationIndex(double[] values, double cutoff) {
    for (int i = 0; i < values.length; i++) {
      if (values[i] < cutoff) {
        return i;
      }
    }
    return values.length - 1;
  }

  /**
   * Derive maps of curves by {@code Gmm} for each {@code Imt} in a
   * {@code Hazard} result.
   */
  public static Map<Imt, Map<Gmm, ? extends XySequence>> curvesByGmm(Hazard hazard) {

    EnumMap<Imt, Map<Gmm, MutableXySequence>> imtMap = Maps.newEnumMap(Imt.class);

    /* Initialize receiver */
    Iterable<RuptureSet<? extends Source>> sources = Iterables.transform(
        hazard.ruptureSetCurves.values(),
        CURVE_SET_TO_SOURCE_SET::apply);
    Set<Gmm> gmms = gmmSet(sources);
    for (Entry<Imt, XySequence> entry : hazard.curves().entrySet()) {
      imtMap.put(entry.getKey(), initCurves(gmms, MutableXySequence.copyOf(entry.getValue())));
    }

    for (HazardCurveSet curveSet : hazard.ruptureSetCurves.values()) {
      for (Entry<Imt, Map<Gmm, MutableXySequence>> imtEntry : curveSet.curveMap.entrySet()) {
        for (Entry<Gmm, MutableXySequence> gmmEntry : imtEntry.getValue().entrySet()) {
          imtMap.get(imtEntry.getKey()).get(gmmEntry.getKey()).add(gmmEntry.getValue());
        }
      }
    }

    return Maps.immutableEnumMap(imtMap);
  }

  /*
   * Scan the supplied source sets for the set of all GMMs used.
   *
   * Note that this is used when consolidating gmm result curves. Results may
   * not span the entire range of GMMs in a model so this is commonly a subset
   */

  private static Set<Gmm> gmmSet(Collection<SourceTree> trees) {
    return trees.stream()
        .flatMap(tree -> tree.gmms().gmms().stream())
        .collect(toCollection(() -> EnumSet.noneOf(Gmm.class)));
  }

  // Consider Stream (above)
  private static Set<Gmm> gmmSet(Iterable<RuptureSet<? extends Source>> ruptureSets) {
    return Sets.immutableEnumSet(
        FluentIterable.from(ruptureSets).transformAndConcat(
            new Function<RuptureSet<? extends Source>, Set<Gmm>>() {
              @Override
              public Set<Gmm> apply(RuptureSet<? extends Source> ruptures) {
                return ruptures.groundMotionModels().gmms();
              }
            }::apply));
  }

  /* Initalize a map of curves, one entry for each of the supplied enum key. */
  // Consider Stream
  private static <K extends Enum<K>> Map<K, MutableXySequence> initCurves(
      final Set<K> keys,
      final MutableXySequence model) {
    return FluentIterable.from(keys).toMap(
        new Function<K, MutableXySequence>() {
          @Override
          public MutableXySequence apply(K key) {
            return emptyCopyOf(model);
          }
        }::apply);
  }

  private static final Function<HazardCurveSet, RuptureSet<? extends Source>> CURVE_SET_TO_SOURCE_SET =
      new Function<HazardCurveSet, RuptureSet<? extends Source>>() {
        @Override
        public RuptureSet<? extends Source> apply(HazardCurveSet curves) {
          return curves.ruptures;
        }
      };

  /**
   * Create the directory at the supplied path, incrementing the path with
   * {@code -#} if the directory already exists. This avoids clobbering and
   * existing directories and their contents. This method is commonly used by
   * calculators prior to supplying an output directory to
   * {@link #create(HazardModel, CalcConfig, boolean, Path)}. Methods either
   * returns the supplied or pissibly incremented path.
   *
   * @param dir to create
   * @return the (possibly) incremented path
   * @throws IOException if an I/O error occurs
   */
  public static Path createDirectory(Path dir) throws IOException {
    int i = 1;
    Path outDir = dir;
    while (Files.exists(outDir)) {
      outDir = outDir.resolveSibling(dir.getFileName() + "-" + i);
      i++;
    }
    Files.createDirectories(outDir);
    return outDir;
  }

}
