package gov.usgs.earthquake.nshmp.calc;

import static com.google.common.base.Preconditions.checkNotNull;
import static com.google.common.base.Preconditions.checkState;
import static gov.usgs.earthquake.nshmp.data.DoubleData.checkInRange;

import com.google.common.base.Strings;
import com.google.common.collect.Range;

import gov.usgs.earthquake.nshmp.NamedLocation;
import gov.usgs.earthquake.nshmp.geo.Location;
import gov.usgs.earthquake.nshmp.gmm.GroundMotionModel;

/**
 * Site characteristics container. Take note of default values; the minimum
 * information required to create a {@code Site} is its location. Not all ground
 * motion models (GMMs) will use all fields and additional fields may be added
 * at any time in the future.
 *
 * <p>Terminology:<ul>
 *
 * <li><b>{@code Vs30}:</b> Average shear wave velocity down to a depth of 30 m,
 * in units of m/s. This value may be <i>inferred</i> or <i>measured</i> (see
 * {@link #vsInferred}).</li>
 *
 * <li><b>{@code z1.0}:</b> Depth to a shear wave velocity of 1.0 km/sec, in
 * units of km.</li>
 *
 * <li><b>{@code z2.5}:</b> Depth to a shear wave velocity of 2.5 km/sec, in
 * units of km.</li>
 *
 * <li><b>{@code zSed}:</b> Continental margin sediment thickness, in units of
 * km.</li>
 *
 * </ul>
 *
 * <p>Both {@code z1.0} and {@code z2.5}, collectively referred to as
 * <i>basin-terms</i>, and {@code zSed} have default values of
 * {@code Double.NaN}. When supplied with the default, those GMMs that support
 * basin terms will use an author defined model, typically based on
 * {@code Vs30}, to compute basin amplification (or demplification).
 *
 * @author U.S. Geological Survey
 */
public class Site {

  /** The name used for a {@code Site} with no supplied name. */
  public static final String NO_NAME = "Unnamed";

  /** Default {@link #vs30()} value: {@code 760 m/s}. */
  public static final double VS_30_DEFAULT = 760.0;

  /** Supported {@link #vs30()} values: {@code [150..2000] m/s}. */
  public static final Range<Double> VS30_RANGE = Range.closed(150.0, 3000.0);

  /** Default {@link #vsInferred()} inferred value: {@code true}. */
  public static final boolean VS_INF_DEFAULT = true;

  /**
   * Default {@link #z1p0()} value: {@code NaN} <br>({@link GroundMotionModel}s
   * will use a default value or model)
   */
  public static final double Z1P0_DEFAULT = Double.NaN;

  /** Supported {@link #z1p0()} values: {@code [0..5] km}. */
  public static final Range<Double> Z1P0_RANGE = Range.closed(0.0, 5.0);

  /**
   * Default {@link #z2p5()} value: {@code NaN} <br>({@link GroundMotionModel}s
   * will use a default value or model)
   */
  public static final double Z2P5_DEFAULT = Double.NaN;

  /** Supported {@link #z2p5()} values: {@code [0..10] km}. */
  public static final Range<Double> Z2P5_RANGE = Range.closed(0.0, 10.0);

  /**
   * Default {@link #zSed()} value: {@code NaN} <br>({@link GroundMotionModel}s
   * will use a default value or model)
   */
  public static final double ZSED_DEFAULT = Double.NaN;

  /** Supported {@link #zSed()} values: {@code [0..25] km}. */
  public static final Range<Double> ZSED_RANGE = Range.closed(0.0, 25.0);

  private final String name;
  private final Location location;
  private final double vs30;
  private final boolean vsInferred;
  private final double z1p0;
  private final double z2p5;
  private final double zSed;

  private Site(Builder builder) {
    this.name = builder.name;
    this.location = builder.location;
    this.vs30 = builder.vs30;
    this.vsInferred = builder.vsInferred;
    this.z1p0 = builder.z1p0;
    this.z2p5 = builder.z2p5;
    this.zSed = builder.zSed;
  }

  /** The site name. */
  public final String name() {
    return name;
  }

  /**
   * The location of this site. This is the only field that must be explicitely
   * set in a {@link Builder}.
   */
  public final Location location() {
    return location;
  }

  /**
   * The average shear-wave velocity down to 30 meters depth.
   *
   * <p>Default: {@code 760.0 m/s}
   */
  public final double vs30() {
    return vs30;
  }

  /**
   * Whether Vs30 was inferred, {@code true}, or measured, {@code false}.
   *
   * <p>Default: {@code true} (inferred)
   */
  public final boolean vsInferred() {
    return vsInferred;
  }

  /**
   * Depth to the shear-wave velocity horizon of 1.0 km/s, in km.
   *
   * <p>Default: {@code NaN} <br>({@link GroundMotionModel}s will use a default
   * value or model)
   */
  public final double z1p0() {
    return z1p0;
  }

  /**
   * Depth to the shear-wave velocity horizon of 2.5 km/s, in km.
   *
   * <p>Default: {@code NaN} <br>({@link GroundMotionModel}s will use a default
   * value or model)
   */
  public final double z2p5() {
    return z2p5;
  }

  /**
   * Continental margin sediment thickness, in km.
   *
   * <p>Default: {@code NaN} <br>({@link GroundMotionModel}s will use a default
   * value or model)
   */
  public final double zSed() {
    return zSed;
  }

  @Override
  public String toString() {
    return new StringBuilder(Strings.padEnd(name, 28, ' '))
        .append(String.format("%.3f %.3f ", location.longitude, location.latitude))
        .append(paramString(vs30, vsInferred, z1p0, z2p5, zSed))
        .toString();
  }

  private static String paramString(
      double vs30, boolean vsInf,
      double z1p0, double z2p5,
      double zSed) {
    return String.format(
        "vs30=%s %s z1.0=%s z2.5=%s zSed=%s",
        vs30, vsInf ? "inferred " : "measured ", z1p0, z2p5, zSed);
  }

  /**
   * Return a new single-use {@link Builder}.
   */
  public static Builder builder() {
    return new Builder();
  }

  /**
   * A reusable {@code Site} builder. In the absence of specifying any site
   * characteristics other than location (required), a default {@code Site} is
   * returned by {@code build()}. Builder instances may be obtained via
   * {@link Site#builder()}.
   *
   * @see Site for default values
   */
  public static class Builder {

    private String name = NO_NAME;
    private Location location;
    private double vs30 = VS_30_DEFAULT;
    private boolean vsInferred = VS_INF_DEFAULT;
    private double z1p0 = Z1P0_DEFAULT;
    private double z2p5 = Z2P5_DEFAULT;
    private double zSed = ZSED_DEFAULT;

    boolean built = false;

    private Builder() {}

    /**
     * The name of the {@code Site}. Prior to setting, the supplied name is
     * stripped of commas and truncated at 72 characters.
     */
    public Builder name(String name) {
      this.name = cleanName(checkNotNull(name));
      return this;
    }

    /** The location of the {@code Site}. */
    public Builder location(Location loc) {
      this.location = checkNotNull(loc);
      return this;
    }

    /** The location and name of the {@code Site}. */
    public Builder location(NamedLocation loc) {
      this.location = checkNotNull(loc.location());
      this.name = loc.name();
      return this;
    }

    /** The average shear-wave velocity down to 30 meters depth. */
    public Builder vs30(double vs30) {
      this.vs30 = checkInRange(VS30_RANGE, Site.Key.VS30, vs30);
      return this;
    }

    /** Whether Vs30 was inferred, {@code true}, or measured {@code false}. */
    public Builder vsInferred(boolean vsInferred) {
      this.vsInferred = vsInferred;
      return this;
    }

    /** Depth to the shear-wave velocity horizon of 1.0 km/s, in km. */
    public Builder z1p0(double z1p0) {
      if (Double.isNaN(z1p0)) {
        this.z1p0 = z1p0;
      } else {
        this.z1p0 = checkInRange(Z1P0_RANGE, Site.Key.Z1P0, z1p0);
      }
      return this;
    }

    /** Depth to the shear-wave velocity horizon of 2.5 km/s, in km. */
    public Builder z2p5(double z2p5) {
      if (Double.isNaN(z2p5)) {
        this.z2p5 = z2p5;
      } else {
        this.z2p5 = checkInRange(Z2P5_RANGE, Site.Key.Z2P5, z2p5);
      }
      return this;
    }

    /** Sediment thickness, in km. */
    public Builder zSed(double zSed) {
      if (Double.isNaN(zSed)) {
        this.zSed = zSed;
      } else {
        this.zSed = checkInRange(ZSED_RANGE, Site.Key.ZSED, zSed);
      }
      return this;
    }

    /**
     * Build the {@code Site}.
     */
    public Site build() {
      checkState(!built, "This instance has already been built");
      checkNotNull(location, "Site location not set");
      built = true;
      return new Site(this);
    }
  }

  private static final int MAX_NAME_LENGTH = 72;

  private static String cleanName(String name) {
    name = name.replaceAll(",", "");
    return name.length() > MAX_NAME_LENGTH ? name.substring(0, MAX_NAME_LENGTH) : name;
  }

  /**
   * Strings used to represent site properties in CSV and JSON files.
   */
  public static final class Key {

    private Key() {}

    /** The site name key. */
    public static final String NAME = "name";
    /** The site latitude key. */
    public static final String LAT = "lat";
    /** The site longitude key. */
    public static final String LON = "lon";
    /** The site Vs30 key. */
    public static final String VS30 = "vs30";
    /** The site Vs inferred key. */
    public static final String VS_INF = "vsInf";
    /** The site z1.0 key. */
    public static final String Z1P0 = "z1p0";
    /** The site z2.5 key. */
    public static final String Z2P5 = "z2p5";
    /** The site zSed key. */
    public static final String ZSED = "zSed";
  }

}
