package gov.usgs.earthquake.nshmp.calc;

import static gov.usgs.earthquake.nshmp.Text.LOG_INDENT;
import static gov.usgs.earthquake.nshmp.geo.BorderType.MERCATOR_LINEAR;
import static java.util.stream.Collectors.toUnmodifiableList;

import java.io.IOException;
import java.nio.file.Path;
import java.util.List;
import java.util.Optional;
import java.util.OptionalDouble;
import java.util.Set;
import java.util.stream.Stream;

import gov.usgs.earthquake.nshmp.data.DelimitedData;
import gov.usgs.earthquake.nshmp.data.DelimitedData.Record;
import gov.usgs.earthquake.nshmp.geo.Bounds;
import gov.usgs.earthquake.nshmp.geo.GriddedRegion;
import gov.usgs.earthquake.nshmp.geo.Location;
import gov.usgs.earthquake.nshmp.geo.LocationList;
import gov.usgs.earthquake.nshmp.geo.Region;
import gov.usgs.earthquake.nshmp.geo.Regions;
import gov.usgs.earthquake.nshmp.geo.json.Feature;
import gov.usgs.earthquake.nshmp.geo.json.FeatureCollection;
import gov.usgs.earthquake.nshmp.geo.json.GeoJson;
import gov.usgs.earthquake.nshmp.geo.json.GeoJson.Type;
import gov.usgs.earthquake.nshmp.geo.json.Properties;
import gov.usgs.earthquake.nshmp.model.SiteData;

/**
 * Factory class that generates {@code Site} lists from different resources.
 *
 * @author U.S. Geological Survey
 */
public final class Sites {

  private Sites() {}

  /**
   * Create an unmodifiable list of sites from the comma-delimted site file
   * designated by {@code path}. Site parameter values in a sites file take
   * precedence over defaults and model values in the supplied siteData.
   *
   * @param path to comma-delimited site data file
   * @param siteData site specific data from model
   * @param vs30 optional vs30 value used when running multiple site classes
   * @throws IOException if a problem is encountered
   */
  public static List<Site> fromCsv(
      Path path,
      SiteData siteData,
      OptionalDouble vs30) throws IOException {

    DelimitedData dd = DelimitedData.comma(path);
    try (Stream<Record> stream = dd.records()) {
      return stream.map(record -> recordToSite(
          record, dd.columnKeys(), siteData, vs30))
          .collect(toUnmodifiableList());
    }
  }

  private static Site recordToSite(
      Record record,
      Set<String> keys,
      SiteData siteData,
      OptionalDouble vs30) {

    /*
     * Site parameter precedence:
     *
     * (4) Defaults from Site.java, (3) site-data from model, (2) sites file,
     * and (1) Vs30 from config.vs30s[]
     *
     * If a sites file contains columns for basin depths, a 'null' value string
     * indicates a GMM should do whatever it defines as default behavior and
     * will override values that may be included in the site data of a model. If
     * the value is empty, then whatever values from (3) or (4) will be used. If
     * the config.hazard.vs30s array is not empty, this indicates we're running
     * batch jobs across multiple site classes and the vs30 of the job will
     * override all lower precedence vs30 values.
     */

    Location loc = Location.create(
        record.getDouble(Site.Key.LON),
        record.getDouble(Site.Key.LAT));

    Site.Builder builder = initBuilder(loc, siteData);

    // Override with values from sites file
    if (keys.contains(Site.Key.NAME)) {
      builder.name(record.get(Site.Key.NAME));
    }
    if (keys.contains(Site.Key.VS30)) {
      builder.vs30(record.getDouble(Site.Key.VS30));
    }
    if (keys.contains(Site.Key.VS_INF)) {
      builder.vsInferred(record.getBoolean(Site.Key.VS_INF));
    }
    if (keys.contains(Site.Key.Z1P0)) {
      builder.z1p0(record.getDouble(Site.Key.Z1P0));
    }
    if (keys.contains(Site.Key.Z2P5)) {
      builder.z2p5(record.getDouble(Site.Key.Z2P5));
    }

    // Override with site class batch value
    vs30.ifPresent(builder::vs30);

    return builder.build();
  }

  /*
   * Initialize a builder with location and possible site specific values for
   * basin and other terms from the model being run.
   */
  private static Site.Builder initBuilder(
      Location loc,
      SiteData siteData) {

    Site.Builder builder = Site.builder().location(loc);
    SiteData.Values values = siteData.get(loc);
    values.z1p0.ifPresent(builder::z1p0);
    values.z2p5.ifPresent(builder::z2p5);
    values.zSed.ifPresent(builder::zSed);
    return builder;
  }

  /**
   * Create an unmodifiable list of sites from the GeoJSON site file designated
   * by {@code path}. The GeoJSON may either be a feature collection of point
   * features or a polygon feature and optional bounds feature.
   *
   * <p>A GeoJSON of points may include site parameters as properties of each
   * feature; these will override default values and model values in the
   * supplied siteData.
   *
   * @param path to GeoJson site data file
   * @param siteData site specific data from model
   * @param vs30 optional vs30 value used when running multiple site classes
   * @throws IOException if a problem is encountered
   */
  public static List<Site> fromGeoJson(
      Path path,
      SiteData siteData,
      OptionalDouble vs30) throws IOException {

    FeatureCollection fc = GeoJson.from(path).toFeatureCollection();
    List<Feature> features = fc.features();

    if (features.get(0).type() == Type.POINT) {
      return fromPoints(features, siteData, vs30);
    } else {
      return fromPolygon(features, siteData, vs30);
    }
  }

  private static List<Site> fromPoints(
      List<Feature> features,
      SiteData siteData,
      OptionalDouble vs30) {

    return features.stream()
        .map(feature -> fromPoint(feature, siteData, vs30))
        .collect(toUnmodifiableList());
  }

  private static Site fromPoint(Feature feature, SiteData siteData, OptionalDouble vs30) {
    Location loc = feature.asPoint();
    Site.Builder builder = initBuilder(loc, siteData);
    Properties props = feature.properties();
    props.getString(Site.Key.NAME).ifPresent(builder::name);
    props.getDouble(Site.Key.VS30).ifPresent(builder::vs30);
    props.getBoolean(Site.Key.VS_INF).ifPresent(builder::vsInferred);
    props.getDouble(Site.Key.Z1P0).ifPresent(builder::z1p0);
    props.getDouble(Site.Key.Z2P5).ifPresent(builder::z2p5);
    vs30.ifPresent(builder::vs30);
    return builder.build();
  }

  private static List<Site> fromPolygon(
      List<Feature> features,
      SiteData siteData,
      OptionalDouble vs30) {

    Optional<Bounds> mapBounds = Optional.empty();
    String boundsName = "";

    /* Possibly set map extents. */
    int mapRegionIndex = 0;
    if (features.size() > 1) {
      mapRegionIndex = 1;
      Feature mapPoly = features.get(0);
      LocationList mapPolyBorder = mapPoly.asPolygonBorder();
      mapBounds = Optional.of(GeoJson.validateBoundingBox(mapPolyBorder).bounds());
      boundsName = mapPoly.properties().getString(Site.Key.NAME).orElse("Map Extents");
    }

    Feature sitesPoly = features.get(mapRegionIndex);
    LocationList sitesPolyBorder = sitesPoly.asPolygonBorder();
    Properties properties = sitesPoly.properties();
    String mapName = properties.getString(Site.Key.NAME).orElse("Unnamed Map");

    /*
     * We special case a 5-coordinate border that defines a mercator recangle so
     * as to create a region that includes sites on the north and east borders.
     */

    Region calcRegion = null;
    try {
      Bounds b = GeoJson.validateBoundingBox(sitesPolyBorder).bounds();
      calcRegion = Regions.createRectangular(mapName, b.min, b.max);
    } catch (IllegalArgumentException iae) {
      calcRegion = Regions.create(mapName, sitesPolyBorder, MERCATOR_LINEAR);
    }

    double spacing = properties.getDouble("spacing").orElseThrow();

    Region mapRegion = calcRegion;
    if (mapBounds.isPresent()) {
      Bounds b = mapBounds.orElseThrow();
      Region r = Regions.createRectangular(boundsName, b.min, b.max);
      mapRegion = Regions.intersectionOf(mapName, r, calcRegion);
    }

    GriddedRegion region = Regions.toGridded(
        mapRegion,
        spacing, spacing,
        GriddedRegion.ANCHOR_0_0);

    return region.nodes().stream()
        .map(loc -> locationToSite(loc, siteData, vs30))
        .collect(toUnmodifiableList());
  }

  private static Site locationToSite(
      Location loc,
      SiteData siteData,
      OptionalDouble vs30) {

    Site.Builder site = initBuilder(loc, siteData);
    vs30.ifPresent(site::vs30);
    return site.build();
  }

  private static final int TO_STRING_LIMIT = 5;
  private static final String SITE_INDENT = LOG_INDENT + "       ";

  /** Site list string representation for logs. */
  public static String toString(List<Site> sites) {
    StringBuilder sb = new StringBuilder();
    sb.append("Sites [size=")
        .append(sites.size())
        .append("]");
    sites.stream()
        .limit(TO_STRING_LIMIT)
        .forEach(site -> sb.append(SITE_INDENT).append(site));
    if (sites.size() > TO_STRING_LIMIT) {
      int delta = sites.size() - TO_STRING_LIMIT;
      sb.append(SITE_INDENT).append("... and ").append(delta).append(" more ...");
    }
    return sb.toString();
  }

}
