package gov.usgs.earthquake.nshmp.data;

import static com.google.common.base.Preconditions.checkElementIndex;

import java.util.Arrays;
import java.util.Iterator;

/**
 * Mutable variant of {@code XySequence}.
 *
 * @author U.S. Geological Survey
 */
final class MutableArrayXySequence extends ArrayXySequence implements MutableXySequence {

  MutableArrayXySequence(double[] xs, double[] ys) {
    super(xs, ys);
  }

  MutableArrayXySequence(XySequence sequence, boolean clear) {
    super(sequence, clear);
  }

  @Override
  public XyPoint min() {
    return new MutablePoint(0);
  }

  @Override
  public XyPoint max() {
    return new MutablePoint(size() - 1);
  }

  @Override
  public MutableXySequence set(int index, double value) {
    checkElementIndex(index, xs.length);
    ys[index] = value;
    return this;
  }

  @Override
  public MutableXySequence add(double term) {
    DoubleData.add(term, ys);
    return this;
  }

  @Override
  public MutableXySequence add(double[] ys) {
    DoubleData.add(this.ys, ys);
    return this;
  }

  @Override
  public MutableXySequence add(XySequence sequence) {
    /* safe covariant cast */
    DoubleData.uncheckedAdd(ys, validateSequence((ArrayXySequence) sequence).ys);
    return this;
  }

  @Override
  public MutableXySequence multiply(double[] ys) {
    DoubleData.multiply(this.ys, ys);
    return this;
  }

  @Override
  public MutableXySequence multiply(double scale) {
    DoubleData.multiply(scale, ys);
    return this;
  }

  @Override
  public MutableXySequence multiply(XySequence sequence) {
    // safe covariant cast
    DoubleData.uncheckedMultiply(ys, validateSequence((ArrayXySequence) sequence).ys);
    return this;
  }

  @Override
  public MutableXySequence complement() {
    DoubleData.complement(ys);
    return this;
  }

  @Override
  public MutableXySequence clear() {
    Arrays.fill(ys, 0.0);
    return this;
  }

  @Override
  public MutableXySequence trim() {
    // safe covariant cast
    return (MutableXySequence) Sequences.trim(this);
  }

  class MutablePoint extends Point {

    MutablePoint(int index) {
      super(index);
    }

    @Override
    public void set(double y) {
      MutableArrayXySequence.this.set(index, y);
    }
  }

  @Override
  public Iterator<XyPoint> iterator() {
    return new XyIterator();
  }

  class XyIterator implements Iterator<XyPoint> {
    private int caret = 0;

    @Override
    public boolean hasNext() {
      return caret < size();
    }

    @Override
    public XyPoint next() {
      return new MutablePoint(caret++);
    }

    @Override
    public void remove() {
      throw new UnsupportedOperationException();
    }
  }

}
