package gov.usgs.earthquake.nshmp.geo;

import static com.google.common.base.Preconditions.checkArgument;
import static com.google.common.base.Preconditions.checkState;
import static gov.usgs.earthquake.nshmp.Text.WHITESPACE_SPLITTER;
import static java.util.stream.Collectors.toUnmodifiableList;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.List;
import java.util.stream.Collectors;
import java.util.stream.StreamSupport;

import com.google.common.collect.Iterables;
import com.google.common.collect.Lists;
import com.google.common.primitives.Doubles;

/**
 * A list of locations.
 *
 * <p>Implementations provided in this package are guaranteed to be immutable.
 * All calls to methods or iterator methods that would cause structural
 * modifications throw an {@code UnsupportedOperationException}.
 *
 * <p>A {@code LocationList} must contain at least 1 location and does not
 * permit {@code null} elements.
 *
 * <p>Consider using a {@link LocationList#builder() builder} if a list is being
 * compiled from numerous locations that are not known in advance, otherwise,
 * use static factory methods. A variety of additional methods exist to create
 * modified forms of a location list (e.g. {@link #resample(double)} ,
 * {@link #reverse()}, and {@link #translate(LocationVector)}.
 *
 * @author U.S. Geological Survey
 */
public interface LocationList extends List<Location> {

  /**
   * Return the first location in the list.
   */
  default Location first() {
    return get(0);
  }

  /**
   * Return the last location in the list.
   */
  default Location last() {
    return get(size() - 1);
  }

  /**
   * Lazily compute the horizontal length of the list in km. Method uses the
   * {@link Locations#horzDistanceFast(Location, Location)} algorithm and
   * ignores depth variation between locations, computing length as though all
   * locations have a depth of 0 km. Repeat calls to this method will
   * recalculate the length each time.
   *
   * @return the length of the line connecting all locations in this list,
   *         ignoring depth variations, or 0 if list only contains 1 location
   */
  default double length() {
    double sum = 0.0;
    if (size() == 1) {
      return sum;
    }
    Location prev = first();
    for (Location loc : Iterables.skip(this, 1)) {
      sum += Locations.horzDistanceFast(prev, loc);
      prev = loc;
    }
    return sum;

    // Consider using this instead; check tests
    // if (size() == 1) {
    // return 0.0;
    // }
    // return IntStream.range(0, size() - 1)
    // .mapToDouble(i -> Locations.horzDistanceFast(get(i), get(i + 1)))
    // .sum();
  }

  /**
   * Lazily compute the average depth of the locations in the list.
   */
  default double depth() {
    double depth = 0.0;
    for (Location loc : this) {
      depth += loc.depth;
    }
    return depth / size();
  }

  /**
   * Compute the distances between each location in the list. The length of the
   * returned array is {@code size() - 1} and may be empty if this list only
   * contains a single location.
   */
  default double[] distances() {
    double[] distances = new double[size() - 1];
    for (int i = 0; i < size() - 1; i++) {
      distances[i] = Locations.horzDistanceFast(get(i), get(i + 1));
    }
    return distances;
  }

  /**
   * Lazily compute the bounds of the locations in the list.
   */
  default Bounds bounds() {
    return Locations.bounds(this);
  }

  /**
   * Split this list into sub-lists of desired {@code length}. The actual length
   * of the returned lists will likely differ, as the lengths of the sub-lists
   * are adjusted up or down to a value closest to the target length that yields
   * sub-lists of equal length.
   *
   * <p>If {@code this.length() < length}, this list is returned.
   *
   * @param length target length of sub-lists
   */
  default List<LocationList> split(double length) {
    checkArgument(
        Doubles.isFinite(length) && length > 0.0,
        "Length must be positive, real number");

    double totalLength = this.length();
    if (totalLength <= length) {
      return List.of(this);
    }

    List<LocationList> partitions = new ArrayList<>();
    double partitionLength = totalLength / Math.rint(totalLength / length);

    double[] distances = distances();
    LocationList.Builder partition = LocationList.builder();
    double residual = 0.0;
    for (int i = 0; i < distances.length; i++) {
      Location start = get(i);
      partition.add(start);
      double distance = distances[i] + residual;
      while (partitionLength < distance) {
        /*
         * Catch the edge case where, on the last segment of a trace that needs
         * to be partitioned we just undershoot the last point. In the absence
         * of this we can end up with a final section of length ≈ 0.
         *
         */
        if (i == distances.length - 1 && distance < 1.5 * partitionLength) {
          break;
        }
        LocationVector v = LocationVector.create(start, get(i + 1));
        Location end = Locations.location(start, v.azimuth, partitionLength - residual);

        partition.add(end);
        partitions.add(partition.build());
        partition = LocationList.builder();
        partition.add(end);
        start = end;
        distance -= partitionLength;
        residual = 0.0;
      }
      residual = distance;
    }
    partition.add(last());
    partitions.add(partition.build());
    return partitions;
  }

  /**
   * Return a new list resampled to the desired maximum spacing. The actual
   * spacing of the returned list will likely differ, as spacing is adjusted
   * down to maintain uniform divisions. The original vertices are also not
   * preserved such that some corners might be clipped if {@code spacing} is
   * large. Buyer beware.
   *
   * <p>If {@code length() < spacing}, this list is returned.
   *
   * @param spacing resample interval
   */
  default LocationList resample(double spacing) {
    checkArgument(
        Doubles.isFinite(spacing) && spacing > 0.0,
        "Spacing must be positive, real number");

    double length = length();
    if (length <= spacing) {
      return this;
    }

    /*
     * Consider using rint() and/or providing a maxFlag which will keep the
     * actual spacing closer to the target spacing, albeit sometimes larger.
     *
     * Consider using round() instead of ceil() which will in some cases be
     * closer to the target spacing, albeit greater.
     */

    spacing = length / Math.ceil(length / spacing);
    List<Location> resampled = Lists.newArrayList();
    Location start = first();
    resampled.add(start);
    double walker = spacing;
    for (Location loc : Iterables.skip(this, 1)) {
      LocationVector v = LocationVector.create(start, loc);
      double distance = Locations.horzDistanceFast(start, loc);
      while (walker <= distance) {
        resampled.add(Locations.location(start, v.azimuth, walker));
        walker += spacing;
      }
      start = loc;
      walker -= distance;
    }
    // replace last point to be exact
    resampled.set(resampled.size() - 1, last());
    return LocationList.copyOf(resampled);
  }

  /**
   * Return a new list with locations in reverse order.
   */
  default LocationList reverse() {
    return copyOf(Lists.reverse(this));
  }

  /**
   * Return a new {@code LocationList} with each location translated according
   * to the supplied vector.
   *
   * @param vector to translate list by
   */
  default LocationList translate(LocationVector vector) {
    return copyOf(this.stream()
        .map(loc -> Locations.location(loc, vector))
        .collect(toUnmodifiableList()));

  }

  /**
   * Create a new list containing the supplied {@code locations}.
   *
   * @param locations to populate list with
   * @throws IllegalArgumentException if {@code locations} is empty
   */
  static LocationList of(Location... locations) {
    return new RegularLocationList(List.of(locations));
  }

  /**
   * Create a new list containing the supplied {@code locations}. If possible,
   * this method avoids copying the supplied locations.
   *
   * @param locations to populate list with
   * @throws IllegalArgumentException if {@code locations} is empty
   */
  static LocationList copyOf(Iterable<Location> locations) {
    if (locations instanceof LocationList) {
      return (LocationList) locations;
    } else if (locations instanceof List) {
      return new RegularLocationList((List<Location>) locations);
    }
    return new RegularLocationList(
        StreamSupport.stream(locations.spliterator(), false)
            .collect(Collectors.toUnmodifiableList()));
  }

  /**
   * Create a new list from the supplied {@code String}. This method assumes
   * that {@code s} is formatted as one or more space-delimited
   * {@code [longitude, latitude, depth]} tuples, for example:
   *
   * <pre>
   *   -118.0,34.0,0.0
   *   -118.0,34.1,0.0
   *   -118.0,34.2,0.0
   * </pre>
   *
   * @param s string to parse
   * @see Location#fromString(String)
   */
  static LocationList fromString(String s) {
    return copyOf(WHITESPACE_SPLITTER.splitToStream(s)
        .map(Location::fromString)
        .collect(toUnmodifiableList()));
  }

  /**
   * Return a new builder.
   */
  static Builder builder() {
    return new Builder();
  }

  /**
   * A reusable {@code LocationList} builder. Repeat calls to {@code build()}
   * will return multiple lists in series with each new list containing all the
   * elements of the one before it. Builder does not permit the addition of
   * {@code null} elements, however, null checking is lazy and not performed
   * until {@code build()} is called.
   *
   * <p>Use {@link LocationList#builder()} to create new builder instances.
   */
  public static final class Builder {

    private List<Location> locations = new ArrayList<>();

    private Builder() {}

    /**
     * Add a location.
     *
     * @param location to add
     * @return this {@code Builder}
     */
    public Builder add(Location location) {
      locations.add(location);
      return this;
    }

    /**
     * Add a location with the specified longitude and latitude, and a depth of
     * 0 km.
     *
     * @param longitude in decimal degrees
     * @param latitude in decimal degrees
     * @return this {@code Builder}
     * @throws IllegalArgumentException if any values are out of range
     * @see Coordinates
     */
    public Builder add(double longitude, double latitude) {
      locations.add(Location.create(longitude, latitude));
      return this;
    }

    /**
     * Add a location with the specified longitude, latitude, and depth.
     *
     * @param longitude in decimal degrees
     * @param latitude in decimal degrees
     * @param depth in km (positive down)
     * @return this {@code Builder}
     * @throws IllegalArgumentException if any values are out of range
     * @see Coordinates
     */
    public Builder add(double longitude, double latitude, double depth) {
      locations.add(Location.create(longitude, latitude, depth));
      return this;
    }

    /**
     * Add each of the supplied locations.
     *
     * @param locations to add
     * @return this {@code Builder}
     */
    public Builder add(Location... locations) {
      return addAll(Arrays.asList(locations));
    }

    /**
     * Add each of the supplied locations.
     *
     * @param locations to add
     * @return this {@code Builder}
     */
    public Builder addAll(Collection<Location> locations) {
      this.locations.addAll(locations);
      return this;
    }

    /**
     * Return a newly created {@code LocationList} based on the contents of the
     * {@code Builder}.
     *
     * @return a new {@code LocationList}
     * @throws IllegalStateException if the list is empty
     * @throws NullPointerException if the list contains {@code null} elements
     */
    public LocationList build() {
      checkState(locations.size() > 0, "Location lists may not by empty");
      return new RegularLocationList(locations);
    }
  }
}
