package gov.usgs.earthquake.nshmp.gmm;

import static gov.usgs.earthquake.nshmp.gmm.Gmm.AB_06_PRIME;
import static gov.usgs.earthquake.nshmp.gmm.Gmm.ASK_14;
import static gov.usgs.earthquake.nshmp.gmm.Gmm.ASK_14_BASIN;
import static gov.usgs.earthquake.nshmp.gmm.Gmm.ATKINSON_08_PRIME;
import static gov.usgs.earthquake.nshmp.gmm.Gmm.ATKINSON_10;
import static gov.usgs.earthquake.nshmp.gmm.Gmm.BCHYDRO_12_SLAB;
import static gov.usgs.earthquake.nshmp.gmm.Gmm.BSSA_14;
import static gov.usgs.earthquake.nshmp.gmm.Gmm.BSSA_14_BASIN;
import static gov.usgs.earthquake.nshmp.gmm.Gmm.CAMPBELL_03;
import static gov.usgs.earthquake.nshmp.gmm.Gmm.CB_14;
import static gov.usgs.earthquake.nshmp.gmm.Gmm.CB_14_BASIN;
import static gov.usgs.earthquake.nshmp.gmm.Gmm.CY_14;
import static gov.usgs.earthquake.nshmp.gmm.Gmm.CY_14_BASIN;
import static gov.usgs.earthquake.nshmp.gmm.Gmm.FRANKEL_96;
import static gov.usgs.earthquake.nshmp.gmm.Gmm.IDRISS_14;
import static gov.usgs.earthquake.nshmp.gmm.Gmm.NGA_EAST_2018;
import static gov.usgs.earthquake.nshmp.gmm.Gmm.NGA_EAST_2023;
import static gov.usgs.earthquake.nshmp.gmm.Gmm.NGA_EAST_SEEDS_2018;
import static gov.usgs.earthquake.nshmp.gmm.Gmm.NGA_EAST_SEEDS_2023;
import static gov.usgs.earthquake.nshmp.gmm.Gmm.PEZESHK_11;
import static gov.usgs.earthquake.nshmp.gmm.Gmm.SILVA_02;
import static gov.usgs.earthquake.nshmp.gmm.Gmm.SOMERVILLE_01;
import static gov.usgs.earthquake.nshmp.gmm.Gmm.TORO_97_MW;
import static gov.usgs.earthquake.nshmp.gmm.Gmm.TP_05;
import static gov.usgs.earthquake.nshmp.gmm.Gmm.WONG_15;

import java.util.Map;
import java.util.Map.Entry;
import java.util.stream.Collectors;

import gov.usgs.earthquake.nshmp.gmm.GmmInput.Constraints;
import gov.usgs.earthquake.nshmp.tree.LogicTree;

/**
 * Convenience class for combined GMM implementations that compute weight
 * averaged ground motions and sigmas for a logic tree of ground motion models.
 * These are NOT intended for use in hazard calculations, but are made available
 * to support comparative deterministic analysis of GMM logic trees.
 */
class CombinedGmm implements GroundMotionModel {

  private static final String NAME = "Combined: ";
  private final Map<Gmm, Double> gmms;
  private final Imt imt;

  /* Supply map of ground motion models initialized to the required IMT. */
  private CombinedGmm(Imt imt, Map<Gmm, Double> gmms) {
    this.gmms = gmms;
    this.imt = imt;
  }

  @Override
  public LogicTree<GroundMotion> calc(GmmInput in) {

    LogicTree.Builder<GroundMotion> builder = LogicTree.builder(NAME);
    gmms.keySet().forEach(key -> builder.addBranch(
        key.name(),
        GroundMotions.combine(key.instance(imt).calc(in)),
        gmms.get(key)));
    return builder.build();
  }

  static Map<Gmm, GroundMotionModel> instancesForImt(
      Imt imt,
      Map<Gmm, Double> gmms) {

    return gmms.entrySet().stream()
        .collect(Collectors.toMap(
            Entry::getKey,
            entry -> entry.getKey().instance(imt)));
  }

  /*
   * Implementations. For each, coefficients are only used to get the set of
   * supported IMTs and therefore reference a model that supports the
   * intersection of all support IMTs.
   */

  private static final Map<Gmm, Double> HI_DEEP_2021 = Map.of(
      ATKINSON_10, 0.4,
      BCHYDRO_12_SLAB, 0.4,
      WONG_15, 0.2);

  static final class HawaiiDeep2021 extends CombinedGmm {

    static final String NAME = CombinedGmm.NAME + "Hawaii Deep 2021 (2.0)";
    static final Constraints CONSTRAINTS = Atkinson_2010.CONSTRAINTS;
    static final CoefficientContainer COEFFS = Atkinson_2010.COEFFS;

    HawaiiDeep2021(Imt imt) {
      super(imt, HI_DEEP_2021);
    }
  }

  private static final Map<Gmm, Double> HI_SHALLOW_2021 = Map.of(
      ATKINSON_10, 0.2,
      ASK_14, 0.2,
      BSSA_14, 0.2,
      CB_14, 0.2,
      CY_14, 0.2);

  static final class HawaiiShallow2021 extends CombinedGmm {

    static final String NAME = CombinedGmm.NAME + "Hawaii Shallow 2021 (2.0)";
    static final Constraints CONSTRAINTS = Atkinson_2010.CONSTRAINTS;
    static final CoefficientContainer COEFFS = Atkinson_2010.COEFFS;

    HawaiiShallow2021(Imt imt) {
      super(imt, HI_SHALLOW_2021);
    }
  }

  static final Map<Gmm, Double> CEUS_2014_FAULT = Map.of(
      AB_06_PRIME, 0.22,
      ATKINSON_08_PRIME, 0.08,
      CAMPBELL_03, 0.11,
      FRANKEL_96, 0.06,
      PEZESHK_11, 0.15,
      SILVA_02, 0.06,
      SOMERVILLE_01, 0.1,
      TP_05, 0.11,
      TORO_97_MW, 0.11);

  /* Need to allow Vs30=3000 through for comparison plots. */
  private static final Constraints CEUS_2014_CONSTRAINTS = Constraints.builder()
      .withDefaults()
      .build();

  /* Fault variant that includes Somerville. */
  static final class Ceus2014 extends CombinedGmm {

    static final String NAME = CombinedGmm.NAME + "CEUS 2014 (4.*)";
    static final Constraints CONSTRAINTS = CEUS_2014_CONSTRAINTS;
    static final CoefficientContainer COEFFS = FrankelEtAl_1996.COEFFS;

    Ceus2014(Imt imt) {
      super(imt, CEUS_2014_FAULT);
    }

    @Override
    public LogicTree<GroundMotion> calc(GmmInput in) {
      GmmInput.Builder b = GmmInput.builder().fromCopy(in);
      b.vs30(in.vs30 <= 760.0 ? 760.0 : 2000.0);
      return super.calc(b.build());
    }
  }

  /* 5.0 */
  private static final Map<Gmm, Double> CEUS_2018 = Map.of(
      NGA_EAST_2018, 0.667,
      NGA_EAST_SEEDS_2018, 0.333);

  static final class Ceus2018 extends CombinedGmm {

    static final String NAME = CombinedGmm.NAME + "CEUS 2018 (5.0)";
    static final Constraints CONSTRAINTS = NgaEast.CONSTRAINTS;
    static final CoefficientContainer COEFFS = NgaEast.COEFFS_SIGMA_PANEL;

    Ceus2018(Imt imt) {
      super(imt, CEUS_2018);
    }
  }

  /* 5.0 Gulf Coast */
  private static final Map<Gmm, Double> CEUS_2018_CPA = Map.of(
      NGA_EAST_2023, 0.667,
      NGA_EAST_SEEDS_2023, 0.333);

  static final class Ceus2018Cpa extends CombinedGmm {

    static final String NAME = CombinedGmm.NAME + "CEUS 2018 (Gulf Coast)";
    static final Constraints CONSTRAINTS = NgaEast.CONSTRAINTS;
    static final CoefficientContainer COEFFS = NgaEast.COEFFS_SIGMA_PANEL;

    Ceus2018Cpa(Imt imt) {
      super(imt, CEUS_2018_CPA);
    }
  }

  private static final Map<Gmm, Double> WUS_2014_4P1 = Map.of(
      ASK_14, 0.22,
      BSSA_14, 0.22,
      CB_14, 0.22,
      CY_14, 0.22,
      IDRISS_14, 0.12);

  /* 4.1 */
  static final class Wus2014_4p1 extends CombinedGmm {

    static final String NAME = CombinedGmm.NAME + "WUS 2014 (4.1)";
    static final Constraints CONSTRAINTS = CampbellBozorgnia_2014.CONSTRAINTS;
    static final CoefficientContainer COEFFS = CampbellBozorgnia_2014.COEFFS;

    Wus2014_4p1(Imt imt) {
      super(imt, WUS_2014_4P1);
    }
  }

  private static final Map<Gmm, Double> WUS_2014_4P2 = Map.of(
      ASK_14, 0.25,
      BSSA_14, 0.25,
      CB_14, 0.25,
      CY_14, 0.25);

  /* 4.2 */
  static final class Wus2014_4p2 extends CombinedGmm {

    static final String NAME = CombinedGmm.NAME + "WUS 2014 (4.2, no Idriss)";
    static final Constraints CONSTRAINTS = CampbellBozorgnia_2014.CONSTRAINTS;
    static final CoefficientContainer COEFFS = CampbellBozorgnia_2014.COEFFS;

    Wus2014_4p2(Imt imt) {
      super(imt, WUS_2014_4P2);
    }
  }

  private static final Map<Gmm, Double> WUS_2018 = Map.of(
      ASK_14_BASIN, 0.25,
      BSSA_14_BASIN, 0.25,
      CB_14_BASIN, 0.25,
      CY_14_BASIN, 0.25);

  /* No Idriss. */
  static final class Wus2018 extends CombinedGmm {

    static final String NAME = CombinedGmm.NAME + "WUS 2018 (5.0, deep basins)";
    static final Constraints CONSTRAINTS = CampbellBozorgnia_2014.CONSTRAINTS;
    static final CoefficientContainer COEFFS = CampbellBozorgnia_2014.COEFFS;

    Wus2018(Imt imt) {
      super(imt, WUS_2018);
    }
  }

}
