package gov.usgs.earthquake.nshmp.gmm;

import static com.google.common.base.Preconditions.checkArgument;
import static com.google.common.base.Preconditions.checkNotNull;

import java.lang.reflect.Constructor;
import java.util.Collection;
import java.util.EnumSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.function.Predicate;

import com.google.common.cache.CacheBuilder;
import com.google.common.cache.CacheLoader;
import com.google.common.cache.LoadingCache;
import com.google.common.collect.Iterables;
import com.google.common.collect.Maps;
import com.google.common.collect.Sets;
import com.google.common.util.concurrent.UncheckedExecutionException;

import gov.usgs.earthquake.nshmp.gmm.CeusMb.AtkinsonBoore_2006_140bar_AB;
import gov.usgs.earthquake.nshmp.gmm.CeusMb.AtkinsonBoore_2006_140bar_J;
import gov.usgs.earthquake.nshmp.gmm.CeusMb.AtkinsonBoore_2006_200bar_AB;
import gov.usgs.earthquake.nshmp.gmm.CeusMb.AtkinsonBoore_2006_200bar_J;
import gov.usgs.earthquake.nshmp.gmm.CeusMb.Campbell_2003_AB;
import gov.usgs.earthquake.nshmp.gmm.CeusMb.Campbell_2003_J;
import gov.usgs.earthquake.nshmp.gmm.CeusMb.FrankelEtAl_1996_AB;
import gov.usgs.earthquake.nshmp.gmm.CeusMb.FrankelEtAl_1996_J;
import gov.usgs.earthquake.nshmp.gmm.CeusMb.SilvaEtAl_2002_AB;
import gov.usgs.earthquake.nshmp.gmm.CeusMb.SilvaEtAl_2002_J;
import gov.usgs.earthquake.nshmp.gmm.CeusMb.TavakoliPezeshk_2005_AB;
import gov.usgs.earthquake.nshmp.gmm.CeusMb.TavakoliPezeshk_2005_J;
import gov.usgs.earthquake.nshmp.gmm.GmmInput.Constraints;
import gov.usgs.earthquake.nshmp.gmm.NgaEast.NgaEast_2018;

/**
 * {@link GroundMotionModel} (Gmm) identifiers. Use these to generate
 * {@link Imt}-specific instances via {@link Gmm#instance(Imt)} and related
 * methods. Single or corporate authored models are identified as
 * NAME_YR_FLAVOR; multi-author models as INITIALS_YR_FLAVOR. FLAVOR is only
 * used for those models with region specific implementations or other variants.
 *
 * @author U.S. Geological Survey
 */
public enum Gmm {

  /*
   * Developer notes:
   *
   * Concrete GroundMotionModel implementations that are identified in enum
   * constructors below must implement a constructor taking a single IMT
   * argument.
   */

  /*
   * Active continent NGA-West1 WUS 2008.
   *
   * '_BASE' models do not include additional epistemic uncertainty.
   */

  /** @see BooreAtkinson_2008 */
  BA_08(
      BooreAtkinson_2008.Usgs.class,
      BooreAtkinson_2008.NAME,
      BooreAtkinson_2008.COEFFS,
      BooreAtkinson_2008.CONSTRAINTS),

  /** @see BooreAtkinson_2008 */
  BA_08_BASE(
      BooreAtkinson_2008.class,
      BooreAtkinson_2008.NAME,
      BooreAtkinson_2008.COEFFS,
      BooreAtkinson_2008.CONSTRAINTS),

  /** @see CampbellBozorgnia_2008 */
  CB_08(
      CampbellBozorgnia_2008.Usgs.class,
      CampbellBozorgnia_2008.NAME,
      CampbellBozorgnia_2008.COEFFS,
      CampbellBozorgnia_2008.CONSTRAINTS),

  /** @see CampbellBozorgnia_2008 */
  CB_08_BASE(
      CampbellBozorgnia_2008.class,
      CampbellBozorgnia_2008.NAME,
      CampbellBozorgnia_2008.COEFFS,
      CampbellBozorgnia_2008.CONSTRAINTS),

  /** @see ChiouYoungs_2008 */
  CY_08(
      ChiouYoungs_2008.Usgs.class,
      ChiouYoungs_2008.NAME,
      ChiouYoungs_2008.COEFFS,
      ChiouYoungs_2008.CONSTRAINTS),

  /** @see ChiouYoungs_2008 */
  CY_08_BASE(
      ChiouYoungs_2008.class,
      ChiouYoungs_2008.NAME,
      ChiouYoungs_2008.COEFFS,
      ChiouYoungs_2008.CONSTRAINTS),

  /*
   * Active continent NGA-West2 WUS 2014.
   *
   * '_BASE' models do not include additional epistemic uncertainty. '_BASIN'
   * models have basin amplification-only effects implemented as well as
   * additional epistemic uncertianty.
   */

  /** @see AbrahamsonEtAl_2014 */
  ASK_14(
      AbrahamsonEtAl_2014.Usgs.class,
      AbrahamsonEtAl_2014.NAME,
      AbrahamsonEtAl_2014.COEFFS,
      AbrahamsonEtAl_2014.CONSTRAINTS),

  /** @see AbrahamsonEtAl_2014 */
  ASK_14_BASE(
      AbrahamsonEtAl_2014.class,
      AbrahamsonEtAl_2014.NAME,
      AbrahamsonEtAl_2014.COEFFS,
      AbrahamsonEtAl_2014.CONSTRAINTS),

  /** @see AbrahamsonEtAl_2014 */
  ASK_14_BASIN(
      AbrahamsonEtAl_2014.Basin.class,
      AbrahamsonEtAl_2014.Basin.NAME,
      AbrahamsonEtAl_2014.Basin.COEFFS,
      AbrahamsonEtAl_2014.Basin.CONSTRAINTS),

  /** @see BooreEtAl_2014 */
  BSSA_14(
      BooreEtAl_2014.Usgs.class,
      BooreEtAl_2014.NAME,
      BooreEtAl_2014.COEFFS,
      BooreEtAl_2014.CONSTRAINTS),

  /** @see BooreEtAl_2014 */
  BSSA_14_BASE(
      BooreEtAl_2014.class,
      BooreEtAl_2014.NAME,
      BooreEtAl_2014.COEFFS,
      BooreEtAl_2014.CONSTRAINTS),

  /** @see BooreEtAl_2014 */
  BSSA_14_BASIN(
      BooreEtAl_2014.Basin.class,
      BooreEtAl_2014.Basin.NAME,
      BooreEtAl_2014.Basin.COEFFS,
      BooreEtAl_2014.Basin.CONSTRAINTS),

  /** @see CampbellBozorgnia_2014 */
  CB_14(
      CampbellBozorgnia_2014.Usgs.class,
      CampbellBozorgnia_2014.NAME,
      CampbellBozorgnia_2014.COEFFS,
      CampbellBozorgnia_2014.CONSTRAINTS),

  /** @see CampbellBozorgnia_2014 */
  CB_14_BASE(
      CampbellBozorgnia_2014.class,
      CampbellBozorgnia_2014.NAME,
      CampbellBozorgnia_2014.COEFFS,
      CampbellBozorgnia_2014.CONSTRAINTS),

  /** @see CampbellBozorgnia_2014 */
  CB_14_BASIN(
      CampbellBozorgnia_2014.Basin.class,
      CampbellBozorgnia_2014.Basin.NAME,
      CampbellBozorgnia_2014.Basin.COEFFS,
      CampbellBozorgnia_2014.Basin.CONSTRAINTS),

  /** @see ChiouYoungs_2014 */
  CY_14(
      ChiouYoungs_2014.Usgs.class,
      ChiouYoungs_2014.NAME,
      ChiouYoungs_2014.COEFFS,
      ChiouYoungs_2014.CONSTRAINTS),

  /** @see ChiouYoungs_2014 */
  CY_14_BASE(
      ChiouYoungs_2014.class,
      ChiouYoungs_2014.NAME,
      ChiouYoungs_2014.COEFFS,
      ChiouYoungs_2014.CONSTRAINTS),

  /** @see ChiouYoungs_2014 */
  CY_14_BASIN(
      ChiouYoungs_2014.Basin.class,
      ChiouYoungs_2014.Basin.NAME,
      ChiouYoungs_2014.Basin.COEFFS,
      ChiouYoungs_2014.Basin.CONSTRAINTS),

  /** @see Idriss_2014 */
  IDRISS_14(
      Idriss_2014.Usgs.class,
      Idriss_2014.NAME,
      Idriss_2014.COEFFS,
      Idriss_2014.CONSTRAINTS),

  /** @see Idriss_2014 */
  IDRISS_14_BASE(
      Idriss_2014.class,
      Idriss_2014.NAME,
      Idriss_2014.COEFFS,
      Idriss_2014.CONSTRAINTS),

  /* Active Continent AK 2007, HI 1998 */

  /** @see AbrahamsonSilva_1997 */
  AS_97(
      AbrahamsonSilva_1997.class,
      AbrahamsonSilva_1997.NAME,
      AbrahamsonSilva_1997.COEFFS,
      AbrahamsonSilva_1997.CONSTRAINTS),

  /** @see BooreEtAl_1997 */
  BJF_97(
      BooreEtAl_1997.class,
      BooreEtAl_1997.NAME,
      BooreEtAl_1997.COEFFS,
      BooreEtAl_1997.CONSTRAINTS),

  /** @see Campbell_1997 */
  CAMPBELL_97(
      Campbell_1997.class,
      Campbell_1997.NAME,
      Campbell_1997.COEFFS,
      Campbell_1997.CONSTRAINTS),

  /** @see CampbellBozorgnia_2003 */
  CB_03(
      CampbellBozorgnia_2003.class,
      CampbellBozorgnia_2003.NAME,
      CampbellBozorgnia_2003.COEFFS,
      CampbellBozorgnia_2003.CONSTRAINTS),

  /** @see MunsonThurber_1997 */
  MT_97(
      MunsonThurber_1997.class,
      MunsonThurber_1997.NAME,
      MunsonThurber_1997.COEFFS,
      MunsonThurber_1997.CONSTRAINTS),

  /** @see SadighEtAl_1997 */
  SADIGH_97(
      SadighEtAl_1997.class,
      SadighEtAl_1997.NAME,
      SadighEtAl_1997.COEFFS_BC_HI,
      SadighEtAl_1997.CONSTRAINTS),

  /* Subduction Interface and Slab WUS 2008 2014 2018, AK 2007 */

  /** @see AtkinsonBoore_2003 */
  AB_03_GLOBAL_INTERFACE(
      AtkinsonBoore_2003.GlobalInterface.class,
      AtkinsonBoore_2003.GlobalInterface.NAME,
      AtkinsonBoore_2003.COEFFS_GLOBAL_INTERFACE,
      AtkinsonBoore_2003.CONSTRAINTS),

  /** @see AtkinsonBoore_2003 */
  AB_03_GLOBAL_SLAB(
      AtkinsonBoore_2003.GlobalSlab.class,
      AtkinsonBoore_2003.GlobalSlab.NAME,
      AtkinsonBoore_2003.COEFFS_GLOBAL_SLAB,
      AtkinsonBoore_2003.CONSTRAINTS),

  /** @see AtkinsonBoore_2003 */
  AB_03_GLOBAL_SLAB_LOW_SAT(
      AtkinsonBoore_2003.GlobalSlabLowMagSaturation.class,
      AtkinsonBoore_2003.GlobalSlabLowMagSaturation.NAME,
      AtkinsonBoore_2003.COEFFS_GLOBAL_SLAB,
      AtkinsonBoore_2003.CONSTRAINTS),

  /** @see AtkinsonBoore_2003 */
  AB_03_CASCADIA_INTERFACE(
      AtkinsonBoore_2003.CascadiaInterface.class,
      AtkinsonBoore_2003.CascadiaInterface.NAME,
      AtkinsonBoore_2003.COEFFS_CASCADIA_INTERFACE,
      AtkinsonBoore_2003.CONSTRAINTS),

  /** @see AtkinsonBoore_2003 */
  AB_03_CASCADIA_SLAB(
      AtkinsonBoore_2003.CascadiaSlab.class,
      AtkinsonBoore_2003.CascadiaSlab.NAME,
      AtkinsonBoore_2003.COEFFS_CASCADIA_SLAB,
      AtkinsonBoore_2003.CONSTRAINTS),

  /** @see AtkinsonBoore_2003 */
  AB_03_CASCADIA_SLAB_LOW_SAT(
      AtkinsonBoore_2003.CascadiaSlabLowMagSaturation.class,
      AtkinsonBoore_2003.CascadiaSlabLowMagSaturation.NAME,
      AtkinsonBoore_2003.COEFFS_CASCADIA_SLAB,
      AtkinsonBoore_2003.CONSTRAINTS),

  /** @see AtkinsonMacias_2009 */
  AM_09_INTERFACE(
      AtkinsonMacias_2009.class,
      AtkinsonMacias_2009.NAME,
      AtkinsonMacias_2009.COEFFS,
      AtkinsonMacias_2009.CONSTRAINTS),

  /** @see AtkinsonMacias_2009 */
  AM_09_INTERFACE_BASIN(
      AtkinsonMacias_2009.Basin.class,
      AtkinsonMacias_2009.Basin.NAME,
      AtkinsonMacias_2009.COEFFS,
      AtkinsonMacias_2009.CONSTRAINTS),

  /** @see BcHydroNga_2018 */
  BCHYDRO_18_NGA_INTERFACE(
      BcHydroNga_2018.Interface.class,
      BcHydroNga_2018.Interface.NAME,
      BcHydroNga_2018.COEFFS,
      BcHydroNga_2018.CONSTRAINTS),

  /** @see BcHydroNga_2018 */
  BCHYDRO_18_NGA_SLAB(
      BcHydroNga_2018.Slab.class,
      BcHydroNga_2018.Slab.NAME,
      BcHydroNga_2018.COEFFS,
      BcHydroNga_2018.CONSTRAINTS),

  /** @see AbrahamsonGulerce_2020 */
  AG_20_GLOBAL_INTERFACE(
      AbrahamsonGulerce_2020.GlobalInterface.class,
      AbrahamsonGulerce_2020.GlobalInterface.NAME,
      AbrahamsonGulerce_2020.COEFFS,
      AbrahamsonGulerce_2020.CONSTRAINTS),

  /** @see AbrahamsonGulerce_2020 */
  AG_20_GLOBAL_SLAB(
      AbrahamsonGulerce_2020.GlobalSlab.class,
      AbrahamsonGulerce_2020.GlobalSlab.NAME,
      AbrahamsonGulerce_2020.COEFFS,
      AbrahamsonGulerce_2020.CONSTRAINTS),

  /** @see AbrahamsonGulerce_2020 */
  AG_20_CASCADIA_INTERFACE(
      AbrahamsonGulerce_2020.CascadiaInterface.class,
      AbrahamsonGulerce_2020.CascadiaInterface.NAME,
      AbrahamsonGulerce_2020.COEFFS,
      AbrahamsonGulerce_2020.CONSTRAINTS),

  /** @see AbrahamsonGulerce_2020 */
  AG_20_CASCADIA_SLAB(
      AbrahamsonGulerce_2020.CascadiaSlab.class,
      AbrahamsonGulerce_2020.CascadiaSlab.NAME,
      AbrahamsonGulerce_2020.COEFFS,
      AbrahamsonGulerce_2020.CONSTRAINTS),

  /** @see AbrahamsonGulerce_2020 */
  AG_20_CASCADIA_ADJ_INTERFACE(
      AbrahamsonGulerce_2020.CascadiaAdjustedInterface.class,
      AbrahamsonGulerce_2020.CascadiaAdjustedInterface.NAME,
      AbrahamsonGulerce_2020.COEFFS,
      AbrahamsonGulerce_2020.CONSTRAINTS),

  /** @see AbrahamsonGulerce_2020 */
  AG_20_CASCADIA_ADJ_SLAB(
      AbrahamsonGulerce_2020.CascadiaAdjustedSlab.class,
      AbrahamsonGulerce_2020.CascadiaAdjustedSlab.NAME,
      AbrahamsonGulerce_2020.COEFFS,
      AbrahamsonGulerce_2020.CONSTRAINTS),

  /** @see AbrahamsonGulerce_2020 */
  AG_20_CASCADIA_INTERFACE_BASIN(
      AbrahamsonGulerce_2020.CascadiaInterfaceBasin.class,
      AbrahamsonGulerce_2020.CascadiaInterfaceBasin.NAME,
      AbrahamsonGulerce_2020.COEFFS,
      AbrahamsonGulerce_2020.CONSTRAINTS),

  /** @see AbrahamsonGulerce_2020 */
  AG_20_CASCADIA_SLAB_BASIN(
      AbrahamsonGulerce_2020.CascadiaSlabBasin.class,
      AbrahamsonGulerce_2020.CascadiaSlabBasin.NAME,
      AbrahamsonGulerce_2020.COEFFS,
      AbrahamsonGulerce_2020.CONSTRAINTS),

  /** @see AbrahamsonGulerce_2020 */
  AG_20_CASCADIA_ADJ_INTERFACE_BASIN(
      AbrahamsonGulerce_2020.CascadiaAdjustedInterfaceBasin.class,
      AbrahamsonGulerce_2020.CascadiaAdjustedInterfaceBasin.NAME,
      AbrahamsonGulerce_2020.COEFFS,
      AbrahamsonGulerce_2020.CONSTRAINTS),

  /** @see AbrahamsonGulerce_2020 */
  AG_20_CASCADIA_ADJ_SLAB_BASIN(
      AbrahamsonGulerce_2020.CascadiaAdjustedSlabBasin.class,
      AbrahamsonGulerce_2020.CascadiaAdjustedSlabBasin.NAME,
      AbrahamsonGulerce_2020.COEFFS,
      AbrahamsonGulerce_2020.CONSTRAINTS),

  /** @see AbrahamsonGulerce_2020 */
  AG_20_ALASKA_INTERFACE(
      AbrahamsonGulerce_2020.AlaskaInterface.class,
      AbrahamsonGulerce_2020.AlaskaInterface.NAME,
      AbrahamsonGulerce_2020.COEFFS,
      AbrahamsonGulerce_2020.CONSTRAINTS),

  /** @see AbrahamsonGulerce_2020 */
  AG_20_ALASKA_SLAB(
      AbrahamsonGulerce_2020.AlaskaSlab.class,
      AbrahamsonGulerce_2020.AlaskaSlab.NAME,
      AbrahamsonGulerce_2020.COEFFS,
      AbrahamsonGulerce_2020.CONSTRAINTS),

  /** @see AbrahamsonGulerce_2020 */
  AG_20_ALASKA_ADJ_INTERFACE(
      AbrahamsonGulerce_2020.AlaskaAdjustedInterface.class,
      AbrahamsonGulerce_2020.AlaskaAdjustedInterface.NAME,
      AbrahamsonGulerce_2020.COEFFS,
      AbrahamsonGulerce_2020.CONSTRAINTS),

  /** @see AbrahamsonGulerce_2020 */
  AG_20_ALASKA_ADJ_SLAB(
      AbrahamsonGulerce_2020.AlaskaAdjustedSlab.class,
      AbrahamsonGulerce_2020.AlaskaAdjustedSlab.NAME,
      AbrahamsonGulerce_2020.COEFFS,
      AbrahamsonGulerce_2020.CONSTRAINTS),

  /** @see ParkerEtAl_2020 */
  PSBAH_20_GLOBAL_INTERFACE(
      ParkerEtAl_2020.GlobalInterface.class,
      ParkerEtAl_2020.GlobalInterface.NAME,
      ParkerEtAl_2020.COEFFS_INTERFACE,
      ParkerEtAl_2020.CONSTRAINTS_INTERFACE),

  /** @see ParkerEtAl_2020 */
  PSBAH_20_GLOBAL_SLAB(
      ParkerEtAl_2020.GlobalSlab.class,
      ParkerEtAl_2020.GlobalSlab.NAME,
      ParkerEtAl_2020.COEFFS_SLAB,
      ParkerEtAl_2020.CONSTRAINTS_SLAB),

  /** @see ParkerEtAl_2020 */
  PSBAH_20_CASCADIA_INTERFACE(
      ParkerEtAl_2020.CascadiaInterface.class,
      ParkerEtAl_2020.CascadiaInterface.NAME,
      ParkerEtAl_2020.COEFFS_INTERFACE,
      ParkerEtAl_2020.CONSTRAINTS_INTERFACE),

  /** @see ParkerEtAl_2020 */
  PSBAH_20_CASCADIA_SLAB(
      ParkerEtAl_2020.CascadiaSlab.class,
      ParkerEtAl_2020.CascadiaSlab.NAME,
      ParkerEtAl_2020.COEFFS_SLAB,
      ParkerEtAl_2020.CONSTRAINTS_SLAB),

  /** @see ParkerEtAl_2020 */
  PSBAH_20_CASCADIA_INTERFACE_BASIN(
      ParkerEtAl_2020.CascadiaInterfaceBasin.class,
      ParkerEtAl_2020.CascadiaInterfaceBasin.NAME,
      ParkerEtAl_2020.COEFFS_INTERFACE,
      ParkerEtAl_2020.CONSTRAINTS_INTERFACE),

  /** @see ParkerEtAl_2020 */
  PSBAH_20_CASCADIA_SLAB_BASIN(
      ParkerEtAl_2020.CascadiaSlabBasin.class,
      ParkerEtAl_2020.CascadiaSlabBasin.NAME,
      ParkerEtAl_2020.COEFFS_SLAB,
      ParkerEtAl_2020.CONSTRAINTS_SLAB),

  /** @see ParkerEtAl_2020 */
  PSBAH_20_ALASKA_INTERFACE(
      ParkerEtAl_2020.AlaskaInterface.class,
      ParkerEtAl_2020.AlaskaInterface.NAME,
      ParkerEtAl_2020.COEFFS_INTERFACE,
      ParkerEtAl_2020.CONSTRAINTS_INTERFACE),

  /** @see ParkerEtAl_2020 */
  PSBAH_20_ALASKA_SLAB(
      ParkerEtAl_2020.AlaskaSlab.class,
      ParkerEtAl_2020.AlaskaSlab.NAME,
      ParkerEtAl_2020.COEFFS_SLAB,
      ParkerEtAl_2020.CONSTRAINTS_SLAB),

  /** @see KuehnEtAl_2020 */
  KBCG_20_GLOBAL_INTERFACE(
      KuehnEtAl_2020.GlobalInterface.class,
      KuehnEtAl_2020.GlobalInterface.NAME,
      KuehnEtAl_2020.COEFFS,
      KuehnEtAl_2020.CONSTRAINTS_INTERFACE),

  /** @see KuehnEtAl_2020 */
  KBCG_20_GLOBAL_SLAB(
      KuehnEtAl_2020.GlobalSlab.class,
      KuehnEtAl_2020.GlobalSlab.NAME,
      KuehnEtAl_2020.COEFFS,
      KuehnEtAl_2020.CONSTRAINTS_SLAB),

  /** @see KuehnEtAl_2020 */
  KBCG_20_CASCADIA_INTERFACE(
      KuehnEtAl_2020.CascadiaInterface.class,
      KuehnEtAl_2020.CascadiaInterface.NAME,
      KuehnEtAl_2020.COEFFS,
      KuehnEtAl_2020.CONSTRAINTS_INTERFACE),

  /** @see KuehnEtAl_2020 */
  KBCG_20_CASCADIA_SLAB(
      KuehnEtAl_2020.CascadiaSlab.class,
      KuehnEtAl_2020.CascadiaSlab.NAME,
      KuehnEtAl_2020.COEFFS,
      KuehnEtAl_2020.CONSTRAINTS_SLAB),

  /** @see KuehnEtAl_2020 */
  KBCG_20_CASCADIA_INTERFACE_BASIN(
      KuehnEtAl_2020.CascadiaInterfaceBasin.class,
      KuehnEtAl_2020.CascadiaInterfaceBasin.NAME,
      KuehnEtAl_2020.COEFFS,
      KuehnEtAl_2020.CONSTRAINTS_INTERFACE),

  /** @see KuehnEtAl_2020 */
  KBCG_20_CASCADIA_SLAB_BASIN(
      KuehnEtAl_2020.CascadiaSlabBasin.class,
      KuehnEtAl_2020.CascadiaSlabBasin.NAME,
      KuehnEtAl_2020.COEFFS,
      KuehnEtAl_2020.CONSTRAINTS_SLAB),

  /** @see KuehnEtAl_2020 */
  KBCG_20_ALASKA_INTERFACE(
      KuehnEtAl_2020.AlaskaInterface.class,
      KuehnEtAl_2020.AlaskaInterface.NAME,
      KuehnEtAl_2020.COEFFS,
      KuehnEtAl_2020.CONSTRAINTS_INTERFACE),

  /** @see KuehnEtAl_2020 */
  KBCG_20_ALASKA_SLAB(
      KuehnEtAl_2020.AlaskaSlab.class,
      KuehnEtAl_2020.AlaskaSlab.NAME,
      KuehnEtAl_2020.COEFFS,
      KuehnEtAl_2020.CONSTRAINTS_SLAB),

  /** @see BcHydro_2012 */
  BCHYDRO_12_INTERFACE(
      BcHydro_2012.Interface.class,
      BcHydro_2012.Interface.NAME,
      BcHydro_2012.COEFFS,
      BcHydro_2012.CONSTRAINTS),

  /** @see BcHydro_2012 */
  BCHYDRO_12_SLAB(
      BcHydro_2012.Slab.class,
      BcHydro_2012.Slab.NAME,
      BcHydro_2012.COEFFS,
      BcHydro_2012.CONSTRAINTS),

  /** @see BcHydro_2012 */
  BCHYDRO_12_INTERFACE_BASIN(
      BcHydro_2012.BasinInterface.class,
      BcHydro_2012.BasinInterface.NAME,
      BcHydro_2012.COEFFS,
      BcHydro_2012.CONSTRAINTS),

  /** @see BcHydro_2012 */
  BCHYDRO_12_SLAB_BASIN(
      BcHydro_2012.BasinSlab.class,
      BcHydro_2012.BasinSlab.NAME,
      BcHydro_2012.COEFFS,
      BcHydro_2012.CONSTRAINTS),

  /** @see BcHydro_2012 */
  BCHYDRO_12_INTERFACE_BACKARC(
      BcHydro_2012.BackarcInterface.class,
      BcHydro_2012.BackarcInterface.NAME,
      BcHydro_2012.COEFFS,
      BcHydro_2012.CONSTRAINTS),

  /** @see BcHydro_2012 */
  BCHYDRO_12_SLAB_BACKARC(
      BcHydro_2012.BackarcSlab.class,
      BcHydro_2012.BackarcSlab.NAME,
      BcHydro_2012.COEFFS,
      BcHydro_2012.CONSTRAINTS),

  /** @see BcHydro_2012 */
  BCHYDRO_12_INTERFACE_BASIN_BACKARC(
      BcHydro_2012.BackarcBasinInterface.class,
      BcHydro_2012.BackarcBasinInterface.NAME,
      BcHydro_2012.COEFFS,
      BcHydro_2012.CONSTRAINTS),

  /** @see BcHydro_2012 */
  BCHYDRO_12_SLAB_BASIN_BACKARC(
      BcHydro_2012.BackarcBasinSlab.class,
      BcHydro_2012.BackarcBasinSlab.NAME,
      BcHydro_2012.COEFFS,
      BcHydro_2012.CONSTRAINTS),

  /** @see YoungsEtAl_1997 */
  YOUNGS_97_INTERFACE(
      YoungsEtAl_1997.Interface.class,
      YoungsEtAl_1997.Interface.NAME,
      YoungsEtAl_1997.COEFFS,
      YoungsEtAl_1997.CONSTRAINTS),

  /** @see YoungsEtAl_1997 */
  YOUNGS_97_SLAB(
      YoungsEtAl_1997.Slab.class,
      YoungsEtAl_1997.Slab.NAME,
      YoungsEtAl_1997.COEFFS,
      YoungsEtAl_1997.CONSTRAINTS),

  /** @see ZhaoEtAl_2006 */
  ZHAO_06_INTERFACE(
      ZhaoEtAl_2006.Interface.class,
      ZhaoEtAl_2006.Interface.NAME,
      ZhaoEtAl_2006.COEFFS,
      ZhaoEtAl_2006.CONSTRAINTS),

  /** @see ZhaoEtAl_2006 */
  ZHAO_06_SLAB(
      ZhaoEtAl_2006.Slab.class,
      ZhaoEtAl_2006.Slab.NAME,
      ZhaoEtAl_2006.COEFFS,
      ZhaoEtAl_2006.CONSTRAINTS),

  /** @see ZhaoEtAl_2006 */
  ZHAO_06_INTERFACE_BASIN(
      ZhaoEtAl_2006.BasinInterface.class,
      ZhaoEtAl_2006.BasinInterface.NAME,
      ZhaoEtAl_2006.COEFFS,
      ZhaoEtAl_2006.CONSTRAINTS),

  /** @see ZhaoEtAl_2006 */
  ZHAO_06_SLAB_BASIN(
      ZhaoEtAl_2006.BasinSlab.class,
      ZhaoEtAl_2006.BasinSlab.NAME,
      ZhaoEtAl_2006.COEFFS,
      ZhaoEtAl_2006.CONSTRAINTS),

  /*
   * Base implementations of the Gmm used in the 2008 CEUS model all work with
   * and assume magnitude = Mw. The method converter() is provided to allow
   * subclasses to impose a conversion if necessary.
   *
   * All CEUS models impose a clamp on median ground motions; see
   * GmmUtils.ceusMeanClip()
   */

  /* Stable continent CEUS 2008 2014 */

  /** @see AtkinsonBoore_2006p */
  AB_06_PRIME(
      AtkinsonBoore_2006p.class,
      AtkinsonBoore_2006p.NAME,
      AtkinsonBoore_2006p.COEFFS,
      AtkinsonBoore_2006p.CONSTRAINTS),

  /** @see AtkinsonBoore_2006 */
  AB_06_140BAR(
      AtkinsonBoore_2006.StressDrop_140bar.class,
      AtkinsonBoore_2006.StressDrop_140bar.NAME,
      AtkinsonBoore_2006.COEFFS_A,
      AtkinsonBoore_2006.CONSTRAINTS),

  /** @see AtkinsonBoore_2006 */
  AB_06_200BAR(
      AtkinsonBoore_2006.StressDrop_200bar.class,
      AtkinsonBoore_2006.StressDrop_200bar.NAME,
      AtkinsonBoore_2006.COEFFS_A,
      AtkinsonBoore_2006.CONSTRAINTS),

  /** @see Atkinson_2008p */
  ATKINSON_08_PRIME(
      Atkinson_2008p.class,
      Atkinson_2008p.NAME,
      Atkinson_2008p.COEFFS,
      Atkinson_2008p.CONSTRAINTS),

  /** @see Campbell_2003 */
  CAMPBELL_03(
      Campbell_2003.class,
      Campbell_2003.NAME,
      Campbell_2003.COEFFS,
      Campbell_2003.CONSTRAINTS),

  /** @see FrankelEtAl_1996 */
  FRANKEL_96(
      FrankelEtAl_1996.class,
      FrankelEtAl_1996.NAME,
      FrankelEtAl_1996.COEFFS,
      FrankelEtAl_1996.CONSTRAINTS),

  /** @see PezeshkEtAl_2011 */
  PEZESHK_11(
      PezeshkEtAl_2011.class,
      PezeshkEtAl_2011.NAME,
      PezeshkEtAl_2011.COEFFS,
      PezeshkEtAl_2011.CONSTRAINTS),

  /** @see SilvaEtAl_2002 */
  SILVA_02(
      SilvaEtAl_2002.class,
      SilvaEtAl_2002.NAME,
      SilvaEtAl_2002.COEFFS,
      SilvaEtAl_2002.CONSTRAINTS),

  /** @see SomervilleEtAl_2001 */
  SOMERVILLE_01(
      SomervilleEtAl_2001.class,
      SomervilleEtAl_2001.NAME,
      SomervilleEtAl_2001.COEFFS,
      SomervilleEtAl_2001.CONSTRAINTS),

  /** @see TavakoliPezeshk_2005 */
  TP_05(
      TavakoliPezeshk_2005.class,
      TavakoliPezeshk_2005.NAME,
      TavakoliPezeshk_2005.COEFFS,
      TavakoliPezeshk_2005.CONSTRAINTS),

  /** @see ToroEtAl_1997 */
  TORO_97_MW(
      ToroEtAl_1997.Mw.class,
      ToroEtAl_1997.Mw.NAME,
      ToroEtAl_1997.COEFFS_MW,
      ToroEtAl_1997.CONSTRAINTS),

  /* Johnston mag converting flavors of CEUS 2008 */

  /** @see AtkinsonBoore_2006 */
  AB_06_140BAR_J(
      AtkinsonBoore_2006_140bar_J.class,
      AtkinsonBoore_2006_140bar_J.NAME,
      AtkinsonBoore_2006.COEFFS_A,
      AtkinsonBoore_2006.CONSTRAINTS),

  /** @see AtkinsonBoore_2006 */
  AB_06_200BAR_J(
      AtkinsonBoore_2006_200bar_J.class,
      AtkinsonBoore_2006_200bar_J.NAME,
      AtkinsonBoore_2006.COEFFS_A,
      AtkinsonBoore_2006.CONSTRAINTS),

  /** @see Campbell_2003 */
  CAMPBELL_03_J(
      Campbell_2003_J.class,
      Campbell_2003_J.NAME,
      Campbell_2003.COEFFS,
      Campbell_2003.CONSTRAINTS),

  /** @see FrankelEtAl_1996 */
  FRANKEL_96_J(
      FrankelEtAl_1996_J.class,
      FrankelEtAl_1996_J.NAME,
      FrankelEtAl_1996.COEFFS,
      FrankelEtAl_1996.CONSTRAINTS),

  /** @see SilvaEtAl_2002 */
  SILVA_02_J(
      SilvaEtAl_2002_J.class,
      SilvaEtAl_2002_J.NAME,
      SilvaEtAl_2002.COEFFS,
      SilvaEtAl_2002.CONSTRAINTS),

  /** @see TavakoliPezeshk_2005 */
  TP_05_J(
      TavakoliPezeshk_2005_J.class,
      TavakoliPezeshk_2005_J.NAME,
      TavakoliPezeshk_2005.COEFFS,
      TavakoliPezeshk_2005.CONSTRAINTS),

  /* Atkinson Boore mag converting flavors of CEUS 2008 */

  /** @see AtkinsonBoore_2006 */
  AB_06_140BAR_AB(
      AtkinsonBoore_2006_140bar_AB.class,
      AtkinsonBoore_2006_140bar_AB.NAME,
      AtkinsonBoore_2006.COEFFS_A,
      AtkinsonBoore_2006.CONSTRAINTS),

  /** @see AtkinsonBoore_2006 */
  AB_06_200BAR_AB(
      AtkinsonBoore_2006_200bar_AB.class,
      AtkinsonBoore_2006_200bar_AB.NAME,
      AtkinsonBoore_2006.COEFFS_A,
      AtkinsonBoore_2006.CONSTRAINTS),

  /** @see Campbell_2003 */
  CAMPBELL_03_AB(
      Campbell_2003_AB.class,
      Campbell_2003_AB.NAME,
      Campbell_2003.COEFFS,
      Campbell_2003.CONSTRAINTS),

  /** @see FrankelEtAl_1996 */
  FRANKEL_96_AB(
      FrankelEtAl_1996_AB.class,
      FrankelEtAl_1996_AB.NAME,
      FrankelEtAl_1996.COEFFS,
      FrankelEtAl_1996.CONSTRAINTS),

  /** @see SilvaEtAl_2002 */
  SILVA_02_AB(
      SilvaEtAl_2002_AB.class,
      SilvaEtAl_2002_AB.NAME,
      SilvaEtAl_2002.COEFFS,
      SilvaEtAl_2002.CONSTRAINTS),

  /** @see TavakoliPezeshk_2005 */
  TP_05_AB(
      TavakoliPezeshk_2005_AB.class,
      TavakoliPezeshk_2005_AB.NAME,
      TavakoliPezeshk_2005.COEFFS,
      TavakoliPezeshk_2005.CONSTRAINTS),

  /** @see ToroEtAl_1997 */
  TORO_97_MB(
      ToroEtAl_1997.Mb.class,
      ToroEtAl_1997.Mb.NAME,
      ToroEtAl_1997.COEFFS_MW,
      ToroEtAl_1997.CONSTRAINTS),

  /* Other */

  /** @see Atkinson_2010 */
  ATKINSON_10(
      Atkinson_2010.UsgsNga.class,
      Atkinson_2010.NAME,
      Atkinson_2010.COEFFS,
      Atkinson_2010.CONSTRAINTS),

  /** @see Atkinson_2010 */
  ATKINSON_10_BASE(
      Atkinson_2010.class,
      Atkinson_2010.NAME,
      Atkinson_2010.COEFFS,
      Atkinson_2010.CONSTRAINTS),

  /** @see Atkinson_2010 */
  ATKINSON_10_CALDERA(
      Atkinson_2010.CalderaSmoothed.class,
      Atkinson_2010.CalderaSmoothed.NAME,
      Atkinson_2010.COEFFS,
      Atkinson_2010.CONSTRAINTS),

  /** @see Atkinson_2015 */
  ATKINSON_15(
      Atkinson_2015.class,
      Atkinson_2015.NAME,
      Atkinson_2015.COEFFS,
      Atkinson_2015.CONSTRAINTS),

  /** @see WongEtAl_2015 */
  WONG_15(
      WongEtAl_2015.class,
      WongEtAl_2015.NAME,
      WongEtAl_2015.COEFFS_760,
      WongEtAl_2015.CONSTRAINTS),

  /* NGA-East for USGS */

  /** @see NgaEast */
  NGA_EAST_2018(
      NgaEast.NgaEast_2018.class,
      NgaEast.NgaEast_2018.NAME,
      NgaEast.COEFFS_SIGMA_PANEL,
      NgaEast.CONSTRAINTS),

  /** @see NgaEast */
  NGA_EAST_2023(
      NgaEast.NgaEast_2023.class,
      NgaEast.NgaEast_2023.NAME,
      NgaEast.COEFFS_SIGMA_PANEL,
      NgaEast.CONSTRAINTS),

  /* NGA-East USGS Seed Tree */

  /** @see NgaEast */
  NGA_EAST_SEEDS_2018(
      NgaEast.UsgsSeeds_2018.class,
      NgaEast.UsgsSeeds_2018.NAME,
      NgaEast.COEFFS_SIGMA_PANEL,
      NgaEast.CONSTRAINTS),

  /** @see NgaEast */
  NGA_EAST_SEEDS_2023(
      NgaEast.UsgsSeeds_2023.class,
      NgaEast.UsgsSeeds_2023.NAME,
      NgaEast.COEFFS_SIGMA_PANEL,
      NgaEast.CONSTRAINTS),

  /* NGA-East Seed Models */

  /** @see NgaEast_2018 */
  NGA_EAST_SEED_1CCSP(
      NgaEast.Seed_1CCSP.class,
      NgaEast.Seed_1CCSP.NAME,
      NgaEast.COEFFS_SIGMA_PANEL,
      NgaEast.CONSTRAINTS),

  /** @see NgaEast_2018 */
  NGA_EAST_SEED_1CVSP(
      NgaEast.Seed_1CVSP.class,
      NgaEast.Seed_1CVSP.NAME,
      NgaEast.COEFFS_SIGMA_PANEL,
      NgaEast.CONSTRAINTS),

  /** @see NgaEast_2018 */
  NGA_EAST_SEED_2CCSP(
      NgaEast.Seed_2CCSP.class,
      NgaEast.Seed_2CCSP.NAME,
      NgaEast.COEFFS_SIGMA_PANEL,
      NgaEast.CONSTRAINTS),

  /** @see NgaEast_2018 */
  NGA_EAST_SEED_2CVSP(
      NgaEast.Seed_2CVSP.class,
      NgaEast.Seed_2CVSP.NAME,
      NgaEast.COEFFS_SIGMA_PANEL,
      NgaEast.CONSTRAINTS),

  /** @see NgaEast_2018 */
  NGA_EAST_SEED_B_A04(
      NgaEast.Seed_B_a04.class,
      NgaEast.Seed_B_a04.NAME,
      NgaEast.COEFFS_SIGMA_PANEL,
      NgaEast.CONSTRAINTS),

  /** @see NgaEast_2018 */
  NGA_EAST_SEED_B_AB14(
      NgaEast.Seed_B_ab14.class,
      NgaEast.Seed_B_ab14.NAME,
      NgaEast.COEFFS_SIGMA_PANEL,
      NgaEast.CONSTRAINTS),

  /** @see NgaEast_2018 */
  NGA_EAST_SEED_B_AB95(
      NgaEast.Seed_B_ab95.class,
      NgaEast.Seed_B_ab95.NAME,
      NgaEast.COEFFS_SIGMA_PANEL,
      NgaEast.CONSTRAINTS),

  /** @see NgaEast_2018 */
  NGA_EAST_SEED_B_BCA10D(
      NgaEast.Seed_B_bca10d.class,
      NgaEast.Seed_B_bca10d.NAME,
      NgaEast.COEFFS_SIGMA_PANEL,
      NgaEast.CONSTRAINTS),

  /** @see NgaEast_2018 */
  NGA_EAST_SEED_B_BS11(
      NgaEast.Seed_B_bs11.class,
      NgaEast.Seed_B_bs11.NAME,
      NgaEast.COEFFS_SIGMA_PANEL,
      NgaEast.CONSTRAINTS),

  /** @see NgaEast_2018 */
  NGA_EAST_SEED_B_SGD02(
      NgaEast.Seed_B_sgd02.class,
      NgaEast.Seed_B_sgd02.NAME,
      NgaEast.COEFFS_SIGMA_PANEL,
      NgaEast.CONSTRAINTS),

  /** @see NgaEast_2018 */
  NGA_EAST_SEED_B20_AB14MOD1(
      NgaEast.Seed_B20_ab14mod1.class,
      NgaEast.Seed_B20_ab14mod1.NAME,
      NgaEast.COEFFS_SIGMA_PANEL,
      NgaEast.CONSTRAINTS),

  /** @see NgaEast_2018 */
  NGA_EAST_SEED_B20_AB14MOD2(
      NgaEast.Seed_B20_ab14mod2.class,
      NgaEast.Seed_B20_ab14mod2.NAME,
      NgaEast.COEFFS_SIGMA_PANEL,
      NgaEast.CONSTRAINTS),

  /** @see NgaEast_2018 */
  NGA_EAST_SEED_B20_BCA10D(
      NgaEast.Seed_B20_bca10d.class,
      NgaEast.Seed_B20_bca10d.NAME,
      NgaEast.COEFFS_SIGMA_PANEL,
      NgaEast.CONSTRAINTS),

  /** @see NgaEast_2018 */
  NGA_EAST_SEED_FRANKEL(
      NgaEast.Seed_Frankel.class,
      NgaEast.Seed_Frankel.NAME,
      NgaEast.COEFFS_SIGMA_PANEL,
      NgaEast.CONSTRAINTS),

  /** @see NgaEast_2018 */
  NGA_EAST_SEED_GRAIZER(
      NgaEast.Seed_Graizer.class,
      NgaEast.Seed_Graizer.NAME,
      NgaEast.COEFFS_SIGMA_PANEL,
      NgaEast.CONSTRAINTS),

  /** @see NgaEast_2018 */
  NGA_EAST_SEED_GRAIZER16(
      NgaEast.Seed_Graizer16.class,
      NgaEast.Seed_Graizer16.NAME,
      NgaEast.COEFFS_SIGMA_PANEL,
      NgaEast.CONSTRAINTS),

  /** @see NgaEast_2018 */
  NGA_EAST_SEED_GRAIZER17(
      NgaEast.Seed_Graizer17.class,
      NgaEast.Seed_Graizer17.NAME,
      NgaEast.COEFFS_SIGMA_PANEL,
      NgaEast.CONSTRAINTS),

  /** @see NgaEast_2018 */
  NGA_EAST_SEED_HA15(
      NgaEast.Seed_HA15.class,
      NgaEast.Seed_HA15.NAME,
      NgaEast.COEFFS_SIGMA_PANEL,
      NgaEast.CONSTRAINTS),

  /** @see NgaEast_2018 */
  NGA_EAST_SEED_PEER_EX(
      NgaEast.Seed_PEER_EX.class,
      NgaEast.Seed_PEER_EX.NAME,
      NgaEast.COEFFS_SIGMA_PANEL,
      NgaEast.CONSTRAINTS),

  /** @see NgaEast_2018 */
  NGA_EAST_SEED_PEER_GP(
      NgaEast.Seed_PEER_GP.class,
      NgaEast.Seed_PEER_GP.NAME,
      NgaEast.COEFFS_SIGMA_PANEL,
      NgaEast.CONSTRAINTS),

  /** @see NgaEast_2018 */
  NGA_EAST_SEED_PZCT15_M1SS(
      NgaEast.Seed_PZCT15_M1SS.class,
      NgaEast.Seed_PZCT15_M1SS.NAME,
      NgaEast.COEFFS_SIGMA_PANEL,
      NgaEast.CONSTRAINTS),

  /** @see NgaEast_2018 */
  NGA_EAST_SEED_PZCT15_M2ES(
      NgaEast.Seed_PZCT15_M2ES.class,
      NgaEast.Seed_PZCT15_M2ES.NAME,
      NgaEast.COEFFS_SIGMA_PANEL,
      NgaEast.CONSTRAINTS),

  /** @see NgaEast_2018 @see PezeshkEtAl_2018 */
  NGA_EAST_SEED_PZCT18_M1SS(
      PezeshkEtAl_2018.Seed_PZCT18_M1SS.class,
      PezeshkEtAl_2018.Seed_PZCT18_M1SS.NAME,
      PezeshkEtAl_2018.COEFFS_M1SS,
      PezeshkEtAl_2018.CONSTRAINTS),

  /** @see PezeshkEtAl_2018 @see PezeshkEtAl_2018 */
  NGA_EAST_SEED_PZCT18_M2ES(
      PezeshkEtAl_2018.Seed_PZCT18_M2ES.class,
      PezeshkEtAl_2018.Seed_PZCT18_M2ES.NAME,
      PezeshkEtAl_2018.COEFFS_M2ES,
      PezeshkEtAl_2018.CONSTRAINTS),

  /** @see NgaEast_2018 */
  NGA_EAST_SEED_SP15(
      NgaEast.Seed_SP15.class,
      NgaEast.Seed_SP15.NAME,
      NgaEast.COEFFS_SIGMA_PANEL,
      NgaEast.CONSTRAINTS),

  /** @see NgaEast_2018 @see ShahjoueiPezeshk_2016 */
  NGA_EAST_SEED_SP16(
      ShahjoueiPezeshk_2016.class,
      ShahjoueiPezeshk_2016.NAME,
      ShahjoueiPezeshk_2016.COEFFS,
      ShahjoueiPezeshk_2016.CONSTRAINTS),

  /** @see NgaEast_2018 */
  NGA_EAST_SEED_YA15(
      NgaEast.Seed_YA15.class,
      NgaEast.Seed_YA15.NAME,
      NgaEast.COEFFS_SIGMA_PANEL,
      NgaEast.CONSTRAINTS),

  /* Combined: must be declared after any dependent models above. */

  /**
   * 2021 Hawaii weight-averaged GMM for deep earthquakes.
   */
  COMBINED_HI_DEEP_2021(
      CombinedGmm.HawaiiDeep2021.class,
      CombinedGmm.HawaiiDeep2021.NAME,
      CombinedGmm.HawaiiDeep2021.COEFFS,
      CombinedGmm.HawaiiDeep2021.CONSTRAINTS),

  /**
   * 2021 Hawaii weight-averaged GMM for shallow earthquakes.
   */
  COMBINED_HI_SHALLOW_2021(
      CombinedGmm.HawaiiShallow2021.class,
      CombinedGmm.HawaiiShallow2021.NAME,
      CombinedGmm.HawaiiShallow2021.COEFFS,
      CombinedGmm.HawaiiShallow2021.CONSTRAINTS),

  /**
   * 2014 CEUS weight-averaged GMM. This is the fault-variant that includes
   * Somerville (2011).
   */
  COMBINED_CEUS_2014(
      CombinedGmm.Ceus2014.class,
      CombinedGmm.Ceus2014.NAME,
      CombinedGmm.Ceus2014.COEFFS,
      CombinedGmm.Ceus2014.CONSTRAINTS),

  /**
   * 2018 CEUS weight-averaged GMM.
   */
  COMBINED_CEUS_2018(
      CombinedGmm.Ceus2018.class,
      CombinedGmm.Ceus2018.NAME,
      CombinedGmm.Ceus2018.COEFFS,
      CombinedGmm.Ceus2018.CONSTRAINTS),

  /**
   * 2018 CEUS weight-averaged GMM (Gulf Coast).
   */
  COMBINED_CEUS_2018_CPA(
      CombinedGmm.Ceus2018Cpa.class,
      CombinedGmm.Ceus2018Cpa.NAME,
      CombinedGmm.Ceus2018Cpa.COEFFS,
      CombinedGmm.Ceus2018Cpa.CONSTRAINTS),

  /**
   * 2018 WUS weight-averaged GMM. These are the basin-amplifying flavors of
   * NGA-West2 and does not include Idriss.
   */
  COMBINED_WUS_2014_41(
      CombinedGmm.Wus2014_4p1.class,
      CombinedGmm.Wus2014_4p1.NAME,
      CombinedGmm.Wus2014_4p1.COEFFS,
      CombinedGmm.Wus2014_4p1.CONSTRAINTS),

  /**
   * 2018 WUS weight-averaged GMM. These are the basin-amplifying flavors of
   * NGA-West2 and does not include Idriss.
   */
  COMBINED_WUS_2014_42(
      CombinedGmm.Wus2014_4p2.class,
      CombinedGmm.Wus2014_4p2.NAME,
      CombinedGmm.Wus2014_4p2.COEFFS,
      CombinedGmm.Wus2014_4p2.CONSTRAINTS),

  /**
   * 2018 WUS weight-averaged GMM. These are the basin-amplifying flavors of
   * NGA-West2 and does not include Idriss.
   */
  COMBINED_WUS_2018(
      CombinedGmm.Wus2018.class,
      CombinedGmm.Wus2018.NAME,
      CombinedGmm.Wus2018.COEFFS,
      CombinedGmm.Wus2018.CONSTRAINTS);

  private final Class<? extends GroundMotionModel> delegate;
  private final String name;
  private final Set<Imt> imts;
  private final Constraints constraints;
  private final LoadingCache<Imt, GroundMotionModel> cache;

  private Gmm(
      Class<? extends GroundMotionModel> delegate,
      String name,
      CoefficientContainer coeffs,
      Constraints constraints) {

    this.delegate = delegate;
    this.name = name;
    this.constraints = constraints;
    imts = coeffs.imts();
    cache = CacheBuilder.newBuilder().build(new CacheLoader<Imt, GroundMotionModel>() {
      @Override
      public GroundMotionModel load(Imt imt) {
        return createInstance(imt);
      }
    });
  }

  private GroundMotionModel createInstance(Imt imt) {
    checkArgument(this.imts.contains(checkNotNull(imt)),
        "Gmm: %s does not support Imt: %s", this.name(), imt.name());
    try {
      Constructor<? extends GroundMotionModel> con = delegate.getDeclaredConstructor(Imt.class);
      GroundMotionModel gmm = con.newInstance(imt);
      return gmm;
    } catch (Exception e) {
      throw new RuntimeException(
          "Problem loading GMM cache; gmm: " + this.name() + " imt: " + imt, e);
    }
  }

  /**
   * Retreive an instance of a {@code GroundMotionModel}, either by creating a
   * new one, or fetching from a cache.
   *
   * @param imt of the retreived instance
   * @throws UncheckedExecutionException if there is an instantiation problem
   */
  public GroundMotionModel instance(Imt imt) {
    return cache.getUnchecked(imt);
  }

  /**
   * Retrieve an immutable map of {@code GroundMotionModel} instances, either by
   * creating new ones, or fetching them from a cache.
   *
   * @param imt of each retreived instance
   * @param gmms to retrieve
   * @throws UncheckedExecutionException if there is an instantiation problem
   */
  public static Map<Gmm, GroundMotionModel> instances(Imt imt, Set<Gmm> gmms) {
    Map<Gmm, GroundMotionModel> instanceMap = Maps.newEnumMap(Gmm.class);
    for (Gmm gmm : gmms) {
      instanceMap.put(gmm, gmm.instance(imt));
    }
    return Maps.immutableEnumMap(instanceMap);
  }

  /**
   * Retrieve immutable maps of {@code GroundMotionModel} instances for a range
   * of {@code Imt}s, either by creating new ones, or fetching them from a
   * cache.
   *
   * @param imts to retrieve instances for
   * @param gmms to retrieve
   * @throws UncheckedExecutionException if there is an instantiation problem
   */
  public static Map<Imt, Map<Gmm, GroundMotionModel>> instances(Set<Imt> imts, Set<Gmm> gmms) {
    Map<Imt, Map<Gmm, GroundMotionModel>> imtMap = Maps.newEnumMap(Imt.class);
    for (Imt imt : imts) {
      imtMap.put(imt, instances(imt, gmms));
    }
    return Maps.immutableEnumMap(imtMap);
  }

  @Override
  public String toString() {
    return name;
  }

  /**
   * Return the {@code Set} of the intensity measure types ({@code Imt}s)
   * supported by this {@code Gmm}.
   */
  public Set<Imt> supportedImts() {
    return imts;
  }

  /**
   * Return the {@code Set} of the intensity measure types ({@code Imt}s)
   * supported by all of the supplied {@code Gmm}s.
   *
   * @param gmms models for which to return common {@code Imt} support
   */
  public static Set<Imt> supportedImts(Collection<Gmm> gmms) {
    Set<Imt> imts = EnumSet.allOf(Imt.class);
    for (Gmm gmm : gmms) {
      imts = Sets.intersection(imts, gmm.supportedImts());
    }
    return EnumSet.copyOf(imts);
  }

  /**
   * Return the set of spectral acceleration {@code Imt}s that are supported by
   * this {@code Gmm}.
   */
  public Set<Imt> responseSpectrumImts() {
    return Sets.intersection(imts, Imt.saImts());
  }

  /**
   * Return the set of spectral acceleration (SA) {@code Imt}s that are common
   * to the supplied {@code Gmm}s.
   *
   * @param gmms models for which to return common SA {@code Imt} support
   */
  public static Set<Imt> responseSpectrumImts(Collection<Gmm> gmms) {
    return Sets.intersection(supportedImts(gmms), Imt.saImts());
  }

  /**
   * Return the {@code Set} of {@code Gmm}s that support the supplied
   * {@code Imt}.
   *
   * @param imt for which to return the {@code Gmm}s that support it
   */
  public static Set<Gmm> supportedGmms(Imt imt) {
    return Sets.newEnumSet(Iterables.filter(
        EnumSet.allOf(Gmm.class),
        new Predicate<Gmm>() {
          @Override
          public boolean test(Gmm gmm) {
            return gmm.imts.contains(imt);
          }
        }::test),
        Gmm.class);
  }

  /**
   * Return the input {@code Constraints} for this {@code Gmm}.
   */
  public Constraints constraints() {
    return constraints;
  }

  @SuppressWarnings("javadoc")
  public enum Group {

    CEUS_14_STABLE_CRUST(
        "2014 Stable Crust (CEUS)",
        List.of(
            ATKINSON_08_PRIME,
            AB_06_PRIME,
            CAMPBELL_03,
            FRANKEL_96,
            PEZESHK_11,
            SILVA_02,
            SOMERVILLE_01,
            TP_05,
            TORO_97_MW)),

    WUS_14_ACTIVE_CRUST(
        "2014 Active Crust (WUS)",
        List.of(
            ASK_14,
            BSSA_14,
            CB_14,
            CY_14,
            IDRISS_14)),

    WUS_18_ACTIVE_CRUST(
        "2018 Active Crust (WUS)",
        List.of(
            ASK_14_BASIN,
            BSSA_14_BASIN,
            CB_14_BASIN,
            CY_14_BASIN)),

    WUS_14_INTERFACE(
        "2014 Subduction Interface (WUS)",
        List.of(
            AB_03_GLOBAL_INTERFACE,
            AM_09_INTERFACE,
            BCHYDRO_12_INTERFACE,
            ZHAO_06_INTERFACE)),

    WUS_18_INTERFACE(
        "2018 Subduction Interface (WUS) beta",
        List.of(
            AM_09_INTERFACE_BASIN,
            BCHYDRO_12_INTERFACE_BASIN,
            ZHAO_06_INTERFACE_BASIN)),

    WUS_14_SLAB(
        "2014 Subduction Intraslab (WUS)",
        List.of(
            AB_03_CASCADIA_SLAB_LOW_SAT,
            AB_03_GLOBAL_SLAB_LOW_SAT,
            BCHYDRO_12_SLAB,
            ZHAO_06_SLAB)),

    WUS_18_SLAB(
        "2018 Subduction Intraslab (WUS) beta",
        List.of(
            BCHYDRO_12_SLAB_BASIN,
            ZHAO_06_SLAB_BASIN)),

    NGA_SUBDUCTION(
        "NGA-Subduction",
        List.of(
            AG_20_GLOBAL_INTERFACE,
            AG_20_GLOBAL_SLAB,
            AG_20_CASCADIA_INTERFACE,
            AG_20_CASCADIA_SLAB,
            AG_20_CASCADIA_ADJ_INTERFACE,
            AG_20_CASCADIA_ADJ_SLAB,
            AG_20_CASCADIA_INTERFACE_BASIN,
            AG_20_CASCADIA_SLAB_BASIN,
            AG_20_CASCADIA_ADJ_INTERFACE_BASIN,
            AG_20_CASCADIA_ADJ_SLAB_BASIN,
            AG_20_ALASKA_INTERFACE,
            AG_20_ALASKA_SLAB,
            AG_20_ALASKA_ADJ_INTERFACE,
            AG_20_ALASKA_ADJ_SLAB,
            KBCG_20_GLOBAL_INTERFACE,
            KBCG_20_GLOBAL_SLAB,
            KBCG_20_CASCADIA_INTERFACE,
            KBCG_20_CASCADIA_SLAB,
            KBCG_20_CASCADIA_INTERFACE_BASIN,
            KBCG_20_CASCADIA_SLAB_BASIN,
            KBCG_20_ALASKA_INTERFACE,
            KBCG_20_ALASKA_SLAB,
            PSBAH_20_GLOBAL_INTERFACE,
            PSBAH_20_GLOBAL_SLAB,
            PSBAH_20_CASCADIA_INTERFACE,
            PSBAH_20_CASCADIA_SLAB,
            PSBAH_20_CASCADIA_INTERFACE_BASIN,
            PSBAH_20_CASCADIA_SLAB_BASIN,
            PSBAH_20_ALASKA_INTERFACE,
            PSBAH_20_ALASKA_SLAB)),

    CEUS_08_STABLE_CRUST(
        "2008 Stable Crust (CEUS)",
        List.of(
            AB_06_140BAR,
            AB_06_200BAR,
            CAMPBELL_03,
            FRANKEL_96,
            SILVA_02,
            SOMERVILLE_01,
            TP_05,
            TORO_97_MW)),

    WUS_08_ACTIVE_CRUST(
        "2008 Active Crust (WUS)",
        List.of(
            BA_08,
            CB_08,
            CY_08)),

    WUS_08_INTERFACE(
        "2008 Subduction Interface (WUS)",
        List.of(
            AB_03_GLOBAL_INTERFACE,
            YOUNGS_97_INTERFACE,
            ZHAO_06_INTERFACE)),

    WUS_08_SLAB(
        "2008 Subduction Intraslab (WUS)",
        List.of(
            AB_03_CASCADIA_SLAB,
            AB_03_GLOBAL_SLAB,
            YOUNGS_97_SLAB)),

    AK_07_ACTIVE_CRUST(
        "2007 Active Crust (AK)",
        List.of(
            AS_97,
            BJF_97,
            CB_03,
            SADIGH_97)),

    AK_07_INTERFACE(
        "2007 Subduction Interface (AK)",
        List.of(
            YOUNGS_97_INTERFACE,
            SADIGH_97)),

    AK_07_SLAB(
        "2007 Subduction Intraslab (AK)",
        List.of(
            YOUNGS_97_SLAB,
            AB_03_GLOBAL_SLAB)),

    HI_98(
        "1998 Active Volcanic (HI)",
        List.of(
            BJF_97,
            CAMPBELL_97,
            MT_97,
            SADIGH_97,
            YOUNGS_97_SLAB)),

    HI_21(
        "2021 Active Volcanic (HI)",
        List.of(
            WONG_15,
            ATKINSON_10,
            ATKINSON_10_CALDERA,
            ASK_14,
            BSSA_14,
            CB_14,
            CY_14,
            BCHYDRO_12_SLAB)),

    NGA_BASE(
        "NGA Models (no uncertainty)",
        List.of(
            BA_08_BASE,
            CB_08_BASE,
            CY_08_BASE,
            ASK_14_BASE,
            BSSA_14_BASE,
            CB_14_BASE,
            CY_14_BASE,
            ATKINSON_10_BASE)),

    OTHER(
        "Others",
        List.of(
            ATKINSON_15,
            AB_03_CASCADIA_INTERFACE,
            BCHYDRO_18_NGA_INTERFACE,
            BCHYDRO_18_NGA_SLAB,
            BCHYDRO_12_INTERFACE_BACKARC,
            BCHYDRO_12_INTERFACE_BASIN_BACKARC,
            BCHYDRO_12_SLAB_BACKARC,
            BCHYDRO_12_SLAB_BASIN_BACKARC)),

    COMBINED(
        "Combined Models",
        List.of(
            COMBINED_HI_DEEP_2021,
            COMBINED_HI_SHALLOW_2021,
            COMBINED_CEUS_2014,
            COMBINED_CEUS_2018,
            COMBINED_CEUS_2018_CPA,
            COMBINED_WUS_2014_41,
            COMBINED_WUS_2014_42,
            COMBINED_WUS_2018)),

    NGA_EAST(
        "NGA-East Logic Trees",
        List.of(
            NGA_EAST_2018,
            NGA_EAST_SEEDS_2018,
            NGA_EAST_2023,
            NGA_EAST_SEEDS_2023)),

    NGA_EAST_SEEDS(
        "NGA-East Seed Models",
        List.of(
            NGA_EAST_SEED_1CCSP,
            NGA_EAST_SEED_1CVSP,
            NGA_EAST_SEED_2CCSP,
            NGA_EAST_SEED_2CVSP,
            NGA_EAST_SEED_B_A04,
            NGA_EAST_SEED_B_AB14,
            NGA_EAST_SEED_B_AB95,
            NGA_EAST_SEED_B_BCA10D,
            NGA_EAST_SEED_B_BS11,
            NGA_EAST_SEED_B_SGD02,
            NGA_EAST_SEED_B20_AB14MOD1,
            NGA_EAST_SEED_B20_AB14MOD2,
            NGA_EAST_SEED_B20_BCA10D,
            NGA_EAST_SEED_FRANKEL,
            NGA_EAST_SEED_GRAIZER,
            NGA_EAST_SEED_GRAIZER16,
            NGA_EAST_SEED_GRAIZER17,
            NGA_EAST_SEED_HA15,
            NGA_EAST_SEED_PEER_EX,
            NGA_EAST_SEED_PEER_GP,
            NGA_EAST_SEED_PZCT15_M1SS,
            NGA_EAST_SEED_PZCT15_M2ES,
            NGA_EAST_SEED_PZCT18_M1SS,
            NGA_EAST_SEED_PZCT18_M2ES,
            NGA_EAST_SEED_SP15,
            NGA_EAST_SEED_SP16,
            NGA_EAST_SEED_YA15));

    private final String name;
    private final List<Gmm> gmms;

    private Group(String name, List<Gmm> gmms) {
      this.name = name;
      this.gmms = gmms;
    }

    @Override
    public String toString() {
      return name;
    }

    /**
     * Return an immutable list of the Gmms in this group, sorted alphabetically
     * by their display name.
     */
    public List<Gmm> gmms() {
      return gmms;
    }
  }

}
