package gov.usgs.earthquake.nshmp.gmm;

import java.util.AbstractList;
import java.util.List;

import gov.usgs.earthquake.nshmp.Maths;
import gov.usgs.earthquake.nshmp.data.DoubleData;
import gov.usgs.earthquake.nshmp.tree.Branch;
import gov.usgs.earthquake.nshmp.tree.LogicTree;

/**
 * Factory class for creating and working with logic trees of ground motions
 * produced by ground motion models (GMMs).
 *
 * @author U.S. Geological Survey
 */
public class GroundMotions {

  private GroundMotions() {}

  /**
   * Returns the weighted combination of ground motions in this tree.
   *
   * <p>Note that the ground motion object returned contains a mean value
   * computed as the natural log of the weighted median ground motion (weighted
   * combination is done in linear space), and a sigma value computed as the
   * square root of the weighted sum of squares (variances) or SRSS. The
   * returned values should not be used for hazard calculations.
   *
   * <p>For single branch logic trees, the ground motion of the lone branch is
   * returned immediately. This can be a useful alternative to
   * {@code tree.get(0).value()}.
   *
   * @return the weighted combination of the ground motions in this tree
   */
  public static GroundMotion combine(LogicTree<GroundMotion> tree) {
    if (tree.size() == 1) {
      return tree.get(0).value();
    }
    double[] means = tree.stream()
        .map(Branch::value)
        .mapToDouble(GroundMotion::mean)
        .toArray();
    double[] sigmas = tree.stream()
        .map(Branch::value)
        .mapToDouble(GroundMotion::sigma)
        .toArray();
    double[] weights = tree.stream()
        .mapToDouble(Branch::weight)
        .toArray();
    double mean = DoubleData.weightedSumLn(means, weights);
    double sigma = Maths.srssWeighted(sigmas, weights);
    return GroundMotion.create(mean, sigma);
  }

  /* Create a logic tree with a single ground motion branch. */
  static LogicTree<GroundMotion> createTree(double mean, double sigma) {
    return new Singleton(GroundMotion.create(mean, sigma));
  }

  /*
   * Create a logic tree of gound motions given multiple means and a single
   * sigma.
   */
  static LogicTree<GroundMotion> createTree(
      String[] μIds, double[] μs, double[] μWts,
      double σ) {

    return new MultiMean(μIds, μs, μWts, σ);
  }

  /*
   * Create a logic tree of gound motions given multiple means and multiple
   * sigmas.
   */
  static LogicTree<GroundMotion> createTree(
      String[] μIds, double[] μs, double[] μWts,
      String[] σIds, double[] σs, double[] σWts) {

    return new MultiMeanMultiSigma(μIds, μs, μWts, σIds, σs, σWts);
  }

  private static final String TREE_NAME = "ground-motion-tree";
  static final String[] EPI_IDS = { "+uₑ", "~uₑ", "-uₑ" };
  static final double[] EPI_WTS = { 0.185, 0.63, 0.185 };

  /* Create a NGA-West symmetric 3-point logic tree of ground motion. */
  static LogicTree<GroundMotion> createNgaTree(
      GroundMotion gm,
      GmmInput in,
      NgaUncertainty model) {

    double epi = model.value(in.Mw, in.rJB);
    double μ = gm.mean();
    double[] μs = new double[] { μ + epi, μ, μ - epi };
    return createTree(EPI_IDS, μs, EPI_WTS, gm.sigma());
  }

  /**
   * Models of additional epistemic uncertainty in ground motion for use with
   * NGA-West1 and NGA-West2 GMMs. Models are parameterized in magnitude and
   * distance.
   */
  enum NgaUncertainty {

    /** Model derived from the PEER NGA-West1 database. */
    NGA_WEST1(new double[] { 0.375, 0.23, 0.4, 0.21, 0.225, 0.36, 0.245, 0.23, 0.31 }),

    /** Model derived from the PEER NGA-West2 database. */
    NGA_WEST2(new double[] { 0.37, 0.25, 0.4, 0.22, 0.23, 0.36, 0.22, 0.23, 0.33 });

    private final double[][] values;

    private NgaUncertainty(double[] values) {
      this.values = initValues(values);
    }

    /**
     * Return the epistemic uncertainty for the supplied magnitude and distance
     * in natural log units of ground motion.
     */
    double value(double m, double r) {
      int mi = (m < 6) ? 0 : (m < 7) ? 1 : 2;
      int ri = (r < 10) ? 0 : (r < 30) ? 1 : 2;
      return values[ri][mi];

    }

    private static double[][] initValues(double[] v) {
      return new double[][] {
          { v[0], v[1], v[2] },
          { v[3], v[4], v[5] },
          { v[6], v[7], v[8] }
      };
    }
  }

  static class Singleton extends
      AbstractList<Branch<GroundMotion>> implements
      LogicTree<GroundMotion> {

    private final GroundMotion gm;

    Singleton(GroundMotion gm) {
      this.gm = gm;
    }

    @Override
    public String name() {
      return TREE_NAME;
    }

    @Override
    public Branch<GroundMotion> sample(double probability) {
      throw new UnsupportedOperationException();
    }

    @Override
    public List<Branch<GroundMotion>> sample(double[] probabilities) {
      throw new UnsupportedOperationException();
    }

    @Override
    public int size() {
      return 1;
    }

    @Override
    public String toString() {
      return LogicTree.toString(this);
    }

    @Override
    public Branch<GroundMotion> get(int index) {
      return new Branch<GroundMotion>() {

        @Override
        public String id() {
          return TREE_NAME;
        }

        @Override
        public GroundMotion value() {
          return gm;
        }

        @Override
        public double weight() {
          return 1.0;
        }

        @Override
        public String toString() {
          return Branch.toString(this);
        }
      };
    }
  }

  static class MultiMean extends
      AbstractList<Branch<GroundMotion>> implements
      LogicTree<GroundMotion> {

    private final String[] ids;
    private final double[] μs;
    private final double[] wts;
    private final double σ;

    MultiMean(String[] ids, double[] μs, double[] wts, double σ) {
      this.ids = ids;
      this.μs = μs;
      this.wts = wts;
      this.σ = σ;
    }

    @Override
    public String name() {
      return TREE_NAME;
    }

    @Override
    public Branch<GroundMotion> sample(double probability) {
      throw new UnsupportedOperationException();
    }

    @Override
    public List<Branch<GroundMotion>> sample(double[] probabilities) {
      throw new UnsupportedOperationException();
    }

    @Override
    public int size() {
      return μs.length;
    }

    @Override
    public String toString() {
      return LogicTree.toString(this);
    }

    @Override
    public Branch<GroundMotion> get(int index) {
      return new Branch<GroundMotion>() {
        @Override
        public String id() {
          return ids[index];
        }

        @Override
        public GroundMotion value() {
          return GroundMotion.create(μs[index], σ);
        }

        @Override
        public double weight() {
          return wts[index];
        }

        @Override
        public String toString() {
          return Branch.toString(this);
        }
      };
    }
  }

  static class MultiMeanMultiSigma extends
      AbstractList<Branch<GroundMotion>> implements
      LogicTree<GroundMotion> {

    private final String[] μIds;
    private final double[] μs;
    private final double[] μWts;

    private final String[] σIds;
    private final double[] σs;
    private final double[] σWts;

    MultiMeanMultiSigma(
        String[] μIds, double[] μs, double[] μWts,
        String[] σIds, double[] σs, double[] σWts) {

      this.μIds = μIds;
      this.μs = μs;
      this.μWts = μWts;
      this.σIds = σIds;
      this.σs = σs;
      this.σWts = σWts;
    }

    @Override
    public String name() {
      return TREE_NAME;
    }

    @Override
    public Branch<GroundMotion> sample(double probability) {
      throw new UnsupportedOperationException();
    }

    @Override
    public List<Branch<GroundMotion>> sample(double[] probabilities) {
      throw new UnsupportedOperationException();
    }

    @Override
    public int size() {
      return μs.length * σs.length;
    }

    @Override
    public String toString() {
      return LogicTree.toString(this);
    }

    @Override
    public Branch<GroundMotion> get(int index) {
      return new Branch<GroundMotion>() {

        final int μIndex;
        final int σIndex;

        {
          μIndex = index / σs.length;
          σIndex = index % σs.length;
        }

        @Override
        public String id() {
          return μIds[μIndex] + " : " + σIds[σIndex];
        }

        @Override
        public GroundMotion value() {
          return GroundMotion.create(μs[μIndex], σs[σIndex]);
        }

        @Override
        public double weight() {
          return μWts[μIndex] * σWts[σIndex];
        }

        @Override
        public String toString() {
          return Branch.toString(this);
        }
      };
    }
  }
}
