package gov.usgs.earthquake.nshmp.model;

import static com.google.common.base.Preconditions.checkNotNull;
import static com.google.common.base.Preconditions.checkState;
import static com.google.common.base.StandardSystemProperty.LINE_SEPARATOR;
import static gov.usgs.earthquake.nshmp.Text.checkName;
import static gov.usgs.earthquake.nshmp.model.SourceType.FAULT_CLUSTER;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import java.util.Map;

import gov.usgs.earthquake.nshmp.geo.Location;
import gov.usgs.earthquake.nshmp.geo.LocationList;
import gov.usgs.earthquake.nshmp.geo.Locations;
import gov.usgs.earthquake.nshmp.mfd.Mfd;

/**
 * Cluster source representation. Each cluster source wraps a
 * {@code FaultRuptureSet} containing one or more fault representations that
 * rupture as independent events but with a similar rate. For example, at New
 * Madrid, each ClusterRuptureSet has 5 ClusterSources, one for each position
 * variant of the model. For each position variant there is one FaultRuptureSet
 * containing the FaultSources in the cluster, each of which may have one, or
 * more, magnitude or other variants represented by its internal list of
 * {@code IncrementalMfd}s.
 *
 * <p>Cluster source hazard is calculated from the joint probabilities of ground
 * motions from the wrapped faults, which is handled internally by a separate
 * calculator and {@link ClusterSource#iterator()} therefore throws an
 * {@code UnsupportedOperationException}.
 *
 * <p>A {@code ClusterSource} cannot be created directly; it may only be created
 * by a private parser.
 *
 * @author Peter Powers
 */
public class ClusterSource implements Source {

  final String name;
  final int id;
  final double rate;
  final List<FaultRuptureSet> faults;

  ClusterSource(Builder builder) {
    this.name = builder.name;
    this.id = builder.id;
    this.rate = builder.rate;
    this.faults = builder.faults;
  }

  @Override
  public String name() {
    return name;
  }

  @Override
  public int size() {
    return faults.size();
  }

  @Override
  public int id() {
    return id; // same as cluster rupture set id
  }

  @Override
  public SourceType type() {
    return FAULT_CLUSTER;
  }

  /**
   * The closest point across the traces of all fault sources that participate
   * in this cluster, relative to the supplied site {@code Location}.
   */
  @Override
  public Location location(Location site) {
    LocationList.Builder locs = LocationList.builder();
    for (FaultRuptureSet fault : faults) {
      locs.add(fault.location(site));
    }
    return Locations.closestPoint(site, locs.build());
  }

  @Override
  public List<Mfd> mfds() {
    // TODO this should be consistent with what is done in Models
    List<Mfd> xyMfds = new ArrayList<>();
    for (FaultRuptureSet frs : faults) {
      checkState(frs.size() == 1); // TODO bad assumption of single source
      for (Mfd mfd : frs.iterator().next().mfds()) {
        xyMfds.add(Mfd.Builder.from(mfd).scale(rate).build());
      }
    }
    return List.copyOf(xyMfds);
  }

  /**
   * {@code (1 / return period)} of this source in years.
   * @return the cluster rate
   */
  public double rate() {
    return rate;
  }

  /**
   * The {@code FaultRuptureSet} of all {@code FaultSource}s that participate in
   * this cluster.
   */
  public List<FaultRuptureSet> faults() {
    return faults;
  }

  /**
   * Overriden to throw an {@code UnsupportedOperationException}. Cluster
   * sources are handled differently than other source types.
   */
  @Override
  public Iterator<Rupture> iterator() {
    throw new UnsupportedOperationException();
  }

  @Override
  public String toString() {
    Map<String, ?> data = Map.of(
        "name", name(),
        "rate", rate());
    StringBuilder sb = new StringBuilder()
        .append(getClass().getSimpleName())
        .append(" ")
        .append(data)
        .append(LINE_SEPARATOR.value());
    for (FaultRuptureSet frs : faults) {
      sb.append("  ")
          .append(frs.toString())
          .append(LINE_SEPARATOR.value());
    }
    return sb.toString();
  }

  /* Single use builder */
  static class Builder {

    boolean built = false;

    String name;
    Integer id;
    Double rate;
    List<FaultRuptureSet> faults;

    Builder name(String name) {
      this.name = checkName(name, "ClusterSource");
      return this;
    }

    Builder id(int id) {
      this.id = id;
      return this;
    }

    Builder rate(double rate) {
      // What sort of value checking should be done for rate (<1 ??)
      this.rate = rate;
      return this;
    }

    Builder faults(List<FaultRuptureSet> faults) {
      checkState(checkNotNull(faults, "Fault source set is null").size() > 0,
          "Fault source set is empty");
      this.faults = faults;
      return this;
    }

    void validateState(String label) {
      checkState(!built, "Single use builder");
      checkNotNull(name, "%s name", label);
      checkNotNull(id, "%s id", label);
      checkNotNull(rate, "%s rate", label);
      checkNotNull(faults, "%s fault rupture sets", label);
      built = true;
    }

    ClusterSource buildClusterSource() {
      validateState("ClusterSource.Builder");
      return new ClusterSource(this);
    }
  }

}
