package gov.usgs.earthquake.nshmp.model;

import static com.google.common.base.Preconditions.checkArgument;
import static com.google.common.base.Preconditions.checkNotNull;
import static com.google.common.base.Preconditions.checkState;
import static gov.usgs.earthquake.nshmp.model.Deserialize.MFD_TREE;
import static gov.usgs.earthquake.nshmp.model.Deserialize.RATE_TREE;
import static gov.usgs.earthquake.nshmp.model.RateType.RECURRENCE;
import static java.lang.Math.log10;
import static java.util.stream.Collectors.collectingAndThen;
import static java.util.stream.Collectors.toList;

import java.math.RoundingMode;
import java.util.Iterator;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Map;
import java.util.Optional;
import java.util.OptionalDouble;
import java.util.Set;
import java.util.function.Predicate;
import java.util.stream.Stream;

import gov.usgs.earthquake.nshmp.Earthquakes;
import gov.usgs.earthquake.nshmp.Faults;
import gov.usgs.earthquake.nshmp.Maths;
import gov.usgs.earthquake.nshmp.geo.Location;
import gov.usgs.earthquake.nshmp.geo.LocationList;
import gov.usgs.earthquake.nshmp.geo.Locations;
import gov.usgs.earthquake.nshmp.geo.json.Feature;
import gov.usgs.earthquake.nshmp.geo.json.Properties;
import gov.usgs.earthquake.nshmp.mfd.Mfd;
import gov.usgs.earthquake.nshmp.mfd.Mfd.Properties.GutenbergRichter;
import gov.usgs.earthquake.nshmp.mfd.Mfd.Properties.Single;
import gov.usgs.earthquake.nshmp.mfd.Mfds;
import gov.usgs.earthquake.nshmp.model.MfdConfig.AleatoryProperties;
import gov.usgs.earthquake.nshmp.model.SourceFeature.Key;
import gov.usgs.earthquake.nshmp.model.SourceFeature.NshmFault.Rates;
import gov.usgs.earthquake.nshmp.tree.Branch;
import gov.usgs.earthquake.nshmp.tree.LogicTree;

/**
 * Crustal finite fault rupture set.
 *
 * @author U.S. Geological Survey
 */
public class FaultRuptureSet extends AbstractRuptureSet<FaultSource> {

  /*
   * Developer notes:
   *
   * The copying of ModelData when processing single branches is likely not
   * necessary as the feature map is all that's being used and in this instnace
   * it's ok to overwrite it each time.
   *
   * General comments on fault earthquake rates. The majority of faults in NSHMs
   * are constrained by slip rates but some are governed by explicit logic trees
   * of slip and/or recurrence rates with pre-defined MFD parameters. In both
   * cases we require moment rate in order to correctly balance MFD logic tree
   * branches of epistemic and aleatory uncertainty
   *
   * This ends up being a little circular for GR with no epistemic uncertainty
   * in that we consrtuct an MFD, scale to it's incremental rate, compute its
   * moment rate, only to then build the MFD again scaling directly to moment
   * rate.
   *
   * We removed aleaProps - MoBalance; 2008 CA a-priori a-faults do not moment
   * balance so we'll need to figure out how to handle that; it's consistent
   * with SSCn SINGLE mag trees; slip-rate based mfds mo-balance, others don't
   *
   */
  final SourceFeature.NshmFault feature;
  final ModelData data; // is this needed beyond the builder?

  final LogicTree<Mfd> mfdTree;
  final Mfd mfdTotal;

  final List<Integer> sectionIds; // reference: actually needed?
  private final List<FaultSource> source; // single source list

  FaultRuptureSet(Builder builder) {
    super(builder);
    this.feature = builder.feature;
    this.data = builder.data;

    this.mfdTree = builder.mfdTree;
    this.mfdTotal = builder.mfdTotal;

    this.sectionIds = builder.sectionIds;
    this.source = List.of(createSource());
  }

  private FaultSource createSource() {
    return new FaultSource.Builder()
        .feature(feature)
        .config(data.sourceConfig().asFault())
        .mfdTree(mfdTree)
        .mfdTotal(mfdTotal)
        .build();
  }

  @Override
  public Iterator<FaultSource> iterator() {
    return source.iterator();
  }

  @Override
  public int size() {
    return source.size(); // TODO should this be # ruptures
  }

  @Override
  public LogicTree<Mfd> mfdTree() {
    return mfdTree;
  }

  @Override
  public Predicate<FaultSource> distanceFilter(Location loc, double distance) {
    return new DistanceFilter(loc, distance);
  }

  /* Not inlined for use by cluster sources */
  static class DistanceFilter implements Predicate<FaultSource> {
    private final Predicate<Location> filter;

    DistanceFilter(Location loc, double distance) {
      filter = Locations.distanceFilter(loc, distance);
    }

    @Override
    public boolean test(FaultSource source) {
      return filter.test(source.feature.trace.first()) ||
          filter.test(source.feature.trace.last());
    }
  }

  /**
   * The closest point on the fault trace, relative to the supplied site
   * {@code Location}.
   */
  @Override
  public Location location(Location site) {
    return Locations.closestPoint(site, feature.trace);
  }

  static Builder builder() {
    return new Builder();
  }

  /* Single use builder */
  static class Builder extends AbstractRuptureSet.Builder {

    private SourceFeature.NshmFault feature;
    private ModelData data;

    private LogicTree<Mfd.Properties> mfdPropsTree;

    /* created on build */
    private LogicTree<Mfd> mfdTree;
    private Mfd mfdTotal;

    private List<Integer> sectionIds;
    private Optional<Map<String, ?>> ruptureProps = Optional.empty();

    private Builder() {
      super(SourceType.FAULT);
    }

    @Override
    Builder name(String name) {
      super.name(name);
      return this;
    }

    @Override
    Builder id(int id) {
      super.id(id);
      return this;
    }

    Builder sections(List<Integer> sections) {
      checkArgument(sections.size() >= 1);
      this.sectionIds = List.copyOf(sections);
      return this;
    }

    Builder properties(Map<String, ?> props) {
      this.ruptureProps = Optional.of(checkNotNull(props));
      return this;
    }

    /* Set the MFD tree. */
    Builder mfdTree(LogicTree<Mfd.Properties> mfdTree) {
      this.mfdPropsTree = mfdTree;
      return this;
    }

    /* Set MFD helper data. */
    Builder modelData(ModelData data) {
      this.data = checkNotNull(data);
      super.setting(data.tectonicSetting());
      super.gmms(data.gmms());
      return this;
    }

    FaultRuptureSet build() {
      validateAndInit("FaultRuptureSet.Builder");
      return new FaultRuptureSet(this);
    }

    private void validateAndInit(String label) {
      validateState(label);

      checkNotNull(data, "%s model data", label);
      checkNotNull(sectionIds, "%s section id list", label);
      checkNotNull(mfdPropsTree, "%s MFD logic tree", label);

      feature = createFeature();

      MfdConfig mfdConfig = data.mfdConfig().orElseThrow();

      /* Filter of unlikely combination of SINGLEs epistemic uncertainty. */
      checkEpistemic(mfdConfig, mfdPropsTree);

      Optional<LogicTree<MoRate>> moRateTree = Optional.empty();

      /* Build mfd-tree from fault slip or recurrence rates. */
      if (data.faultRateTree().isPresent() && !data.isFaultSourceTree()) {

        /* Create Mo-rate tree. */
        moRateTree = Optional.of(createMoRateTree(feature, data));

        /* Update MFD properties tree based on source magnitude. */
        mfdPropsTree = updateMfdsForSlip();

      } else if (data.rateTree().isPresent() && !data.isClusterModel()) {

        /* Update MFD properties tree to include M and R branches. */
        mfdPropsTree = updateMfdsForRecurrence();
      }

      mfdTree = createMfdTree(mfdPropsTree, moRateTree, mfdConfig);
      mfdTotal = ModelTrees.reduceMfdTree(mfdTree);
    }

    private SourceFeature.NshmFault createFeature() {

      Map<Integer, SourceFeature.NshmFault> featureMap =
          data.faultFeatureMap().orElseThrow();
      checkState(featureMap.keySet().containsAll(sectionIds));

      if (sectionIds.size() == 1) {
        // If rupture set for single feature has different name
        // it is currently not written into feature; only if ruptureProps
        // present is feature updated; can we get rid of name in single
        // features?
        var fault = featureMap.get(sectionIds.get(0));
        if (ruptureProps.isPresent()) {

          Properties.Builder props = Properties.fromFeature(fault.source);
          ruptureProps.orElseThrow().forEach(props::put);
          updateState(props, fault.states);
          Feature newFeature = Feature.copyOf(fault.source)
              .properties(props.build())
              .build();
          fault = new SourceFeature.NshmFault(newFeature);
        }
        return fault;
      }

      /* Collect participating features */
      List<SourceFeature.NshmFault> features = sectionIds.stream()
          .map(featureMap::get)
          .collect(toList());

      /* Consolidate states */
      List<String> states = Stream.concat(
          ruptureSetState(),
          features.stream()
              .map(f -> f.states)
              .flatMap(List::stream))
          .distinct()
          .collect(toList());

      /* Join section traces */
      LocationList trace = features.stream()
          .map(f -> f.source)
          .map(Feature::asLineString)
          .flatMap(LocationList::stream)
          .distinct()
          .collect(collectingAndThen(
              toList(),
              LocationList::copyOf));

      /* Update properties */
      Properties.Builder props = Properties.fromFeature(features.get(0).source)
          .put(Key.NAME, name)
          .put(Key.STATES, states);
      ruptureProps.ifPresent(p -> p.forEach(props::put));

      /* Build new feature from stitched traces and props of 1st section */
      Feature feature = Feature.lineString(trace)
          .id(id)
          .properties(props.build())
          .build();

      return new SourceFeature.NshmFault(feature);
    }

    /* For single feature rupture sets. */
    private void updateState(Properties.Builder props, List<String> states) {
      String state = (String) ruptureProps.orElseThrow().get(Key.STATE);
      if (state != null) {
        props.put(Key.STATE, state);
        Set<String> stateSet = new LinkedHashSet<>();
        stateSet.add(state);
        stateSet.addAll(states);
        props.put(Key.STATES, List.copyOf(stateSet));
      }
    }

    /* For multi feature rupture sets. */
    private Stream<String> ruptureSetState() {
      return ruptureProps.isEmpty()
          ? Stream.empty()
          : Optional.ofNullable((String) ruptureProps.orElseThrow().get(Key.STATE)).stream();
    }

    /* Create mfd-tree from a logic tree of SINGLE MFDs and a rate-tree. */
    private LogicTree<Mfd.Properties> updateMfdsForRecurrence() {
      ModelTrees.checkType(mfdPropsTree, Mfd.Type.SINGLE);
      LogicTree.Builder<Mfd.Properties> propsTree = LogicTree.builder(mfdPropsTree.name());
      LogicTree<Double> rateTree = data.rateTree().orElseThrow();
      for (Branch<Double> rBranch : rateTree) {
        /*
         * Need to be careful here; this catch is for additive New Madrid
         * branches; null rates have been converted to 0; perhaps we should use
         * 0 in JSON and still have this 0 check here; we also use 0 rates to
         * deal with imbalanced geodetic logic tree branches so maybe just
         * document these cases
         *
         */
        double rate = (rBranch.value() <= 0.0)
            ? 0.0
            : (rBranch.value() <= 1.0)
                ? rBranch.value() // rate in 1/yr
                : Maths.roundToDigits(1.0 / rBranch.value(), 4); // rate in yr
        // Above line and rounding should go away, we'll create labels
        // from 1/rate; all rate-trees will have annual rate, not years
        for (Branch<Mfd.Properties> mBranch : mfdPropsTree) {
          Single props = mBranch.value().getAsSingle();
          String id = String.join(":", props.type().name(), mBranch.id(), rBranch.id());
          double weight = mBranch.weight() * rBranch.weight();
          propsTree.addBranch(id, new Single(props.magnitude(), rate), weight);
        }
      }
      return propsTree.build();
    }

    /*
     * The properties picked up from a mfd-tree for faults with slip rates are
     * only used to set GR mMin and b-value; all other fields are just required
     * placeholders. This method uses the 'characteristic' magnitude of the
     * feature to set the SINGLE magnitude and GR Δm and mMax.
     */
    private LogicTree<Mfd.Properties> updateMfdsForSlip() {
      LogicTree.Builder<Mfd.Properties> propsTree = LogicTree.builder(mfdPropsTree.name());
      double Mw = feature.magnitude.orElseThrow();
      for (Branch<Mfd.Properties> branch : mfdPropsTree) {
        switch (branch.value().type()) {
          case GR:
            MfdConfig mfdConfig = data.mfdConfig().orElseThrow();
            double mCutoff = mfdConfig.minMagnitude.orElseThrow();
            propsTree.addBranch(
                branch.id(),
                updateGr(branch.value(), Mw, mCutoff, mfdConfig.nshmBinModel),
                branch.weight());
            break;
          case SINGLE:
            propsTree.addBranch(
                branch.id(),
                new Mfd.Properties.Single(Mw),
                branch.weight());
            break;
          default:
            throw new UnsupportedOperationException(
                branch.value().type() + " MFD not supported");
        }
      }
      return propsTree.build();
    }

    /*
     * Update GR properties for fault 'characteristic' magnitude, or return
     * SINGLE properties instead if mMax < mCutoff or nm = 1.
     */
    private static Mfd.Properties updateGr(
        Mfd.Properties props,
        double mMax,
        double mCutoff,
        boolean nshmBinModel) {

      Mfd.Properties.GutenbergRichter grProps = props.getAsGr();

      /* Return SINGLE for m <= GR minimum. */
      if (mMax <= mCutoff) {
        return new Mfd.Properties.Single(mMax);
      }

      double Rm = Maths.round(mMax - grProps.mMin(), 6);

      /* Return custom GR if GR only has one magnitude bin. */
      if (Rm <= 0.1) {
        double Δm = Maths.round(mMax - grProps.mMin(), 4);
        return new Mfd.Properties.GutenbergRichter(1.0, grProps.b(), Δm, grProps.mMin(), mMax);
      }

      /*
       * USGS NSHM Gutenberg-Richter discretization algorithm. Cutoff for nm = 4
       * in fltrate*.f is >0.3, but in reality M6.8 events have 4 GR magnitudes
       * in input files. This applies to most WUS faults, however a few use the
       * discretization specified in the MFD logic tree (e.g. Seattle)
       */
      int nm = Rm > 0.7 ? 8 : Rm > 0.5 ? 6 : Rm >= 0.3 ? 4 : 2;
      double Δm = grProps.Δm();
      if (nshmBinModel) {
        Δm = Rm / nm;
        if (Δm == 0.09625 ||
            // -- Pyramid Lake, NV
            Δm == 0.12375 ||
            // -- Wassuk Range
            // -- Kawich - Hot Creek... NV
            Δm == 0.10875 ||
            // -- Smith Valley, NV
            Δm == 0.09875 ||
            // -- Southwest Reese Rvr Vly, NV
            // -- Tijeras... NM
            // -- Eastern Bear Lake, ID
            // -- Black Rock, NV
            Δm == 0.11625 ||
            // -- Schell Creek Rge, NV
            Δm == 0.10375 ||
            // -- Sevier - Toroweap (northern), UT
            Δm == 0.10125 ||
            // -- East Cache, UT
            Δm == 0.11125 ||
            // -- Picuris - Pecos, NM
            Δm == 0.11375 ||
            // -- Montanan Mts - Desert Vly, NV
            // -- Hoppin Peaks, NV
            Δm == 0.11875)
        // -- Madison, MT
        {
          Δm = Maths.round(Δm, 4, RoundingMode.DOWN);
        } else {
          Δm = Maths.round(Δm, 4);
        }
      }

      // double Δm = nshmBinModel ? Maths.round(Rm / nm, 4) : grProps.Δm();
      // System.out.println(grProps.mMin() + " " + mMax + " " + Δm);
      return new Mfd.Properties.GutenbergRichter(1.0, grProps.b(), Δm, grProps.mMin(), mMax);
    }

    /*
     * Consider short citcuiting singletons if feature as represented in model
     * is consistent with rupture set requirements
     *
     * For documentation; there are many options avilable when defining both
     * fault and grid based sources that may necessarily be able to be combined.
     * For example, normal fault dip variants are difficult to handle in the
     * context of a cluster model.
     */

  }

  /* Logic tree of moment rates. */
  private static LogicTree<Mfd> createMfdTree(
      LogicTree<Mfd.Properties> mfdPropsTree,
      Optional<LogicTree<MoRate>> moRateTree,
      MfdConfig mfdConfig) {

    LogicTree.Builder<Mfd> mfdTree = LogicTree.builder(MFD_TREE);
    if (moRateTree.isPresent()) {
      LogicTree<MoRate> moTree = moRateTree.orElseThrow();
      moTree.forEach(branch -> addMfds(mfdTree, mfdPropsTree, Optional.of(branch), mfdConfig));
    } else {
      addMfds(mfdTree, mfdPropsTree, Optional.empty(), mfdConfig);
    }
    return mfdTree.build();
  }

  private static void addMfds(
      LogicTree.Builder<Mfd> mfdTree, // receiver tree
      LogicTree<Mfd.Properties> mfdPropsTree,
      Optional<Branch<MoRate>> moBranch,
      MfdConfig mfdConfig) {

    /*
     * Working with branch list because weights may not sum to 1.0 as we loop
     * over moment rate branches to build receiver mfdTree. It's simpler to
     * update branch names and weights here than do it in both Single and GR MFD
     * factories below.
     *
     * Update properties tree with mo rate branch name and weight.
     */
    List<Branch<Mfd.Properties>> propsBranchList = moBranch.isPresent()
        ? ModelTrees.toBranchList(mfdPropsTree, moBranch.orElseThrow())
        : mfdPropsTree;
    Optional<MoRate> moRateIn = moBranch.map(Branch::value);

    for (Branch<Mfd.Properties> mfdBranch : propsBranchList) {
      double moRate = computeMoRate(mfdBranch.value(), moRateIn);

      switch (mfdBranch.value().type()) {
        case SINGLE:
          addSingleMfds(mfdTree, mfdBranch, moRate, mfdConfig);
          break;
        case GR:
          addGrMfds(mfdTree, mfdBranch, moRate, mfdConfig);
          break;
        case INCR:
          // INCR does not support epi and alea uncurtainty
          // and doesn't require special handling or mo-balancing
          Mfd mfd = mfdBranch.value().toBuilder().build();
          mfdTree.addBranch(mfdBranch.id(), mfd, mfdBranch.weight());
          break;
        default:
          throw new UnsupportedOperationException(
              mfdBranch.value().type() + " MFD not supported");
      }
    }
  }

  /*
   * Container class for slip-rate based moRate.
   *
   * Optional fields, if present, are used downstream to override slip-based
   * moRates to yield better agreement with legacy codes.
   */
  private static class MoRate {

    final double moRate;
    final OptionalDouble graValue;
    final OptionalDouble singleRate;

    MoRate(
        double moRate,
        OptionalDouble graValue,
        OptionalDouble singleRate) {

      this.moRate = moRate;
      this.graValue = graValue;
      this.singleRate = singleRate;
    }
  }

  private static double computeMoRate(Mfd.Properties props, Optional<MoRate> moRateData) {

    /* No mo-rate data supplied, compute from properties */
    if (moRateData.isEmpty()) {
      switch (props.type()) {
        case SINGLE:
          Mfd.Properties.Single single = props.getAsSingle();
          return single.rate() * Earthquakes.magToMoment(single.magnitude());
        case GR:
          Mfd.Properties.GutenbergRichter gr = props.getAsGr();
          return grMomentRate(gr);
        case INCR:
          return Mfds.momentRate(props.toBuilder().build());
        default:
          throw new UnsupportedOperationException(
              props.type() + " MFD not supported");
      }
    }

    /* Return supplied mo-rate or compute from updated properties. */
    MoRate rateData = moRateData.orElseThrow();
    switch (props.type()) {
      case SINGLE:
        if (rateData.singleRate.isPresent()) {
          return rateData.singleRate.orElseThrow() *
              Earthquakes.magToMoment(props.getAsSingle().magnitude());
        }
        return rateData.moRate;
      case GR:
        if (rateData.graValue.isPresent()) {
          Mfd.Properties.GutenbergRichter grNew =
              GutenbergRichter.copyOfWithRate(props.getAsGr(), rateData.graValue.orElseThrow());
          return grMomentRate(grNew);
        }
        return rateData.moRate;
      default:
        throw new UnsupportedOperationException(
            props.type() + " MFD not supported");
    }
  }

  /*
   * Create a logic tree of moment rates for a fault source given a feature and
   * rate data. If present, a rate-dip may be used to compute a constraining
   * moment rate (e.g. WUS normal faults)
   */
  private static LogicTree<MoRate> createMoRateTree(
      SourceFeature.NshmFault feature,
      ModelData data) {

    double dip = feature.dip;
    double width = feature.width;

    // Needs revisiting for 2008 and/or
    // future models if slip rate is to be used
    // rateDip should not be present for 2014/2018 models

    /* Dip slip model dependent parameters. */
    if (feature.rateDip.isPresent()) {
      dip = feature.rateDip.orElseThrow();
      double height = feature.lowerDepth - feature.upperDepth;
      width = Faults.projectVertical(dip, height);
    }

    double length = feature.length;
    double area = length * width * 1e6;
    double Mw = feature.magnitude.orElseThrow();
    double mo = Earthquakes.magToMoment(Mw);

    Map<String, Rates> rateMap = feature.rateMap;
    RateType rateType = feature.rateType;
    LogicTree.Builder<MoRate> moRateTree = LogicTree.builder(RATE_TREE);

    for (Branch<String> branch : data.faultRateTree().orElseThrow()) {

      /*
       * Verify missing fault rates and imbalanced logic trees. In some cases
       * (e.g. TX) need to give GEO branch weight of 1.0. To do this we should
       * actually put geo rate in BIRD and ZENG branches to preserve logic tree.
       * In other cases (e.g. OR offshore) we need empty branches because GEO
       * weight is 0.8 and imbalanced; maybe we just handle this with setting
       * rate to 0, as below.
       */
      Rates rates = rateMap.get(branch.id());

      double rate = (rates != null) ? rates.rate : 0.0;

      if (rateType == RateType.VERTICAL_SLIP) {
        rate = Faults.projectVertical(dip, rate);
      }

      double moRateValue = (rateType == RECURRENCE)
          ? mo * rate
          : Earthquakes.moment(area, rate / 1000.0);

      MoRate moRate = (rates != null)
          ? new MoRate(moRateValue, rates.graValue, rates.singleRate)
          : new MoRate(moRateValue, OptionalDouble.empty(), OptionalDouble.empty());
      moRateTree.addBranch(branch.id(), moRate, branch.weight());
    }
    return moRateTree.build();
  }

  /* Fault GR */
  private static void addGrMfds(
      LogicTree.Builder<Mfd> mfdTree, // receiver tree
      Branch<Mfd.Properties> mfdBranch,
      double moRate,
      MfdConfig mfdConfig) {

    Mfd.Properties.GutenbergRichter gr = mfdBranch.value().getAsGr();

    String mfdId = mfdBranch.id();
    double mfdWt = mfdBranch.weight();

    // Does this ever throw?
    int nMag = magCount(gr.mMin(), gr.mMax(), gr.Δm());
    checkState(nMag > 0, "GR MFD with no magnitudes [%s]", mfdBranch.id());

    /*
     * Optional epistemic uncertainty; no aleatory with GR. This allows
     * epistemic uncertainty because the cutoff is based on mMin (6.5), but in
     * reality a few faults end up with no magnitudes in their loe
     */
    boolean epistemic = hasEpistemic(mfdConfig, gr.mMax() - gr.Δm() / 2);

    /*
     * Edge cases (Rush Peak in brange.3dip.gr.in) suggest data.mMin should be
     * used instead of unc.epiCutoff
     *
     * Note in documentation that epi ±0.2 is applied to mMax-Δm/2, not mMax
     *
     * Conduct sensitivity study for epi being applied to actual mMax, not
     * centered mMax for GR MFD
     */

    if (epistemic) {

      for (Branch<Double> epiBranch : mfdConfig.epistemicTree.orElseThrow()) {

        /*
         * Given changes to how GR MFD sequences are built (centered and bounded
         * by the upper a lower bin edges), M + epi - Δm/2 commonly cuts off the
         * upper bin. As a result we add Δm/2 knowing it will be removed (e.g.
         * for M7, M + epi 7.2 as upper limit of sequence).
         */
        // double mMaxEpi = gr.mMax() + epiBranch.value() + gr.Δm() / 2;

        // reverting to fortran methodolgy
        int nMagEpi = magCount(gr.mMin(), gr.mMax() + epiBranch.value(), gr.Δm());
        double mMaxEpi = Maths.round(gr.mMin() + nMagEpi * gr.Δm() + 0.0001, 6);

        // System.out.println(gr.mMin());
        // System.out.println(gr.mMax());
        // System.out.println(gr.mMin() + " " + gr.mMax() + " " + gr.Δm() + " "
        // + mMaxEpi);

        double weightEpi = mfdWt * epiBranch.weight();

        String id = ModelTrees.combineBranchIds(mfdId, epiBranch.id());

        // Check if ever thrown

        if (nMagEpi == 0) {

          System.err.println("                 WARNING: Empty GR MFD epi branch");
          /*
           * There are a handful of sources in legacy models for which the lower
           * epistemic branch of the GR logic tree yields no bins. Prior codes
           * simply skipped these ruptures, however, we now need a complete
           * logic tree if MFDs. ALthough we could use a SINGLE Mfd here, we use
           * GR to preserve the original MFD parameterization.
           */
          double mMax = Maths.round(gr.mMin() + gr.Δm() + 0.0001, 6);
          Mfd mfd = Mfd.newGutenbergRichterBuilder(
              gr.b(), gr.Δm(), gr.mMin(), mMax)
              .scaleToMomentRate(0.0)
              .build();
          mfdTree.addBranch(id, mfd, weightEpi);

        } else {

          Mfd mfd = GutenbergRichter.copyOfWithMax(gr, mMaxEpi)
              .toBuilder()
              .scaleToMomentRate(moRate)
              .build();

          mfdTree.addBranch(id, mfd, weightEpi);
        }
      }
    } else {

      Mfd mfd = gr.toBuilder()
          .scaleToMomentRate(moRate)
          .build();
      mfdTree.addBranch(mfdId, mfd, mfdWt);
    }
  }

  /*
   * Determines the number of magnitude bins for the supplied arguments
   * consistent with approach used in legacy fortran codes. If Δm does not
   * divide evenly into {@code mMax - mMin}, and the result of this method is
   * used to build a Gutenberg-Richter MFD, the maximum magnitude of the MFD may
   * not equal the {@code mMax} supplied here. Method assumes mMin and mMax are
   * bin edges, not centers.
   */
  private static int magCount(double mMin, double mMax, double Δm) {
    checkArgument(Δm > 0.0);
    return (int) ((mMax - mMin - Δm) / Δm + 1.4);
    // return (int) ((mMax - mMin) / Δm + 1.4); (old) assumed bin centers
  }

  private static boolean checkEpi(double mMin, double mMax, double Δm, double Δepi) {
    return (mMax - Δm / 2 + Δepi) > (mMin + Δm / 2);
  }

  /* Fault SINGLE */
  private static void addSingleMfds(
      LogicTree.Builder<Mfd> mfdTree, // receiver tree
      Branch<Mfd.Properties> mfdBranch,
      double moRate,
      MfdConfig mfdConfig) {

    double magnitude = mfdBranch.value().getAsSingle().magnitude();

    /* Optional epistemic uncertainty. */
    boolean epistemic = mfdConfig.epistemicTree.isPresent();

    /* Optional aleatory variability. */
    Optional<AleatoryProperties> aleaProps = mfdConfig.aleatoryProperties;

    /* mMax epistemic uncertainty branches */
    if (epistemic) {

      for (Branch<Double> epiBranch : mfdConfig.epistemicTree.orElseThrow()) {

        double mEpi = Maths.round(magnitude + epiBranch.value(), 6);
        double wtEpi = mfdBranch.weight() * epiBranch.weight();
        String id = ModelTrees.combineBranchIds(mfdBranch.id(), epiBranch.id());

        Mfd mfd = createSingleMfd(mEpi, moRate, aleaProps);
        mfdTree.addBranch(id, mfd, wtEpi);

      }

    } else { /* No epistemic uncertainty */

      Mfd mfd = createSingleMfd(magnitude, moRate, aleaProps);
      mfdTree.addBranch(mfdBranch.id(), mfd, mfdBranch.weight());

    }
  }

  private static Mfd createSingleMfd(
      double magnitude, double moRate,
      Optional<AleatoryProperties> aleaProps) {

    Mfd.Builder builder = (aleaProps.isPresent())
        ? newGaussianBuilder(magnitude, aleaProps.orElseThrow())
        : Mfd.newSingleBuilder(magnitude);
    return builder.scaleToMomentRate(moRate).build();
  }

  /* Logic only applies to GR MFDs. */
  private static boolean hasEpistemic(MfdConfig mfdConfig, double magnitude) {
    boolean epistemic = mfdConfig.epistemicTree.isPresent();
    /* No epistemic uncertainty if: m - Δepi < mfdConfig.minMagnitude */
    if (epistemic) {
      double mEpiCutoff = mfdConfig.minMagnitude.orElseThrow();
      double mEpiOffset = mfdConfig.minEpiOffset.orElseThrow();
      epistemic = (magnitude + mEpiOffset) > mEpiCutoff;
    }
    return epistemic;
  }

  /*
   * It is unlikely that a logic tree of SINGLE MFDs will be combined with MFD
   * epistemic uncertainty on magnitude, usually one or the other representation
   * is preferred.
   */
  static void checkEpistemic(MfdConfig mfdConfig, LogicTree<Mfd.Properties> mfdTree) {
    boolean multipleSingleBranches = (mfdTree.size() > 1) &&
        ModelTrees.checkTreeTypes(mfdTree, Mfd.Type.SINGLE);
    boolean hasEpistemic = mfdConfig.epistemicTree.isPresent();
    if (multipleSingleBranches && hasEpistemic) {
      System.err.println(
          "                 WARNING: Multiple SINGLE MFDs with epistemic uncertainty");
    }
    // checkState(
    // !(multipleSingleBranches && hasEpistemic),
    // "Multiple SINGLE MFDs incompatable with M epistemic uncertainty");
  }

  /*
   * The expected magnitude for the supplied length rounded to two decimal
   * places. Note that the function used here is not the inverted form of that
   * used in RuptureScaling.NSHM_FAULT_WC94_LENGTH.dimensions() See Wells &
   * Coppersmith (1994) Table 2A.
   */
  static double magnitudeWc94(double length) {
    return Maths.round(5.08 + 1.16 * log10(length), 2);
  }

  /* Gaussian MFD builder. */
  static Mfd.Builder newGaussianBuilder(double m, MfdConfig.AleatoryProperties aleaProps) {
    return Mfd.newGaussianBuilder(
        m,
        aleaProps.size,
        aleaProps.σ,
        aleaProps.nσ);
  }

  /* Expected moment rate of a GR MFD. */
  static double grMomentRate(Mfd.Properties.GutenbergRichter gr) {

    // Ensure that mMax is large neough to match
    // the MFD that will be created for epi branches
    gr = Mfd.Properties.GutenbergRichter.copyOfWithMax(gr, gr.mMax() + 0.001);

    double mMin = gr.mMin() + gr.Δm() / 2;
    double incrRate = Mfds.gutenbergRichterRate(gr.a(), gr.b(), mMin);
    return Mfds.momentRate(gr.toBuilder()
        .scaleToIncrementalRate(incrRate)
        .build());
  }
}
