package gov.usgs.earthquake.nshmp.model;

import static com.google.common.base.Preconditions.checkArgument;
import static com.google.common.base.Preconditions.checkNotNull;
import static com.google.common.base.Preconditions.checkState;
import static gov.usgs.earthquake.nshmp.data.DoubleData.checkInRange;
import static gov.usgs.earthquake.nshmp.data.DoubleData.checkWeights;

import java.util.Map;
import java.util.Objects;
import java.util.Set;

import com.google.common.collect.Maps;
import com.google.common.collect.Range;

import gov.usgs.earthquake.nshmp.gmm.Gmm;
import gov.usgs.earthquake.nshmp.gmm.GroundMotionModel;
import gov.usgs.earthquake.nshmp.tree.Branch;
import gov.usgs.earthquake.nshmp.tree.LogicTree;

/**
 * Wrapper class for {@link GroundMotionModel}s associated with a
 * {@link RuptureSet}. In addition to carrying {@code Map}s of
 * {@code GroundMotionModel}s and their associated weights, this class
 * encapsulates any additional epistemic uncertainty that should be applied to
 * the ground motions computed from these models, as well as model weight
 * variants associated with different distance scales.
 *
 * <p>A {@code GmmSet} can not be created directly; it may only be created by a
 * private parser. Currently, although {@code hashCode} and {@code equals} are
 * consistent, they do not consider any supplied uncertainty model.
 *
 * @author U.S. Geological Survey
 */
public final class GmmSet {

  /*
   * There is a lot of null checking and handling in this class that would be
   * better served by Optional
   */

  private final Map<Gmm, Double> weightMapLo;
  private final double maxDistLo;
  private final Map<Gmm, Double> weightMapHi; // may be null
  private final double maxDistHi; // used by distance filters
  private final boolean singular;
  private final int hashCode;

  // private final UncertType uncertainty;
  // private final double[] epiWeights;
  //
  // private final UncertaintyModel epiModel;

  GmmSet(Builder builder) {

    this.weightMapLo = builder.gmmWtMapLo;
    this.maxDistLo = builder.maxDistanceLo;
    this.weightMapHi = builder.gmmWtMapHi;
    this.maxDistHi = (this.weightMapHi != null)
        ? builder.maxDistanceHi
        : maxDistLo;
    this.singular = this.weightMapHi == null;

    this.hashCode = Objects.hash(
        this.weightMapLo, this.maxDistLo,
        this.weightMapHi, this.maxDistHi);

    // this.uncertainty = (builder.epiModel == null)
    // ? UncertType.NONE
    // : UncertType.MULTI;

    // this.epiWeights = builder.uncWeights;

    // if (uncertainty == UncertType.NONE) {
    // this.epiValue = Double.NaN;
    // this.epiValues = null;
    // } else {
    // this.epiValue = Double.NaN;
    // this.epiValues = initEpiValues(builder.uncValues);
    // }

    // this.epiModel = builder.epiModel;

    // System.out.println(epiModel);
    // System.out.println(epiValues);
    // System.out.println(epiWeights);
    // System.out.println(uncertainty);
  }

  /**
   * The {@code Set} of {@link GroundMotionModel} identifiers.
   */
  public Set<Gmm> gmms() {
    return weightMapLo.keySet();
  }

  /**
   * The {@code Map} of {@link GroundMotionModel} identifiers and associated
   * weights to use at a given {@code distance} from a {@code Site}.
   * @param distance
   */
  public Map<Gmm, Double> gmmWeightMap(double distance) {
    /*
     * Note that below (maxDistance) is used by all distance filters, so here,
     * as long as distance isn't < maxDistLo, weightMapHi will be used.
     */
    return singular ? weightMapLo : distance <= maxDistLo ? weightMapLo : weightMapHi;
  }

  /**
   * The maximum distance for which the contained {@link GroundMotionModel}s are
   * applicable.
   */
  public double maxDistance() {
    return maxDistHi;
  }

  @Override
  public int hashCode() {
    return hashCode;
  }

  @Override
  public boolean equals(Object obj) {
    if (this == obj) {
      return true;
    }
    if (!(obj instanceof GmmSet)) {
      return false;
    }
    GmmSet that = (GmmSet) obj;
    return Objects.equals(this.weightMapLo, that.weightMapLo) &&
        this.maxDistLo == that.maxDistLo &&
        Objects.equals(this.weightMapHi, that.weightMapHi) &&
        this.maxDistHi == that.maxDistHi;
  }

  /* Single use builder */
  static class Builder {

    static final String ID = "GmmSet.Builder";
    boolean built = false;

    private static final Range<Double> MAX_DIST_RANGE = Range.closed(50.0, 1000.0);

    private Map<Gmm, Double> gmmWtMapLo;
    private Map<Gmm, Double> gmmWtMapHi;
    private Double maxDistanceLo;
    private double maxDistanceHi;

    // private UncertaintyModel epiModel;
    // private double[] uncWeights;

    // leave maxDistanceHi as primitive unless validation required
    // at some later date; GmmSet throws NPE if Double useds

    Builder primaryModels(LogicTree<Gmm> tree, GmmConfig config) {
      checkNotNull(tree);
      checkNotNull(config);

      // setBuilder.uncertainty(uncValues, uncWeights);
      gmmWtMapLo = tree.stream()
          .collect(Maps.toImmutableEnumMap(
              Branch::value,
              Branch::weight));

      maxDistanceLo = config.maxDistance;

      return this;
    }

    Builder secondaryModels(Map<Gmm, Double> gmmWtMap, GmmConfig config) {
      checkArgument(checkNotNull(gmmWtMap, "Map is null").size() > 0, "Map is empty");
      checkWeights(gmmWtMap.values());
      gmmWtMapHi = Maps.immutableEnumMap(gmmWtMap);

      secondaryMaxDistance(config.maxDistance);

      return this;
    }

    Builder secondaryMaxDistance(double maxDistance) {
      maxDistanceHi = checkInRange(MAX_DIST_RANGE, "Max distance", maxDistance);
      return this;
    }

    void validateState(String id) {

      // at a minimum the '*Lo' fields must be set
      checkState(!built, "This %s instance as already been used", id);
      checkState(gmmWtMapLo != null, "%s primary weight map not set", id);
      checkState(maxDistanceLo != null, "%s primary max distance not set", id);

      if (gmmWtMapHi != null) {
        // hi gmms must be same as or subset of lo gmms
        checkState(gmmWtMapLo.keySet().containsAll(gmmWtMapHi.keySet()),
            "%s secondary models must be a subset of primary models", id);
        // maxDistanceHi must also be set and greater than maxDistanceLo
        checkNotNull(maxDistanceHi,
            "%s secondary distance must be set for secondary models", id);
        checkState(maxDistanceHi > maxDistanceLo,
            "%s secondary distance [%s] ≤ primary distance [%s]", id, maxDistanceHi,
            maxDistanceLo);
      }

      // if (epiModel != null) {
      // checkNotNull(uncWeights, "%s uncertainty weights not set", id);
      // }
      // if (uncWeights != null) {
      // checkNotNull(epiModel, "%s uncertainty values not set", id);
      // }

      built = true;
    }

    GmmSet build() {

      validateState(ID);
      try {
        GmmSet gmmSet = new GmmSet(this);
        return gmmSet;
      } catch (Exception e) {
        e.printStackTrace();
        return null;
      }
    }

  }

}
