package gov.usgs.earthquake.nshmp.model;

import static com.google.common.base.Preconditions.checkArgument;
import static com.google.common.base.Preconditions.checkNotNull;
import static com.google.common.base.Preconditions.checkState;

import java.nio.file.Path;
import java.util.Map;
import java.util.Optional;

import gov.usgs.earthquake.nshmp.mfd.Mfd;
import gov.usgs.earthquake.nshmp.tree.LogicTree;

/**
 * Stateful container of model data.
 *
 * This class is NOT thread safe. Use of copyOf() when parsing a model permits
 * overriding settings in nested model directories.
 *
 * @author U.S. Geological Survey
 */
class ModelData {

  private TectonicSetting setting;
  private GmmSet gmms;
  private SourceConfig sourceConfig;
  private boolean clusterModel = false;
  private boolean faultSourceTree = false;

  private Optional<MfdConfig> mfdConfig = Optional.empty();
  private Optional<Map<String, LogicTree<Mfd.Properties>>> mfdMap = Optional.empty();

  private Optional<LogicTree<Double>> rateTree = Optional.empty();
  private Optional<LogicTree<String>> faultRateTree = Optional.empty();

  private Optional<Map<Integer, SourceFeature.NshmFault>> faultFeatureMap = Optional.empty();
  private Optional<Map<Integer, SourceFeature.Interface>> interfaceFeatureMap = Optional.empty();
  private Optional<Map<Integer, SourceFeature.Grid>> gridFeatureMap = Optional.empty();

  private Optional<Path> gridDataDir = Optional.empty();
  private Optional<Path> gridRateFile = Optional.empty();

  private ModelData() {}

  /* Initial instances require GMMs and source config. */
  static ModelData create(
      TectonicSetting setting,
      GmmSet gmms,
      SourceConfig sourceConfig) {

    ModelData data = new ModelData();
    data.tectonicSetting(setting);
    data.gmms(gmms);
    data.sourceConfig(sourceConfig);
    return data;
  }

  static ModelData copyOf(ModelData original) {
    ModelData copy = new ModelData();
    copy.setting = original.setting;
    copy.gmms = original.gmms;
    copy.sourceConfig = original.sourceConfig;
    copy.clusterModel = original.clusterModel;
    copy.faultSourceTree = original.faultSourceTree;
    copy.mfdConfig = original.mfdConfig;
    copy.mfdMap = original.mfdMap;
    copy.rateTree = original.rateTree;
    copy.faultRateTree = original.faultRateTree;
    copy.gridDataDir = original.gridDataDir;
    copy.gridRateFile = original.gridRateFile;
    copy.faultFeatureMap = original.faultFeatureMap;
    copy.interfaceFeatureMap = original.interfaceFeatureMap;
    copy.gridFeatureMap = original.gridFeatureMap;
    return copy;
  }

  void tectonicSetting(TectonicSetting setting) {
    this.setting = checkNotNull(setting);
  }

  TectonicSetting tectonicSetting() {
    return setting;
  }

  void gmms(GmmSet gmms) {
    this.gmms = checkNotNull(gmms);
  }

  GmmSet gmms() {
    return gmms;
  }

  void sourceConfig(SourceConfig sourceConfig) {
    this.sourceConfig = checkNotNull(sourceConfig);
  }

  SourceConfig sourceConfig() {
    return sourceConfig;
  }

  void clusterModel() {
    this.clusterModel = true;
  }

  boolean isClusterModel() {
    return clusterModel;
  }

  void faultSourceTree() {
    this.faultSourceTree = true;
  }

  boolean isFaultSourceTree() {
    return faultSourceTree;
  }

  void mfdConfig(MfdConfig mfdConfig) {
    this.mfdConfig = Optional.of(checkNotNull(mfdConfig));
  }

  Optional<MfdConfig> mfdConfig() {
    return mfdConfig;
  }

  void mfdMap(Map<String, LogicTree<Mfd.Properties>> mfdMap) {
    this.mfdMap = Optional.of(checkNotNull(mfdMap));
  }

  Optional<Map<String, LogicTree<Mfd.Properties>>> mfdMap() {
    return mfdMap;
  }

  void gridDataDirectory(Path gridDataDir) {
    checkState(this.gridDataDir.isEmpty());
    this.gridDataDir = Optional.of(checkNotNull(gridDataDir));
  }

  Optional<Path> gridDataDirectory() {
    return gridDataDir;
  }

  /* Resolves against grid data dir. */
  void gridRateFile(Path gridRateFile) {
    checkNotNull(gridRateFile);
    checkArgument(gridRateFile.getNameCount() == 1);
    this.gridRateFile = Optional.of(gridDataDir.orElseThrow().resolve(gridRateFile));
  }

  Optional<Path> gridRateFile() {
    return gridRateFile;
  }

  void rateTree(LogicTree<Double> rateTree) {
    this.rateTree = Optional.of(checkNotNull(rateTree));
  }

  Optional<LogicTree<Double>> rateTree() {
    return rateTree;
  }

  void faultRateTree(LogicTree<String> faultRateTree) {
    this.faultRateTree = Optional.of(checkNotNull(faultRateTree));
  }

  Optional<LogicTree<String>> faultRateTree() {
    return faultRateTree;
  }

  void faultFeatureMap(Map<Integer, SourceFeature.NshmFault> faultFeatureMap) {
    checkNotNull(faultFeatureMap);
    checkArgument(faultFeatureMap.size() > 0);
    this.faultFeatureMap = Optional.of(faultFeatureMap);
  }

  Optional<Map<Integer, SourceFeature.NshmFault>> faultFeatureMap() {
    return faultFeatureMap;
  }

  void interfaceFeatureMap(Map<Integer, SourceFeature.Interface> interfaceFeatureMap) {
    checkNotNull(interfaceFeatureMap);
    checkArgument(interfaceFeatureMap.size() > 0);
    this.interfaceFeatureMap = Optional.of(interfaceFeatureMap);
  }

  Optional<Map<Integer, SourceFeature.Interface>> interfaceFeatureMap() {
    return interfaceFeatureMap;
  }

  void gridFeatureMap(Map<Integer, SourceFeature.Grid> gridFeatureMap) {
    checkNotNull(gridFeatureMap);
    checkArgument(gridFeatureMap.size() > 0);
    this.gridFeatureMap = Optional.of(gridFeatureMap);
  }

  Optional<Map<Integer, SourceFeature.Grid>> gridFeatureMap() {
    return gridFeatureMap;
  }

}
