package gov.usgs.earthquake.nshmp.model;

import static java.util.stream.Collectors.groupingBy;
import static java.util.stream.Collectors.mapping;
import static java.util.stream.Collectors.toList;

import java.nio.file.Path;
import java.util.ArrayList;
import java.util.Collection;
import java.util.List;
import java.util.Map.Entry;
import java.util.NoSuchElementException;
import java.util.stream.Collectors;

import com.google.gson.Gson;
import com.google.gson.GsonBuilder;

import gov.usgs.earthquake.nshmp.mfd.Mfd;
import gov.usgs.earthquake.nshmp.model.SourceTree.Leaf;
import gov.usgs.earthquake.nshmp.tree.Branch;

/**
 * Factory class for querying source models.
 *
 * @author U.S. Geological Survey
 */
public class Models {

  static final Gson GSON = new GsonBuilder()
      .setPrettyPrinting()
      .create();

  public static void main(String[] args) {
    HazardModel model = ModelLoader.load(Path.of("../nshm-conus"));
    // HazardModel model = ModelLoader.load(Path.of("../nshm-hawaii"));

    // JSON of all tree info for a model
    System.out.println(GSON.toJson(trees(model)));
    System.out.println();

    // Cascadia interface tree
    // SourceTree tree = model.tree(3199).orElseThrow();

    // see what the list of nodes for each branch looks like
    // we don't use this but it's good for illustrating
    // internal structure of a SourceTree
    // tree.nodes().stream().forEach(System.out::println);
    // System.out.println();

    // SourceTree tree = model.tree(8099).orElseThrow();
    // tree.nodes().stream().forEach(System.out::println);
    // System.out.println();
    // System.out.println(tree);
    System.out.println();
    // tree.stream().forEach(System.out::println);

    // System.out.println(GSON.toJson(tree(model, 8099))); // ceus grid
    // System.out.println(GSON.toJson(tree(model, 8230))); // intraslab
    // System.out.println(GSON.toJson(tree(model, 3199))); // Cascadia
    //
    // System.out.println(GSON.toJson(tree(model, 3209))); // Meers
    // System.out.println(GSON.toJson(tree(model, 3709))); // Charleston
    // System.out.println(GSON.toJson(tree(model, 3600))); // Commerce Lineament
    // System.out.println(GSON.toJson(tree(model, 3809))); // E. Rift Margin S
    // System.out.println(GSON.toJson(tree(model, 3099))); // New Madrid

    // System.out.println(GSON.toJson(tree(model, 2539))); // Wasatch
    // System.out.println(GSON.toJson(tree(model, 8309))); // UCERF3
    // System.out.println(GSON.toJson(tree(model, 8129))); // UCERF3 grid

    // System.out.println(GSON.toJson(tree(model, 2591))); // Stonewall Ant.
    System.out.println(GSON.toJson(tree(model, 2535))); // Steens

  }

  /**
   * Returns an object for JSON serialization with the name and ID of all source
   * logic tree groups in the supplied model organized by tectonic setting and
   * source type.
   *
   * @param model to extract logic tree data from
   */
  public static Object trees(HazardModel model) {
    return model.trees().asMap().entrySet().stream()
        .map(Models::toSettingGroup)
        .collect(toList());
  }

  /**
   * Returns an object for JSON serialization the details of a logic tree in the
   * supplied model.
   *
   * @param model to query
   * @param id of desired source tree
   * @throws NoSuchElementException if the id is not present in the model
   */
  public static Object tree(HazardModel model, int id) {
    return model.tree(id).map(t -> new Tree(t, model.root())).orElseThrow();
  }

  private static SettingGroup toSettingGroup(Entry<TectonicSetting, Collection<SourceTree>> entry) {
    return new SettingGroup(
        entry.getKey(),
        toSourceGroup(entry.getValue()));
  }

  private static List<SourceGroup> toSourceGroup(Collection<SourceTree> trees) {
    return trees.stream()
        .collect(groupingBy(
            SourceTree::type,
            mapping(
                TreeInfo::new,
                Collectors.toList())))
        .entrySet().stream()
        .map(e -> new SourceGroup(e.getKey(), e.getValue()))
        .collect(toList());
  }

  static final class SettingGroup {

    final TectonicSetting setting;
    final List<SourceGroup> data;

    SettingGroup(TectonicSetting setting, List<SourceGroup> data) {
      this.setting = setting;
      this.data = data;
    }
  }

  static final class SourceGroup {

    final SourceType type;
    final List<TreeInfo> data;

    SourceGroup(SourceType type, List<TreeInfo> data) {
      this.type = type;
      this.data = data;
    }
  }

  static final class TreeInfo {

    final int id;
    final String name;

    TreeInfo(SourceTree tree) {
      this.id = tree.id();
      this.name = tree.name();
    }
  }

  static final class Tree {

    final int id;
    final String name;
    final String path;
    final TectonicSetting setting;
    final SourceType type;
    final List<SourceBranch> branches;

    Tree(SourceTree tree, Path modelRoot) {
      this.id = tree.id();
      this.name = tree.name();
      this.path = modelRoot.relativize(tree.path()).toString();
      this.setting = tree.setting();
      this.type = tree.type();

      List<SourceBranch> branches = new ArrayList<>();

      for (Entry<Leaf, List<Branch<Path>>> e : tree.branches().entrySet()) {
        List<Branch<Path>> pathList = e.getValue();
        String branchWts = weightsString(pathList);
        Branch<Path> leafBranch = pathList.get(pathList.size() - 1);
        // System.out.println(pathList.size());
        // System.out.println(leafBranch.value() + " " +
        // leafBranch.value().getNameCount());

        Path branchPath = leafBranch.value().getNameCount() == 1
            ? leafBranch.value()
            : tree.path().relativize(leafBranch.value());

        SourceBranch branch = new SourceBranch(
            branchWts,
            branchPath.toString(),
            leafBranch.id(),
            tree.weights().get(e.getKey()),
            e.getKey().ruptures);
        branches.add(branch);
      }
      this.branches = branches;
    }
  }

  private static String weightsString(List<Branch<Path>> branches) {
    return branches.stream()
        .map(b -> b.id() + "[" + b.weight() + "]")
        .collect(Collectors.joining(":"));
  }

  static final class SourceBranch {

    final String branch;
    final String path;
    final String name;
    final double weight;
    final List<MfdBranch> mfds;

    SourceBranch(
        String branch,
        String path,
        String name,
        double weight,
        RuptureSet<? extends Source> ruptureSet) {

      this.branch = branch;
      this.path = path;
      this.name = name;
      this.weight = weight;
      this.mfds = ruptureSet.mfdTree().stream()
          .map(MfdBranch::new)
          .collect(toList());
    }
  }

  static final class MfdBranch {
    final String id; // TODO change to name
    final double weight;
    final Mfd mfd;

    MfdBranch(Branch<Mfd> branch) {
      this.id = branch.id();
      this.weight = branch.weight();
      this.mfd = branch.value();
    }
  }
}
