package gov.usgs.earthquake.nshmp.model;

import static gov.usgs.earthquake.nshmp.Faults.checkRake;

import java.util.Map;

import gov.usgs.earthquake.nshmp.fault.surface.RuptureSurface;
import gov.usgs.earthquake.nshmp.geo.Location;

/**
 * A {@code Rupture} is a proxy for an actual earthquake and encapsulates all
 * the source information required by a ground motion model (Gmm).
 *
 * @author U.S. Geological Survey
 */
public class Rupture {

  double mag;
  double rake;
  double rate;
  Location hypocenter; // needed?
  RuptureSurface surface;

  /* for internal use only */
  Rupture() {}

  private Rupture(double mag, double rate, double rake, RuptureSurface surface,
      Location hypocenter) {
    this.mag = mag;
    this.rate = rate;
    this.rake = checkRake(rake);
    this.surface = surface;
    this.hypocenter = hypocenter;
  }

  /**
   *
   * @param mag moment magnitude
   * @param rate of occurrence (annual)
   * @param rake slip direction on rupture surface
   * @param surface of the rupture
   * @return a new {@code Rupture}
   */
  public static Rupture create(double mag, double rate, double rake, RuptureSurface surface) {
    return new Rupture(mag, rate, rake, surface, null);
  }

  /**
   * The {@code Rupture} magnitude.
   */
  public double mag() {
    return mag;
  }

  /**
   * The {@code Rupture} rake.
   */
  public double rake() {
    return rake;
  }

  /**
   * The {@code Rupture} surface.
   */
  public RuptureSurface surface() {
    return surface;
  }

  /**
   * The {@code Rupture} rate.
   */
  public double rate() {
    return rate;
  }

  @Override
  public String toString() {
    Map<String, ?> data = Map.of(
        "mag", mag,
        "rake", rake,
        "rate", rate,
        "hypo", hypocenter == null ? "null" : hypocenter,
        "surface", surface);
    return getClass().getSimpleName() + " " + data;
  }

}
