package gov.usgs.earthquake.nshmp.model;

import java.util.function.Predicate;

import gov.usgs.earthquake.nshmp.geo.Location;
import gov.usgs.earthquake.nshmp.gmm.GroundMotionModel;
import gov.usgs.earthquake.nshmp.mfd.Mfd;
import gov.usgs.earthquake.nshmp.tree.LogicTree;

/**
 * A rupture set; usually some physical or pseudo-representation of a fault and
 * associated {@link Mfd}s governing the size and rate of all possible
 * {@link Rupture}s.
 *
 * @author U.S. Geological Survey
 */
public interface RuptureSet<T extends Source> extends Iterable<T> {

  /**
   * The display name of the rupture set.
   */
  String name();

  /**
   * The numeric ID of this rupture set.
   */
  int id();

  /**
   * The {@code TectonicSetting} identifier.
   */
  TectonicSetting setting();

  /**
   * The {@code SourceType} identifier.
   */
  SourceType type();

  /**
   * Return the {@link GroundMotionModel}s associated with this
   * {@code RuptureSet} as a {@link GmmSet}.
   *
   * @see GmmSet
   */
  GmmSet groundMotionModels();

  /**
   * The number of {@link Rupture}s in this rupture set.
   */
  int size();

  /**
   * The weight of the {@link SourceTree} branch to which this rupture set is
   * attached.
   */
  double weight();

  /**
   * The logic tree of MFDs in this rupture set.
   */
  LogicTree<Mfd> mfdTree();

  /**
   * The {@code Location} of this source relative to the supplied {@code site}
   * location. The details of what this method returns are implementation
   * secific.
   */
  Location location(Location site); // TODO move to utility ??

  /**
   * Return an {@code Iterable} over those {@code Source}s that are within the
   * cutoff distance of the supplied {@code Location}. The cutoff distance is
   * derived from the {@code GroundMotionModel}s associated with this
   * {@code RuptureSet}.
   *
   * @param loc {@code Location} of interest
   */
  Iterable<T> iterableForLocation(Location loc);

  /**
   * Return an {@code Iterable} over those {@code Source}s that are within
   * {@code distance} of the supplied {@code Location}.
   *
   * @param loc {@code Location} of interest
   * @param distance limit
   */
  Iterable<T> iterableForLocation(Location loc, double distance);

  /**
   * Return a {@link Predicate} that evaluates to {@code true} if this source is
   * within {@code distance} of the supplied {@link Location}. This
   * {@code Predicate} performs a quick distance calculation and is used to
   * determine whether this source should be included in a hazard calculation.
   *
   * @param loc {@code Location} of interest
   * @param distance limit
   */
  Predicate<T> distanceFilter(Location loc, double distance);

  /**
   * Return the total
   * @return
   */
  // Mfd mfd();

  // instead of Source
  // interface .Model

  // where possible get id and name from feature?
}
