package gov.usgs.earthquake.nshmp.model;

import static com.google.common.base.Preconditions.checkArgument;
import static com.google.common.base.Preconditions.checkNotNull;
import static gov.usgs.earthquake.nshmp.Earthquakes.checkCrustalDepth;
import static gov.usgs.earthquake.nshmp.Faults.checkAseismicityFactor;
import static gov.usgs.earthquake.nshmp.Faults.checkDip;
import static gov.usgs.earthquake.nshmp.Faults.checkDipDirection;
import static gov.usgs.earthquake.nshmp.Faults.checkRake;
import static gov.usgs.earthquake.nshmp.Faults.checkSlipRate;
import static gov.usgs.earthquake.nshmp.Faults.checkTrace;
import static gov.usgs.earthquake.nshmp.Text.checkName;
import static java.util.stream.Collectors.toUnmodifiableMap;

import java.awt.geom.Area;
import java.lang.reflect.Type;
import java.net.URL;
import java.nio.file.Path;
import java.util.ArrayList;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.NoSuchElementException;
import java.util.Optional;
import java.util.OptionalDouble;

import com.google.common.reflect.TypeToken;
import com.google.gson.JsonElement;

import gov.usgs.earthquake.nshmp.Earthquakes;
import gov.usgs.earthquake.nshmp.Faults;
import gov.usgs.earthquake.nshmp.geo.Location;
import gov.usgs.earthquake.nshmp.geo.LocationList;
import gov.usgs.earthquake.nshmp.geo.Locations;
import gov.usgs.earthquake.nshmp.geo.json.Feature;
import gov.usgs.earthquake.nshmp.geo.json.FeatureCollection;
import gov.usgs.earthquake.nshmp.geo.json.GeoJson;
import gov.usgs.earthquake.nshmp.geo.json.Properties;

/**
 * An earthquake source feature.
 *
 * <p>This class provides access to the <a href="https://geojson.org">
 * GeoJSON</a> features that define earthquake source geomtries in a NSHM. A
 * feature may define one or more fault traces using a GeoJSON
 * {@code LineString} or {@code MultiLineString} geometry, or a closed area
 * using a {@code Polygon}, along with various {@code properties} specific to
 * each source type.
 *
 * @author U.S. Geological Survey
 */
public abstract class SourceFeature {

  /** The feature numeric id. */
  public final int id;

  /** The feature name. */
  public final String name;

  /** The country ISO 3166-1 code (alpha-2) of this feature. */
  public final String country;

  /**
   * The primary state, province, or territory ISO 3166-2 code in which this
   * feature is located. If a feature spans multiple states, they are listed in
   * the {@code states} array. Most features require a state identifier, but it
   * may be absent for features spanning very large areas (e.g. grid sources).
   */
  public final Optional<String> state;

  /**
   * An immutable list of all states in which the feature is located. If
   * {@code states().size() > 1}, the states are generally ordered relative to
   * decreasing percentage of the feature (trace or area) in each state.
   */
  public final List<String> states;

  /** The GeoJSON feature from which this source feature was initialized. */
  public final Feature source;

  private SourceFeature(Feature feature) {
    id = feature.idAsInt().orElseThrow();
    checkArgument(id > 0, "Feature ID [%s] < 1", id);
    Properties props = feature.properties();
    name = checkName(props.getString(Key.NAME).orElseThrow(), "Feature");
    country = props.getString(Key.COUNTRY).orElse("US");
    state = props.getString(Key.STATE);
    states = List.of(props.get(Key.STATES, String[].class).orElse(
        state.isPresent() ? new String[] { state.orElseThrow() } : new String[0]));
    checkArgument(
        state.isEmpty() || states.get(0).equals(state.orElseThrow()),
        "Inconsistent 'state' and 'states' fields");
    this.source = feature;
  }

  @Override
  public String toString() {
    return String.format("Feature: %s %s %s", name, state.orElse(""), id);
  }

  /**
   * Create a new crustal fault feature from the supplied GeoJSON.
   *
   * @param geojson to parse
   * @throws NoSuchElementException if the GeoJSON feature {@code id}, or any of
   *         the required properties members, are missing, {@code null}, or
   *         unparseable as their expected value type.
   * @throws IllegalArgumentException if the GeoJSON geometry type is not
   *         {@code LineString}, the feature {@code id} is negative, or any
   *         properties are empty or out of range.
   */
  public static SourceFeature.Fault newFaultSection(Path geojson) {
    return geojsonToSection(GeoJson.from(geojson));
  }

  /**
   * Create a new crustal fault feature from the supplied GeoJSON.
   *
   * @param geojson to parse
   * @throws NoSuchElementException if the GeoJSON feature {@code id}, or any of
   *         the required properties members, are missing, {@code null}, or
   *         unparseable as their expected value type.
   * @throws IllegalArgumentException if the GeoJSON geometry type is not
   *         {@code LineString}, the feature {@code id} is negative, or any
   *         properties are empty or out of range.
   */
  public static SourceFeature.Fault newFaultSection(URL geojson) {
    return geojsonToSection(GeoJson.from(geojson));
  }

  private static SourceFeature.Fault geojsonToSection(GeoJson geojson) {
    /*
     * At present, a crustal fault may be represented as a Feature or
     * FeatureCollection. We try to read each section as a Feature and if
     * GeoJson.toFeature() throws an UnsupportedOperationException with a known
     * message we try to create a FeatureCollection. All other parsing
     * exceptions will be thrown as is.
     */
    try {
      return new Fault(geojson.toFeature(), Optional.empty());
    } catch (UnsupportedOperationException uoe) {
      if (uoe.getMessage().startsWith("GeoJSON is not a")) {
        FeatureCollection fc = geojson.toFeatureCollection();
        return new Fault(
            fc.features().get(0),
            Optional.of(fc.features().get(1).asPolygonBorder()));
      } else {
        throw uoe;
      }
    }
  }

  /**
   * Create a new subduction interface section from the supplied JSON.
   *
   * @param geojson to parse
   * @throws NoSuchElementException if the GeoJSON feature {@code id}, or any of
   *         the required properties members, are missing, {@code null}, or
   *         unparseable as their expected value type.
   * @throws IllegalArgumentException if the GeoJSON geometry type is not
   *         {@code MultiLineString}, the section traces are not ordered with
   *         increasing depth, the feature {@code id} is negative, or any
   *         properties are empty or out of range.
   */
  public static SourceFeature.Interface newInterfaceSection(Path geojson) {
    return new Interface(GeoJson.from(geojson).toFeature(), true);
  }

  /**
   * Create a new décollement section from the supplied JSON.
   *
   * @param geojson to parse
   * @throws NoSuchElementException if the GeoJSON feature {@code id}, or any of
   *         the required properties members, are missing, {@code null}, or
   *         unparseable as their expected value type.
   * @throws IllegalArgumentException if the GeoJSON geometry type is not
   *         {@code LineString}, the feature {@code id} is negative, or any
   *         properties are empty or out of range.
   */
  public static SourceFeature.Decollement newDecollementSection(Path geojson) {
    return new Decollement(GeoJson.from(geojson).toFeature());
  }

  /**
   * Create a new zone source polygon from the supplied JSON.
   *
   * @param geojson to parse
   * @throws NoSuchElementException if the GeoJSON feature {@code id}, or any of
   *         the required properties members, are missing, {@code null}, or
   *         unparseable as their expected value type.
   * @throws IllegalArgumentException if the GeoJSON geometry type is not
   *         {@code LineString}, the feature {@code id} is negative, or any
   *         properties are empty or out of range.
   */
  public static SourceFeature.Zone newSourceZone(Path geojson) {
    return new Zone(GeoJson.from(geojson).toFeature());
  }

  /**
   * Create a new grid source polygon from the supplied JSON.
   *
   * @param geojson to parse
   * @throws NoSuchElementException if the GeoJSON feature {@code id}, or any of
   *         the required properties members, are missing, {@code null}, or
   *         unparseable as their expected value type.
   * @throws IllegalArgumentException if the GeoJSON geometry type is not
   *         {@code LineString}, the feature {@code id} is negative, or any
   *         properties are empty or out of range.
   */
  public static SourceFeature.Grid newSourceGrid(Path geojson) {
    return new Grid(GeoJson.from(geojson).toFeature());
  }

  static SourceFeature.NshmFault newNshmFault(Path geojson) {
    return new NshmFault(GeoJson.from(geojson).toFeature());
  }

  static final class Key {
    static final String NAME = "name";
    static final String COUNTRY = "country";
    static final String STATE = "state";
    static final String STATES = "states";
    static final String DIP = "dip";
    static final String DIP_DIRECTION = "dip-direction";
    static final String RATE_DIP = "rate-dip";
    static final String M = "m";
    static final String MAGNITUDE = "magnitude";
    static final String MAGNITUDE_NOTE = "magnitude-note";
    static final String LENGTH = "length";
    static final String WIDTH = "width";
    static final String DEPTH = "depth";
    static final String UPPER_DEPTH = "upper-depth";
    static final String LOWER_DEPTH = "lower-depth";
    static final String RAKE = "rake";
    static final String RATE = "rate";
    static final String RATE_MAP = "rate-map";
    static final String RATE_TYPE = "rate-type";
    static final String GR_A_VALUE = "gr-a-value";
    static final String SINGLE_RATE = "single-rate";
    static final String STRIKE = "strike";
    static final String MFD_TREE = "mfd-tree";
    static final String ASEISMICITY = "aseismicity";
    static final String INDEX = "index";
    static final String INDICES = "indices";
    static final String PARENT_ID = "parent-id";
  }

  /** A section of a crustal fault. */
  public static final class Fault extends SourceFeature {

    /** The fault section trace. */
    public final LocationList trace;

    /** The fault section dip. */
    public final double dip;

    /** The fault section upper depth. */
    public final double upperDepth;

    /** The fault section lower depth. */
    public final double lowerDepth;

    /** The fault section rake. */
    public final double rake;

    /** The fault section rate. */
    public final double rate;

    /** The fault section {@code RateType}. */
    public final RateType rateType;

    /** The optional fault section zone polygon. */
    public final Optional<LocationList> zone;

    private Fault(Feature section, Optional<LocationList> zone) {
      super(section);
      trace = section.asLineString();
      checkTrace(trace);
      Properties props = section.properties();
      dip = checkDip(props.getDouble(Key.DIP).orElseThrow());
      upperDepth = checkCrustalDepth(props.getDouble(Key.UPPER_DEPTH).orElseThrow());
      lowerDepth = checkCrustalDepth(props.getDouble(Key.LOWER_DEPTH).orElseThrow());
      rake = checkRake(props.getDouble(Key.RAKE).orElseThrow());
      rateType = RateType.valueOf(props.getString(Key.RATE_TYPE).orElseThrow());
      rate = (rateType != RateType.RECURRENCE)
          ? checkSlipRate(props.getDouble(Key.RATE).orElseThrow())
          : 0.0;
      this.zone = zone;
      props.getString(Key.STATE).orElseThrow();
    }
  }

  /** A section of a crustal fault in a NSHM. */
  static final class NshmFault extends SourceFeature {

    /** The fault section trace. */
    final LocationList trace;

    /** The fault section dip. */
    final double dip;

    /** The fault section upper depth. */
    final double upperDepth;

    /** The fault section lower depth. */
    final double lowerDepth;

    /** The fault section rake. */
    final double rake;

    /** The fault section length. */
    final double length;

    /** The fault section width. */
    final double width;

    /** The fault section {@code RateType}. */
    final RateType rateType;

    /** The mapping of rates from different deformation models. */
    final Map<String, Rates> rateMap;

    /*
     * Fields for use in FaultRuptureSet.Builder. length, width, and magnitude
     * are be provided for easy Gmm parameterization reference; rateDip will
     * only be present for normal faults with a dip-tree.
     *
     * Note that these fields generally exist for singleton features and are not
     * expected or currently handled when stitching features together.
     */
    final OptionalDouble rateDip;
    final OptionalDouble magnitude;
    final Optional<MagnitudeNote> magnitudeNote;

    NshmFault(Feature section) {
      super(section);
      trace = section.asLineString();
      Properties props = section.properties();

      /* Required properties. */
      dip = props.getDouble(Key.DIP).orElseThrow();
      checkDip(dip);
      upperDepth = checkCrustalDepth(props.getDouble(Key.UPPER_DEPTH).orElseThrow());
      lowerDepth = checkCrustalDepth(props.getDouble(Key.LOWER_DEPTH).orElseThrow());
      rake = checkRake(props.getDouble(Key.RAKE).orElseThrow());
      rateType = RateType.valueOf(props.getString(Key.RATE_TYPE).orElseThrow());
      rateMap = nshmRateMap(props);
      props.getString(Key.STATE).orElseThrow();

      /* Properties computed if absent. */
      length = props.getDouble(Key.LENGTH).orElseGet(trace::length);
      width = props.getDouble(Key.WIDTH).orElseGet(this::computeWidth);

      /* Optional properties. */
      rateDip = props.getDouble(Key.RATE_DIP);
      magnitude = props.getDouble(Key.MAGNITUDE);
      magnitudeNote = props.get(Key.MAGNITUDE_NOTE, MagnitudeNote.class);
    }

    private double computeWidth() {
      return Faults.projectVertical(dip, lowerDepth - upperDepth);
    }

    /* Utility container for rate values. */
    static class Rates {

      /* Slip or recurrence rate */
      final double rate;

      /* Pre-computed Gutenberg-Richter MFD a-value. */
      final OptionalDouble graValue;

      /* Pre-computed single MFD rate. */
      final OptionalDouble singleRate;

      Rates(double rate) {
        this.rate = rate;
        this.graValue = OptionalDouble.empty();
        this.singleRate = OptionalDouble.empty();
      }

      Rates(JsonElement json) {
        Map<String, Double> rateMap = Deserialize.fromJson(json, RATE_DATA_TOKEN);
        this.rate = checkNotNull(rateMap.get(Key.RATE));
        Double graValue = rateMap.get(Key.GR_A_VALUE);
        this.graValue = (graValue != null)
            ? OptionalDouble.of(graValue)
            : OptionalDouble.empty();
        Double singleRate = rateMap.get(Key.SINGLE_RATE);
        this.singleRate = (singleRate != null)
            ? OptionalDouble.of(singleRate)
            : OptionalDouble.empty();
      }
    }

    private static Type RATE_MAP_TOKEN = new TypeToken<Map<String, JsonElement>>() {}.getType();
    private static Type RATE_DATA_TOKEN = new TypeToken<Map<String, Double>>() {}.getType();

    /*
     * A NSHM feature can have either single rate (maps to 'GEO'), no rate
     * because type is RECURRENCE (maps to 'GEO' with rate of 0), a rate-map
     * with multiple rates whose type can be SLIP or RECURRENCE.
     */
    private static Map<String, Rates> nshmRateMap(Properties props) {
      OptionalDouble rate = props.getDouble(Key.RATE);
      RateType type = props.get(Key.RATE_TYPE, RateType.class).orElseThrow();
      if (rate.isPresent()) {
        return Map.of("GEO", new Rates(rate.orElseThrow()));
      }

      Optional<Map<String, Rates>> rateMap = readRateMap(props);
      if (rateMap.isEmpty() && type == RateType.RECURRENCE) {
        return Map.of("GEO", new Rates(0.0));
      }
      return rateMap.orElseThrow();
    }

    private static Optional<Map<String, Rates>> readRateMap(Properties props) {
      Optional<Map<String, JsonElement>> rateElements = props.get(Key.RATE_MAP, RATE_MAP_TOKEN);
      if (rateElements.isEmpty()) {
        return Optional.empty();
      }
      Map<String, Rates> rateMap = rateElements.orElseThrow().entrySet().stream()
          .filter(e -> !e.getValue().isJsonNull())
          .collect(toUnmodifiableMap(
              Entry::getKey,
              e -> new Rates(e.getValue())));
      return Optional.of(rateMap);
    }
  }

  /** A sub-section of a crustal fault in a NSHM fault-system. */
  static final class SystemSection extends SourceFeature {

    /** Index of section referenced by rupture sets. */
    final int index;

    /** ID of the parent fault section. */
    final int parentId;

    /** The fault subsection trace. */
    final LocationList trace;

    /** The fault subsection dip. */
    final double dip;

    /** The fault subsection dip. */
    final double dipDirection;

    /** The fault subsection upper depth. */
    final double upperDepth;

    /** The fault subsection lower depth. */
    final double lowerDepth;

    /** The fault subsection aseismicity factor. */
    final double aseismicity;

    SystemSection(Feature subsection) {
      super(subsection);
      trace = subsection.asLineString();
      Properties props = subsection.properties();
      index = props.getInt(Key.INDEX).orElseThrow();
      parentId = props.getInt(Key.PARENT_ID).orElseThrow();
      dip = checkDip(props.getDouble(Key.DIP).orElseThrow());
      dipDirection = checkDipDirection(
          trace,
          props.getDouble(Key.DIP_DIRECTION).orElseThrow());
      upperDepth = checkCrustalDepth(props.getDouble(Key.UPPER_DEPTH).orElseThrow());
      lowerDepth = checkCrustalDepth(props.getDouble(Key.LOWER_DEPTH).orElseThrow());
      aseismicity = checkAseismicityFactor(props.getDouble(Key.ASEISMICITY).orElseThrow());
      props.getString(Key.STATE).orElseThrow();
    }
  }

  /** A section of a subduction interface. */
  public static final class Decollement extends SourceFeature {

    /**
     * An immutable list of traces that define the surface (edge to edge) of the
     * section. Unlike interface sections, the traces are not required to
     * increase in depth.
     */
    public final List<LocationList> traces;

    private Decollement(Feature section) {
      super(section);
      this.traces = section.asMultiLineString();
      section.properties().getString(Key.STATE).orElseThrow();
    }
  }

  /** A section of a subduction interface. */
  public static final class Interface extends SourceFeature {

    /**
     * An immutable list of traces that define the surface (top to bottom) of
     * the section.
     */
    public final List<LocationList> traces;

    Interface(Feature section, boolean checkState) {
      super(section);
      this.traces = section.asMultiLineString();
      checkTraces(traces);
      if (checkState) {
        section.properties().getString(Key.STATE).orElseThrow();
      }
    }
  }

  /** A source zone. */
  public static final class Zone extends SourceFeature {

    /** The source zone border. */
    public final LocationList border;

    /** The optional strike of sources in the zone. */
    public final OptionalDouble strike;

    private Zone(Feature feature) {
      super(feature);
      this.border = feature.asPolygonBorder();
      Properties props = feature.properties();
      this.strike = props.getDouble(Key.STRIKE);
      props.getJsonElement(Key.MFD_TREE).orElseThrow();
      props.getJsonElement(Key.RATE).orElseThrow();
      props.getString(Key.STATE).orElseThrow();
      // Consider requiring rateType and add RateType.TREE
      // if TREE, then rate = 0.0 or is optional
      // Consider allowing rate to be absent in
      // nshm-fault-sections if NONE as opposed to 0.0
    }
  }

  /** A gridded seismicity source. */
  public static final class Grid extends SourceFeature {

    /** The grid border. */
    public final LocationList border;

    /** The optional strike of sources in the grid. */
    public final OptionalDouble strike;

    private Area area;

    private Grid(Feature feature) {
      super(feature);
      this.border = feature.asPolygonBorder();
      Properties props = feature.properties();
      this.strike = props.getDouble(Key.STRIKE);

      // To be explicit, consider requiring rateType and add RateType.TREE
      // if TREE, then rate = 0.0 or is optional
      // Consider allowing rate to be absent in
      // nshm-fault-sections if NONE as opposed to 0.0

      // Is there any check upstream of the first and last points
      // in the border being the same?

      this.area = new Area(Locations.toPath(border));
    }

    /**
     * Test if a location is inside the grid border polygon.
     */
    public boolean contains(Location location) {
      return area.contains(location.longitude, location.latitude);
    }
  }

  private static void checkTraces(List<LocationList> traces) {
    List<Double> maxDepths = new ArrayList<>();
    /* Check that depths are valid and save max for each trace. */
    for (LocationList trace : traces) {
      double maxDepth = trace.stream()
          .mapToDouble(loc -> loc.depth)
          .map(Earthquakes::checkInterfaceDepth)
          .max()
          .orElseThrow();
      maxDepths.add(maxDepth);
    }
    /* Check that each trace after the first is deeper than the previous. */
    for (int i = 1; i < traces.size(); i++) {
      double minDepth = maxDepths.get(i - 1);
      traces.get(i).stream()
          .mapToDouble(loc -> loc.depth)
          .forEach(depth -> checkArgument(
              depth >= minDepth,
              "Invalid interface depths"));
    }
  }
}
