package gov.usgs.earthquake.nshmp.model;

import static com.google.common.base.Preconditions.checkArgument;
import static com.google.common.base.Preconditions.checkNotNull;
import static com.google.common.base.Preconditions.checkState;
import static gov.usgs.earthquake.nshmp.Text.NEWLINE;
import static gov.usgs.earthquake.nshmp.Text.checkName;
import static java.util.stream.Collectors.toList;
import static java.util.stream.Collectors.toUnmodifiableMap;

import java.nio.file.Path;
import java.util.AbstractList;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Optional;
import java.util.stream.Collectors;
import java.util.stream.StreamSupport;

import com.google.common.base.Strings;
import com.google.common.collect.Iterables;
import com.google.common.graph.EndpointPair;
import com.google.common.graph.Graphs;
import com.google.common.graph.MutableNetwork;
import com.google.common.graph.NetworkBuilder;
import com.google.common.graph.Traverser;

import gov.usgs.earthquake.nshmp.Maths;
import gov.usgs.earthquake.nshmp.tree.Branch;
import gov.usgs.earthquake.nshmp.tree.LogicTree;

/**
 * Top level wrapper class used for any source logic trees. For example, UCERF3
 * alternate fault models, New Madrid, Wasatch, or any zone or grid logic tree
 * based models.
 *
 * @author U.S. Geological Survey
 */
public class SourceTree extends
    AbstractList<Branch<RuptureSet<? extends Source>>> implements
    LogicTree<RuptureSet<? extends Source>> {

  /*
   * Developer notes:
   *
   * Root --> Node --> ... --> Node --> Leaf
   *
   * We use Guava's graph library to track source model logic trees. The
   * heirarchy of logic trees in a source model is mapped into a Network where
   * nodes are connected with unique edges that are the branches of the
   * component logic trees. The network is a directed tree and each node has
   * only one incident edge. Leaf weights are rounded to 8 decimal places.
   */

  private final Path path;
  private final String name;
  private final int id;
  private final TectonicSetting setting;
  private final SourceType type;
  private final GmmSet gmms;
  private final List<List<Node>> nodes; // for string representation
  private final Map<Leaf, List<Branch<Path>>> branches;
  private final Map<Leaf, Double> weights;
  private final List<Leaf> leaves;
  // private final Network<Node, Branch<Path>> tree;

  SourceTree(Builder builder) {
    this.path = builder.path;
    this.name = builder.name;
    this.id = builder.id;
    this.setting = builder.setting;
    this.type = builder.type;
    this.gmms = builder.gmms;
    this.nodes = builder.nodes;
    this.branches = builder.branches;
    this.weights = builder.leafWeights;
    this.leaves = builder.leaves;
    // this.tree = ImmutableNetwork.copyOf(builder.tree);
  }

  /** The model path to this tree. */
  public Path path() {
    return path;
  }

  @Override
  public String name() {
    return name;
  }

  /** The ID of this tree. */
  public int id() {
    return id;
  }

  /** The tectonic setting of this tree. */
  public TectonicSetting setting() {
    return setting;
  }

  /** The type of sources in this tree. */
  public SourceType type() {
    return type;
  }

  /** The ground motion models to use with this tree. */
  public GmmSet gmms() {
    return gmms;
  }

  /** The leaf nodes of this tree mapped to their branch paths. */
  public Map<Leaf, List<Branch<Path>>> branches() {
    return branches;
  }

  /** The leaf nodes of this tree mapped to the total branch weight. */
  public Map<Leaf, Double> weights() {
    return weights;
  }

  /**
   * The list of {@code root --> node --> ... --> node --> leaf} node mappings
   * that reflects the internal structure of this tree.
   */
  public List<String> nodes() {
    return nodes.stream()
        .map(Object::toString)
        .collect(toList());
  }

  @Override
  public Branch<RuptureSet<? extends Source>> sample(double probability) {
    throw new UnsupportedOperationException();
  }

  @Override
  public List<Branch<RuptureSet<? extends Source>>> sample(double[] probabilities) {
    throw new UnsupportedOperationException();
  }

  @Override
  public Branch<RuptureSet<? extends Source>> get(int index) {
    return new Branch<RuptureSet<? extends Source>>() {

      @Override
      public String id() {
        return leaves.get(index).toString();
      }

      @Override
      public RuptureSet<? extends Source> value() {
        return leaves.get(index).ruptures;
      }

      @Override
      public double weight() {
        return weights.get(leaves.get(index));
      }
    };
  }

  @Override
  public int size() {
    return leaves.size();
  }

  @Override
  public String toString() {
    return this.stream()
        .sorted((b1, b2) -> b1.value().name().compareTo(b2.value().name()))
        .map(branch -> "  ↳         " +
            branch.value() +
            Strings.padEnd(" weight: " + branch.weight(), 20, ' ') +
            branch.id())
        .collect(Collectors.joining(NEWLINE));
  }

  static Builder builder() {
    return new Builder();
  }

  private static final Root ROOT = new Root();

  static class Builder {

    private boolean built = false;

    private int index = 1;

    /* Required fields. */
    private Path path;
    private String name;
    private Integer id;
    private TectonicSetting setting;
    private SourceType type;
    private GmmSet gmms;

    private MutableNetwork<Node, Branch<Path>> tree;
    private List<Leaf> leafList = new ArrayList<>();

    /* Created on build. */
    private List<List<Node>> nodes;
    private Map<Leaf, List<Branch<Path>>> branches;
    private Map<Leaf, Double> leafWeights;
    private List<Leaf> leaves;

    private Builder() {}

    /* Set the model path to the root of the tree. */
    Builder path(Path path) {
      checkState(this.path == null, "Tree path already set");
      this.path = checkNotNull(path);
      return this;
    }

    /* Set the name of the tree. */
    Builder name(String name) {
      checkState(this.name == null, "Tree name already set");
      this.name = checkName(name, "Tree name");
      return this;
    }

    /* Set the ID of the tree. */
    Builder id(int id) {
      checkState(this.id == null, "Tree ID already set");
      this.id = id;
      return this;
    }

    /* Set the type of source in the tree. */
    Builder setting(TectonicSetting setting) {
      checkState(this.setting == null, "Tree tectonic setting already set");
      this.setting = checkNotNull(setting);
      return this;
    }

    /* Set the type of source in the tree. */
    Builder type(SourceType type) {
      checkState(this.type == null, "Tree source type already set");
      this.type = checkNotNull(type);
      return this;
    }

    /* Set the GMMs to be used with the sources in the tree. */
    Builder gmms(GmmSet gmms) {
      checkState(this.gmms == null, "Tree GMMs already set");
      this.gmms = checkNotNull(gmms);
      return this;
    }

    /* Set the logic tree that stems from the root node. */
    Builder root(LogicTree<Path> root) {
      checkState(this.tree == null, "Tree root node already set");
      this.tree = NetworkBuilder.directed()
          .allowsSelfLoops(false)
          .allowsParallelEdges(false)
          .build();
      for (Branch<Path> branch : root) {
        tree.addEdge(ROOT, new Node(index++), branch);
      }
      return this;
    }

    /* Attach the branches of a logic tree to the supplied branch. */
    Builder addBranches(Branch<Path> parent, LogicTree<Path> children) {
      checkState(this.tree != null, "Tree root node is not set");
      checkArgument(tree.edges().contains(parent), "Missing parent branch");
      EndpointPair<Node> endpoints = tree.incidentNodes(parent);
      // parent target is new source for each branch being added
      Node source = endpoints.target();
      for (Branch<Path> child : children) {
        tree.addEdge(source, new Node(index++), child);
      }
      return this;
    }

    /*
     * Add a leaf to the source tree. This method replaces the target Node
     * associated with the supplied branch with a terminal Leaf node.
     */
    Builder addLeaf(Branch<Path> parent, RuptureSet<? extends Source> ruptures) {
      checkState(this.tree != null, "Tree root node is not set");
      checkArgument(tree.edges().contains(parent), "Missing parent branch");
      EndpointPair<Node> endpoints = tree.incidentNodes(parent);
      Node target = endpoints.target();
      Node source = endpoints.source();
      tree.removeNode(target);
      Leaf leaf = new Leaf(target.index, ruptures); // transfer assigned index
      tree.addEdge(source, leaf, parent);
      leafList.add(leaf);
      return this;
    }

    /* Create a new source tree from this builder. */
    SourceTree build() {
      checkNotNull(path);
      checkNotNull(name);
      checkNotNull(id);
      checkNotNull(setting);
      checkNotNull(type);
      checkNotNull(gmms);
      checkState(tree.nodes().size() > 0, "Empty tree");
      checkState(leafList.size() > 0, "Leafless tree");
      buildTreeFields();
      checkState(!built, "Single use builder");
      built = true;
      return new SourceTree(this);
    }

    private void buildTreeFields() {

      /*
       * Generate lists of logic tree branches (branch paths) from the root of a
       * tree to each of the supplied leaf nodes. Because the tree is a directed
       * graph, traversing a transposed view starting with each leaf yields the
       * required node path.
       */
      Traverser<Node> traverser = Traverser.forTree(Graphs.transpose(tree));
      Map<Leaf, List<Branch<Path>>> branchListsMap = new HashMap<>();
      List<List<Node>> nodeLists = new ArrayList<>();

      for (Leaf leaf : leafList) {
        Iterable<Node> it = traverser.depthFirstPostOrder(leaf);
        List<Node> nodeList = StreamSupport.stream(it.spliterator(), false)
            .collect(Collectors.toList());
        checkState(nodeList.size() > 1); // 2 nodes minimum [root -> leaf]
        nodeLists.add(nodeList);

        List<Branch<Path>> branchList = new ArrayList<>();
        Node source = nodeList.get(0);
        for (Node target : Iterables.skip(nodeList, 1)) {
          branchList.add(tree.edgeConnecting(source, target).orElseThrow());
          source = target;
        }
        branchListsMap.put(leaf, List.copyOf(branchList));
      }
      nodes = List.copyOf(nodeLists);
      branches = Map.copyOf(branchListsMap);

      /*
       * Create map of leaves and their weights. Note that the values of the
       * returned map will not sum to one when the source tree contains one or
       * more LogicGroups.
       */
      leafWeights = branches.entrySet().stream().collect(
          toUnmodifiableMap(
              Entry::getKey,
              e -> leafWeight(e.getValue())));
      leaves = List.copyOf(branches.keySet());

      /*
       * Assign weights to RuptureSets. It would be nice if a RuptureSet didn't
       * know about it's source tree weight, but it's much easier in the current
       * calculation pipeline if they do.
       */
      leaves.stream().forEach(leaf -> {
        ((AbstractRuptureSet<?>) leaf.ruptures).setLeafWeight(leafWeights.get(leaf));
      });
    }

    /* Compute cumulative weight of a source branch from root to leaf. */
    private static double leafWeight(List<Branch<Path>> branchList) {
      double weight = branchList.stream()
          .mapToDouble(Branch::weight)
          .reduce(1, (a, b) -> a * b);
      return Maths.round(weight, 8);
    }
  }

  static class Node {

    final int index;

    Node(int index) {
      this.index = index;
    }

    @Override
    public String toString() {
      return "node:" + index;
    }
  }

  static final class Root extends Node {

    Root() {
      super(0);
    }

    @Override
    public String toString() {
      return "root:" + index;
    }
  }

  static final class Leaf extends Node {

    final RuptureSet<? extends Source> ruptures;

    Leaf(int index, RuptureSet<? extends Source> ruptures) {
      super(index);
      this.ruptures = ruptures;
    }

    @Override
    public String toString() {
      return "leaf:" + index;
    }
  }

  static final class Info {

    final int id;
    final Optional<String> name;

    Info(int id, Optional<String> name) {
      this.id = id;
      this.name = name;
    }
  }

}
