package gov.usgs.earthquake.nshmp.tree;

import static com.google.common.base.Preconditions.checkState;

import java.util.AbstractList;
import java.util.Iterator;
import java.util.List;

/**
 * A logic group is a collection of branches, the weights of which are not
 * required to sum to one.
 *
 * Logic groups are used to represent additive alternatives that would otherwise
 * be disallowed by a {@link LogicTree}. Although a {@code LogicGroup}
 * implements {@code LogicTree}, its sampling methods are overriden to throw an
 * {@code UnsupportedOperationException}. A logic group branch <i>may</i>
 * include a scale factor that is used in lieu of a branch weight; in most
 * real-world cases, the scale value is one. The scale factor may also be zero
 * to represent {@code null} branches.
 *
 * @param <T> the type of value stored in a branch
 * @author U.S. Geological Survey
 */
public final class LogicGroup<T> extends AbstractList<Branch<T>> implements LogicTree<T> {

  private static final String NAME = "logic-group";
  private final List<Branch<T>> branches;

  LogicGroup(List<Branch<T>> branches) {
    this.branches = List.copyOf(branches);
  }

  @Override
  public Branch<T> get(int index) {
    return branches.get(index);
  }

  /** Return the name of this group. */
  @Override
  public String name() {
    return NAME;
  }

  /** Return the number of branches in this group. */
  @Override
  public int size() {
    return branches.size();
  }

  @Override
  public String toString() {
    return LogicTree.toString(this);
  }

  @Override
  public Iterator<Branch<T>> iterator() {
    return branches.iterator();
  }

  /**
   * Not supported for branch groups.
   *
   * @throws UnsupportedOperationException
   */
  @Override
  public Branch<T> sample(double probability) {
    throw new UnsupportedOperationException();
  }

  /**
   * Not supported for branch groups.
   *
   * @throws UnsupportedOperationException
   */
  @Override
  public List<Branch<T>> sample(double[] probabilities) {
    throw new UnsupportedOperationException();
  }

  /**
   * Return a new, single-use {@code LogicGroup} builder.
   */
  public static <T> Builder<T> builder() {
    return new Builder<T>();
  }

  /**
   * A single-use {@code LogicGroup} builder.
   */
  public static class Builder<T> extends LogicTree.Builder<T> {

    private Builder() {
      super(NAME);
    }

    /**
     * Add a branch to the logic group.
     *
     * @param id of the branch
     * @param value of the branch
     * @param scale factor to apply in lieu of a branch weight, may be zero
     * @return this {@code Builder} object
     * @throws NullPointerException if {@code id} or {@code value} is null
     * @throws IllegalArgumentException if {@code id} is empty or whitespace
     */
    @Override
    public Builder<T> addBranch(String id, T value, double scale) {
      branches.add(new GroupBranch<>(id, value, scale));
      return this;
    }

    /**
     * Return a newly created {@code LogicGroup}.
     *
     * @throws IllegalStateException if this method has already been called
     * @throws IllegalStateException if the logic tree is empty (i.e. no
     *         branches have been added)
     */
    @Override
    public LogicGroup<T> build() {
      checkState(!built, "Single use builder");
      built = true;
      checkState(!branches.isEmpty(), "Empty logic group");
      return new LogicGroup<T>(List.copyOf(branches));
    }
  }
}
