package gov.usgs.earthquake.nshmp.tree;

import static com.google.common.base.Preconditions.checkArgument;
import static com.google.common.base.Preconditions.checkState;

import java.util.ArrayList;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

import gov.usgs.earthquake.nshmp.Text;
import gov.usgs.earthquake.nshmp.data.DoubleData;

/**
 * A logic tree.
 *
 * <p>Logic trees are implemented as immutable lists of branches, where each
 * branch has an ID, value and weight. Factory methods are provided that return
 * builders for specialized branch implementations.
 *
 * @param <T> the type of value stored in the branches of the tree
 * @author U.S. Geological Survey
 */
public interface LogicTree<T> extends List<Branch<T>> {

  /** Return the name of this tree. */
  String name();

  /**
   * Return a logic tree branch corresponding to the supplied probability.
   *
   * <p><b>Note:</b> this method does not check that {@code 0.0 ≤ P < 1.0}. For
   * {@code P < 0.0}, method returns the first branch of the tree; for
   * {@code P >= 1.0} method returns the last branch of the tree.
   *
   * @param probability in the range [0..1)
   */
  Branch<T> sample(double probability);

  /**
   * Return a list of logic tree branches corresponding to the supplied
   * probabilities.
   *
   * <p><b>Note:</b> this method does not check that each {@code 0.0 ≤ P < 1.0}.
   * For {@code P < 0.0}, method returns the first branch of the tree; for
   * {@code P >= 1.0} method returns the last branch of the tree.
   *
   * @param probabilities array with values in the range [0..1)
   */
  List<Branch<T>> sample(double[] probabilities);

  /**
   * String representation of a logic tree for use by concrete implementations.
   *
   * @param tree to render as a string
   */
  public static <T> String toString(LogicTree<T> tree) {
    StringBuilder sb = new StringBuilder(tree.name());
    for (int i = 0; i < tree.size(); i++) {
      switch (i) {
        case 0:
          sb.append((tree.size() == 1) ? " ─ " : " ┬ ");
          break;
        default:
          sb.append(" ".repeat(tree.name().length()));
          sb.append((i == tree.size() - 1) ? " └ " : " ├ ");
      }
      sb.append(tree.get(i));
      sb.append((i == tree.size() - 1) ? "" : Text.NEWLINE);
    }
    return sb.toString();
  }

  /**
   * A logic tree with a single branch.
   *
   * @param <T> the value type of each branch
   * @param name of the tree
   * @param id of the lone branch
   * @param value of the lone branch
   */
  static <T> LogicTree<T> singleton(String name, String id, T value) {
    return LogicTree.<T> builder(name)
        .addBranch(id, value, 1.0)
        .build();
  }

  /**
   * Return a new, single-use {@code LogicTree} builder.
   *
   * @param <T> the value type of each branch
   * @param name of the tree
   */
  static <T> Builder<T> builder(String name) {
    return new Builder<T>(name);
  }

  /**
   * Return a new, single-use {@code LogicTree} builder where each
   * {@link Branch#id()} of the built tree is an {@link Enum#name()} and each
   * {@link Branch#value()} is any value type.
   *
   * @param <E> the enum ID type of each branch
   * @param <T> the value type of each branch
   * @param name of the tree
   */
  static <E extends Enum<E>, T> EnumBuilder<E, T> enumBuilder(String name) {
    return new EnumBuilder<E, T>(name);
  }

  /**
   * Return a new, single-use {@code LogicTree} builder where each
   * {@link Branch#id()} of the built tree is an {@link Enum#name()} and each
   * {@link Branch#value()} is the enum value itself.
   *
   * @param <E> the enum value type of each branch
   * @param name of the tree
   */
  static <E extends Enum<E>> EnumValueBuilder<E> enumValueBuilder(String name) {
    return new EnumValueBuilder<E>(name);
  }

  /**
   * Return a new, single-use {@code LogicTree} builder where each
   * {@link Branch#id()} and {@link Branch#value()} of the built tree is the
   * same string.
   *
   * @param name of the tree
   */
  static StringValueBuilder stringValueBuilder(String name) {
    return new StringValueBuilder(name);
  }

  /** A single-use {@code LogicTree} builder. */
  static class Builder<T> {

    final String name;
    List<Branch<T>> branches = new ArrayList<>();
    Set<String> ids = new HashSet<>();
    double[] cumulativeWeights;
    boolean built = false;

    /**
     * Sole package constructor; use {@link LogicTree#builder(String)} instead.
     * @param name of the tree
     */
    protected Builder(String name) {
      this.name = Text.checkName(name, "Logic tree");
    }

    /**
     * Add a branch to the tree.
     *
     * @param id of the branch
     * @param value of the branch
     * @param weight of the branch
     * @return this {@code Builder} object
     * @throws NullPointerException if {@code id} or {@code value} is null
     * @throws IllegalArgumentException if {@code id} is empty or whitespace
     * @throws IllegalArgumentException if a branch with {@code id} has already
     *         been added
     * @throws IllegalArgumentException if {@code weight} is outside the range
     *         (0..1]
     */
    public Builder<T> addBranch(String id, T value, double weight) {
      branches.add(new RegularBranch<>(id, value, weight));
      checkIdExists(id);
      return this;
    }

    /**
     * Return a newly created {@code LogicTree}.
     *
     * @throws IllegalStateException if this method has already been called
     * @throws IllegalStateException if the logic tree is empty (i.e. no
     *         branches have been added)
     * @throws IllegalStateException if branch weights do not sum to
     *         {@code 1.0 ± 0.00001}
     */
    public LogicTree<T> build() {
      checkState(!built, "Single use builder");
      built = true;
      checkState(!branches.isEmpty(), "Empty logic tree");
      double[] weights = weightArray(branches);
      DoubleData.checkWeightSum(weights);
      cumulativeWeights = DoubleData.round(
          RegularTree.WEIGHT_SCALE,
          DoubleData.cumulate(weights));
      return new RegularTree<T>(this);
    }

    private static <T> double[] weightArray(List<Branch<T>> branches) {
      return branches.stream()
          .mapToDouble(Branch::weight)
          .toArray();
    }

    void checkIdExists(String id) {
      checkArgument(!ids.contains(id), "Duplicate branch [%s]", id);
      ids.add(id);
    }
  }

  /**
   * A single-use {@code LogicTree} builder where each {@link Branch#id()} of
   * the built tree is an {@link Enum#name()} and each {@link Branch#value()} is
   * any value type.
   */
  static class EnumBuilder<E extends Enum<E>, T> extends Builder<T> {

    private EnumBuilder(String name) {
      super(name);
    }

    /**
     * Add a branch to the tree.
     *
     * @param id of the branch
     * @param value of the branch
     * @param weight of the branch
     * @return this {@code Builder} object
     * @throws NullPointerException if {@code id} or {@code value} is null
     * @throws IllegalArgumentException if a branch with {@code id} has already
     *         been added
     * @throws IllegalArgumentException if {@code weight} is outside the range
     *         (0..1]
     */
    public EnumBuilder<E, T> addBranch(E id, T value, double weight) {
      branches.add(new EnumBranch<E, T>(id, value, weight));
      checkIdExists(id.name());
      return this;
    }
  }

  /**
   * A single-use {@code LogicTree} builder where each {@link Branch#id()} of
   * the built tree is an {@link Enum#name()} and each {@link Branch#value()} is
   * the enum value itself.
   */
  static class EnumValueBuilder<E extends Enum<E>> extends Builder<E> {

    private EnumValueBuilder(String name) {
      super(name);
    }

    /**
     * Add a branch to the tree.
     *
     * @param id of the branch
     * @param weight of the branch
     * @return this {@code Builder} object
     * @throws NullPointerException if {@code id} is null
     * @throws IllegalArgumentException if a branch with {@code id} has already
     *         been added
     * @throws IllegalArgumentException if {@code weight} is outside the range
     *         (0..1]
     */
    public EnumValueBuilder<E> addBranch(E id, double weight) {
      branches.add(new EnumValueBranch<E>(id, weight));
      checkIdExists(id.name());
      return this;
    }
  }

  /**
   * A single-use {@code LogicTree} builder where each {@link Branch#id()} and
   * {@link Branch#value()} of the built tree is the same string.
   */
  static class StringValueBuilder extends Builder<String> {

    private StringValueBuilder(String name) {
      super(name);
    }

    /**
     * Add a branch to the tree.
     *
     * @param id of the branch
     * @param weight of the branch
     * @return this {@code Builder} object
     * @throws NullPointerException if {@code id} is null
     * @throws IllegalArgumentException if {@code id} is empty or whitespace
     * @throws IllegalArgumentException if a branch with {@code id} has already
     *         been added
     * @throws IllegalArgumentException if {@code weight} is outside the range
     *         (0..1]
     */
    public StringValueBuilder addBranch(String id, double weight) {
      branches.add(new StringValueBranch(id, weight));
      checkIdExists(id);
      return this;
    }
  }
}
