package gov.usgs.earthquake.nshmp;

import static org.junit.jupiter.api.Assertions.assertEquals;
import static org.junit.jupiter.api.Assertions.assertThrows;

import org.junit.jupiter.api.Test;
import org.junit.jupiter.params.ParameterizedTest;
import org.junit.jupiter.params.provider.ValueSource;

final class EarthquakesTests {

  @ParameterizedTest
  @ValueSource(doubles = { 0.0, 5.0, 700.0 })
  final void testCheckDepth_inRange(double z) {
    assertEquals(z, Earthquakes.checkDepth(z), 0.0);
  }

  @ParameterizedTest
  @ValueSource(doubles = { -10.0, -0.1, 700.1, 1000.0 })
  final void testCheckDepth_outOfRange(double z) {
    assertThrows(IllegalArgumentException.class, () -> {
      Earthquakes.checkDepth(z);
    });
  }

  @ParameterizedTest
  @ValueSource(doubles = { 0.0, 5.0, 40.0 })
  final void testCheckCrustalDepth_inRange(double z) {
    assertEquals(z, Earthquakes.checkCrustalDepth(z), 0.0);
  }

  @ParameterizedTest
  @ValueSource(doubles = { -10.0, -0.1, 40.1, 1000.0 })
  final void testCheckCrustalDepth_outOfRange(double z) {
    assertThrows(IllegalArgumentException.class, () -> {
      Earthquakes.checkCrustalDepth(z);
    });
  }

  @ParameterizedTest
  @ValueSource(doubles = { 0.1, 5.0, 60.0 })
  final void testCheckCrustalWidth_inRange(double z) {
    assertEquals(z, Earthquakes.checkCrustalWidth(z), 0.0);
  }

  @ParameterizedTest
  @ValueSource(doubles = { -10.0, 0.0, 60.1, 1000.0 })
  final void testCheckCrustalWidth_outOfRange(double z) {
    assertThrows(IllegalArgumentException.class, () -> {
      Earthquakes.checkCrustalWidth(z);
    });
  }

  @ParameterizedTest
  @ValueSource(doubles = { 0.0, 5.0, 60.0 })
  final void testCheckInterfaceDepth_inRange(double z) {
    assertEquals(z, Earthquakes.checkInterfaceDepth(z), 0.0);
  }

  @ParameterizedTest
  @ValueSource(doubles = { -10.0, -0.1, 60.1, 1000.0 })
  final void testCheckInterfaceDepth_outOfRange(double z) {
    assertThrows(IllegalArgumentException.class, () -> {
      Earthquakes.checkInterfaceDepth(z);
    });
  }

  @ParameterizedTest
  @ValueSource(doubles = { 0.1, 40.0, 200.0 })
  final void testCheckInterfaceWidth_inRange(double z) {
    assertEquals(z, Earthquakes.checkInterfaceWidth(z), 0.0);
  }

  @ParameterizedTest
  @ValueSource(doubles = { -10.0, 0.0, 200.1, 1000.0 })
  final void testCheckInterfaceWidth_outOfRange(double z) {
    assertThrows(IllegalArgumentException.class, () -> {
      Earthquakes.checkInterfaceWidth(z);
    });
  }

  @ParameterizedTest
  @ValueSource(doubles = { 20.0, 30.0, 700.0 })
  final void testCheckSlabDepth_inRange(double z) {
    assertEquals(z, Earthquakes.checkSlabDepth(z), 0.0);
  }

  @ParameterizedTest
  @ValueSource(doubles = { -10.0, 19.9, 700.1, 1000.0 })
  final void testCheckSlabDepth_outOfRange(double z) {
    assertThrows(IllegalArgumentException.class, () -> {
      Earthquakes.checkSlabDepth(z);
    });
  }

  @ParameterizedTest
  @ValueSource(doubles = { 1e-9, 0.5, 1 - 1e-9 })
  final void testRecurrenceRate_inRange(double z) {
    assertEquals(z, Earthquakes.checkRecurrenceRate(z), 0.0);
  }

  @ParameterizedTest
  @ValueSource(doubles = { -10.0, 0.0, 1.0, 1000.0 })
  final void testRecurrenceRate_outOfRange(double z) {
    assertThrows(IllegalArgumentException.class, () -> {
      Earthquakes.checkRecurrenceRate(z);
    });
  }

  @ParameterizedTest
  @ValueSource(doubles = { -2.0, 6.5, 9.7 })
  final void testCheckMagnitude_inRange(double m) {
    assertEquals(m, Earthquakes.checkMagnitude(m), 0.0);
  }

  @ParameterizedTest
  @ValueSource(doubles = { -10.0, -2.1, 9.8, 1000.0 })
  final void testCheckMagnitude_outOfRange(double m) {
    assertThrows(IllegalArgumentException.class, () -> {
      Earthquakes.checkMagnitude(m);
    });
  }

  @Test
  final void testMagToMoment() {
    double moExpect = 6.3095734e18;
    double magInput = 6.5;
    double tolerance = 1e11; // 8 sig figs
    assertEquals(moExpect, Earthquakes.magToMoment(magInput), tolerance);
  }

  @Test
  final void testMomentToMag() {
    double magExpect = 6.5;
    double moInput = 6.3095734e18;
    double tolerance = 1e-8; // 8 sig figs
    assertEquals(magExpect, Earthquakes.momentToMag(moInput), tolerance);
  }

  @Test
  final void testMoment() {
    double fLength = 20 * 1000.0; // 20 km
    double fWidth = 10 * 1000.0; // 10 km
    double slip = 0.02; // 20 mm
    double moExpect = 1.2e17;
    assertEquals(moExpect, Earthquakes.moment(fLength * fWidth, slip), 0.0);
  }

  @Test
  final void testSlip() {
    double fLength = 20 * 1000.0; // 20 km
    double fWidth = 10 * 1000.0; // 10 km
    double moment = 1.2e17;
    double slipExpect = 0.02; // 20 mm
    assertEquals(slipExpect, Earthquakes.slip(fLength * fWidth, moment), 0.0);
  }

}
