package gov.usgs.earthquake.nshmp;

import static org.junit.jupiter.api.Assertions.assertEquals;
import static org.junit.jupiter.api.Assertions.assertNotEquals;
import static org.junit.jupiter.api.Assertions.assertThrows;

import java.nio.file.Files;
import java.nio.file.Path;
import java.util.List;
import java.util.function.Function;

import org.junit.jupiter.api.Test;

import com.google.gson.Gson;
import com.google.gson.GsonBuilder;

import gov.usgs.earthquake.nshmp.geo.json.Feature;
import gov.usgs.earthquake.nshmp.geo.json.GeoJson;

class TextTests {

  private enum TestEnum {
    ABC,
    DEF,
    G_HI_J;

    @Override
    public String toString() {
      return this.name().toLowerCase();
    }
  }

  static final Gson GSON = new GsonBuilder()
      .setPrettyPrinting()
      .disableHtmlEscaping()
      .serializeNulls()
      .create();

  public static void main(String[] args) {
    Path geojson = Path.of("../nshm-fault-sections/fault-sections/AZ/Algodones.geojson");
    System.out.println(Files.exists(geojson));
    Feature f = GeoJson.from(geojson).toFeature();
    String s = GSON.toJson(f);
    System.out.println(s);
    String s2 = Text.cleanCoordinates(s);
    System.out.println(s2);
    System.out.println(dirtyCoords);
    String cleanCoords = Text.cleanCoordinates(dirtyCoords);
    System.out.println(cleanCoords);
  }

  @Test
  final void testCheckName() {
    assertThrows(IllegalArgumentException.class, () -> {
      Text.checkName("   ", "empty name");
    });

    assertEquals("test name", Text.checkName("test name", "test name"));
  }

  @Test
  final void testSplitsAndJoin() {
    String str = "Test string";
    assertEquals(str,
        Text.join(Text.split(str, Text.Delimiter.SPACE), Text.Delimiter.SPACE));

    str = "test";
    assertEquals(str,
        Text.join(Text.split(str, Text.Delimiter.SPACE), Text.Delimiter.SPACE));

    // duplicate delimiters do not produce an empty field
    str = "test_double__underscore_delimited";
    assertNotEquals(str,
        Text.join(Text.split(str, Text.Delimiter.UNDERSCORE),
            Text.Delimiter.UNDERSCORE));
    str = "Test string 1 with   extra spaces";
    assertNotEquals(str,
        Text.join(Text.split(str, Text.Delimiter.SPACE), Text.Delimiter.SPACE));

    str = "test string with μ β θν° unicode";
    assertEquals(str,
        Text.join(Text.split(str, Text.Delimiter.SPACE), Text.Delimiter.SPACE));

    str = "test:colon:delimited";
    assertEquals(str,
        Text.join(Text.split(str, Text.Delimiter.COLON), Text.Delimiter.COLON));

    str = "test,comma,delimited";
    assertEquals(str,
        Text.join(Text.split(str, Text.Delimiter.COMMA), Text.Delimiter.COMMA));

    str = "test-dash-delimited";
    assertEquals(str,
        Text.join(Text.split(str, Text.Delimiter.DASH), Text.Delimiter.DASH));

    str = "test.period.delimited";
    assertEquals(str,
        Text.join(Text.split(str, Text.Delimiter.PERIOD), Text.Delimiter.PERIOD));

    str = "test/slash/delimited";
    assertEquals(str,
        Text.join(Text.split(str, Text.Delimiter.SLASH), Text.Delimiter.SLASH));

    str = "test_underscore_delimited";
    assertEquals(str,
        Text.join(Text.split(str, Text.Delimiter.UNDERSCORE),
            Text.Delimiter.UNDERSCORE));

    str = "1 2 3 4";
    assertEquals(str,
        Text.join(Text.split(str, Text.Delimiter.SPACE), Text.Delimiter.SPACE));
    assertEquals(str,
        Text.join(Text.splitToList(str, Text.Delimiter.SPACE), Text.Delimiter.SPACE));
    assertNotEquals(str,
        Text.join(Text.splitToDoubleList(str, Text.Delimiter.SPACE),
            Text.Delimiter.SPACE));

    str = "1.0 2.0 3.0 4.0";
    assertEquals(str,
        Text.join(Text.splitToDoubleList(str, Text.Delimiter.SPACE),
            Text.Delimiter.SPACE));
    str = "1.1 2.02 3.3 4.0";
    assertEquals(str,
        Text.join(Text.splitToDoubleList(str, Text.Delimiter.SPACE),
            Text.Delimiter.SPACE));
    // trailing zeros are not reproduced
    str = "1.10 2.02 3.3 4.0";
    assertNotEquals(str,
        Text.join(Text.splitToDoubleList(str, Text.Delimiter.SPACE),
            Text.Delimiter.SPACE));

    // doubles below 1e-3 are returned in scientific notation, with capital E
    str = "1.0E-4 2.02 3.3 4.0";
    assertEquals(str,
        Text.join(Text.splitToDoubleList(str, Text.Delimiter.SPACE),
            Text.Delimiter.SPACE));
    str = "1.0E-3 2.02 3.3 4.0";
    assertNotEquals(str,
        Text.join(Text.splitToDoubleList(str, Text.Delimiter.SPACE),
            Text.Delimiter.SPACE));

    List<Double> listDoubles = List.of(1e-4, 2.02, 3.3);
    assertEquals(listDoubles, Text.splitToDoubleList(
        Text.join(listDoubles, Text.Delimiter.COLON), Text.Delimiter.COLON));
  }

  @Test
  final void testEnumsToString() {
    // System.out.println("Print enum as default: " + TestEnum.ABC);
    // System.out.println("Pring enum as name(): " + TestEnum.ABC.name());
    // System.out.println("Print enum toString(): " + TestEnum.ABC.toString());
    // System.out.println();

    // This should be more robust, create expected string with alternative
    // method?
    List<TestEnum> v = List.of(TestEnum.values());
    String enumString = "[ABC,DEF,G_HI_J]";
    assertEquals(enumString, Text.enumsToString(v, TestEnum.class));
  }

  @Test
  final void testEnumLabels() {
    assertEquals("abc", Text.enumLabelWithSpaces(TestEnum.ABC, false));
    assertEquals("Abc", Text.enumLabelWithSpaces(TestEnum.ABC, true));
    assertEquals("g hi j", Text.enumLabelWithSpaces(TestEnum.G_HI_J, false));
    assertEquals("G Hi J", Text.enumLabelWithSpaces(TestEnum.G_HI_J, true));

    assertEquals("abc", Text.enumLabelWithDashes(TestEnum.ABC, false));
    assertEquals("Abc", Text.enumLabelWithDashes(TestEnum.ABC, true));
    assertEquals("g-hi-j", Text.enumLabelWithDashes(TestEnum.G_HI_J, false));
    assertEquals("G-Hi-J", Text.enumLabelWithDashes(TestEnum.G_HI_J, true));
  }

  @Test
  final void testToStringMethods() {
    List<Double> doubles = List.of(1e-5, 2e-4, 3e-3, 0.0, 1.0001, 10.0);
    String fmt_f = "%f";
    String expected_f0 = "[0.000010,0.000200,0.003000,0.0,1.000100,10.000000]";

    assertEquals(expected_f0, Text.toString(doubles, fmt_f, ",", true));

    assertEquals(
        expected_f0.substring(1, expected_f0.length() - 1),
        Text.toString(doubles, fmt_f, ",", false));
    assertEquals(
        expected_f0.replaceAll(",", "|"),
        Text.toString(doubles, fmt_f, "|", true));
    assertEquals(
        expected_f0,
        Text.toString(doubles.stream().mapToDouble(Double::doubleValue), fmt_f, ",", true));
  }

  @Test
  final void testFormatDoubleFunction() {
    Function<Double, String> fmtFunc = Text.formatDoubleFunction("%.3f");
    assertEquals("1.001", fmtFunc.apply(1.0011));
    assertEquals("0.0", fmtFunc.apply(0.00000000));

    // compact (clean) zero-values are only written if the original value is
    // zero, not if it is zero after formatting
    assertEquals("0.000", fmtFunc.apply(0.000000001));
  }

  @Test
  final void testReadDouble() {
    String string = "0.0   1.0  2.0 3.0  4   5.0\n";
    for (double i = 0.0; i <= 5.0; i++) {
      assertEquals(i, Text.readDouble(string, (int) i));
    }
    assertThrows(IndexOutOfBoundsException.class, () -> {
      Text.readDouble(string, 10);
    });
  }

  @Test
  final void testCleanCoordinates() {
    assertEquals(cleanCoords, Text.cleanCoordinates(dirtyCoords));
  }

  static final String dirtyCoords =
      "    \"coordinates\": [\n" +
          "      [\n" +
          "        -114.40759,\n" +
          "        32.36229\n" +
          "      ],\n" +
          "      [\n" +
          "        -114.42494,\n" +
          "        32.39011\n" +
          "      ],\n" +
          "      [\n" +
          "        -114.44586,\n" +
          "        32.42941\n" +
          "      ],\n" +
          "      [\n" +
          "        -114.47051,\n" +
          "        32.45656\n" +
          "      ],\n" +
          "      [\n" +
          "        -114.5132,\n" +
          "        32.48507\n" +
          "      ],\n" +
          "      [\n" +
          "        -114.55289,\n" +
          "        32.5366\n" +
          "      ],\n" +
          "      [\n" +
          "        -114.57051,\n" +
          "        32.55287\n" +
          "      ],\n" +
          "      [\n" +
          "        -114.60994,\n" +
          "        32.58949\n" +
          "      ]\n" +
          "    ]\n" +
          "  },\n" +
          "";

  static final String cleanCoords =
      "    \"coordinates\": [\n" +
          "      [-114.40759, 32.36229],\n" +
          "      [-114.42494, 32.39011],\n" +
          "      [-114.44586, 32.42941],\n" +
          "      [-114.47051, 32.45656],\n" +
          "      [-114.5132, 32.48507],\n" +
          "      [-114.55289, 32.5366],\n" +
          "      [-114.57051, 32.55287],\n" +
          "      [-114.60994, 32.58949]\n" +
          "    ]\n" +
          "  },\n" +
          "";

}
