package gov.usgs.earthquake.nshmp.data;

import static org.junit.jupiter.api.Assertions.assertArrayEquals;
import static org.junit.jupiter.api.Assertions.assertEquals;

import java.util.ArrayList;
import java.util.List;
import java.util.stream.Collectors;
import java.util.stream.DoubleStream;

import org.junit.jupiter.api.Test;

import com.google.common.primitives.Doubles;

class InterpolatorTests {

  static final double[] X = { 0.01, 0.1, 0.3, 0.5, 0.7, 0.9, 1.0 };
  static final double[] Y = { 1, 0.9, 0.7, 0.5, 0.3, 0.1, 1e-2 };

  static final List<Double> X_LIST;
  static final List<Double> Y_LIST;

  static final XySequence XY_SEQUENCE;

  static final double[] X_TARGET = { 0.01, 0.11, 0.4, 0.5, 0.8, 0.99, 1 };
  static final int[] LOWER_BIN_INDEX = { 0, 1, 2, 3, 4, 5, 5 };

  // Expected results calculated in Excel
  static final double[] Y_TARGET_LINX_LINY = { 1.00, 0.89, 0.60, 0.50, 0.20, 0.019, 0.01 };
  static final double[] Y_TARGET_LOGX_LINY = {
      1.00, 0.882648987129, 0.587365841075, 0.50, 0.193733604125, 0.018585096809, 0.01 };
  static final double[] Y_TARGET_LINX_LOGY = {
      1.00, 0.888761607855, 0.591607978310, 0.50, 0.173205080757, 0.012589254118, 0.01 };
  static final double[] Y_TARGET_LOGX_LOGY = {
      1.00, 0.880589847272, 0.579165919197, 0.50, 0.167344511862, 0.012456325964, 0.01 };

  static final Interpolator LINEAR_LINEAR_INTERPOLATOR = Interpolator.builder()
      .build();
  static final Interpolator LINEAR_LINEAR_DOWN_INTERPOLATOR = Interpolator.builder()
      .decreasingY()
      .build();
  static final Interpolator LINEAR_LOG_INTERPOLATOR = Interpolator.builder()
      .logy()
      .build();
  static final Interpolator LOG_LINEAR_INTERPOLATOR = Interpolator.builder()
      .logx()
      .build();
  static final Interpolator LOG_LOG_INTERPOLATOR = Interpolator.builder()
      .logx()
      .logy()
      .build();

  static final double TOL = 1e-10;

  static {
    assert (X.length == Y.length);
    assert (X_TARGET.length == LOWER_BIN_INDEX.length);
    assert (X_TARGET.length == Y_TARGET_LINX_LINY.length);
    assert (X_TARGET.length == Y_TARGET_LOGX_LINY.length);
    assert (X_TARGET.length == Y_TARGET_LINX_LOGY.length);
    assert (X_TARGET.length == Y_TARGET_LOGX_LOGY.length);

    X_LIST = DoubleStream.of(X).boxed().collect(Collectors.toCollection(ArrayList::new));
    Y_LIST = DoubleStream.of(Y).boxed().collect(Collectors.toCollection(ArrayList::new));
    XY_SEQUENCE = XySequence.create(X, Y);
  }

  @Test
  final void testFindX_DoubleDoubleDoubleDoubleDouble() {
    for (int i = 0; i < X_TARGET.length; i++) {
      int j = LOWER_BIN_INDEX[i];
      int k = j + 1;
      double actualX = Interpolator.findX(X[j], Y[j], X[k], Y[k], Y_TARGET_LINX_LINY[i]);
      assertEquals(X_TARGET[i], actualX, TOL);
    }
  }

  @Test
  final void testFindY_DoubleDoubleDoubleDoubleDouble() {
    for (int i = 0; i < X_TARGET.length; i++) {
      int j = LOWER_BIN_INDEX[i];
      int k = j + 1;
      double actualY = Interpolator.findY(X[j], Y[j], X[k], Y[k], X_TARGET[i]);
      assertEquals(Y_TARGET_LINX_LINY[i], actualY, TOL);
    }
  }

  @Test
  final void testFindX_DoubleArrayDoubleArrayDouble() {
    for (int i = 0; i < X_TARGET.length; i++) {
      // need to use an interpolator with decreasingY() since we're using the
      // same data and swapping X & Y
      double actualX = LINEAR_LINEAR_DOWN_INTERPOLATOR.findX(X, Y, Y_TARGET_LINX_LINY[i]);
      assertEquals(X_TARGET[i], actualX, TOL);
    }
  }

  @Test
  final void testFindX_ListOfDoubleListOfDoubleDouble() {
    for (int i = 0; i < X_TARGET.length; i++) {
      // need to use an interpolator with decreasingY() since we're using the
      // same data and swapping X & Y
      double actualX = LINEAR_LINEAR_DOWN_INTERPOLATOR.findX(X_LIST, Y_LIST, Y_TARGET_LINX_LINY[i]);
      assertEquals(X_TARGET[i], actualX, TOL);
    }
  }

  @Test
  final void testFindX_XySequenceDouble() {
    for (int i = 0; i < X_TARGET.length; i++) {
      // need to use an interpolator with decreasingY() since we're using the
      // same data and swapping X & Y
      double actualX = LINEAR_LINEAR_DOWN_INTERPOLATOR.findX(XY_SEQUENCE, Y_TARGET_LINX_LINY[i]);
      assertEquals(X_TARGET[i], actualX, TOL);
    }
  }

  @Test
  final void testFindY_DoubleArrayDoubleArrayDouble() {
    for (int i = 0; i < X_TARGET.length; i++) {
      double actualY = LINEAR_LINEAR_INTERPOLATOR.findY(X, Y, X_TARGET[i]);
      assertEquals(Y_TARGET_LINX_LINY[i], actualY, TOL);
    }
  }

  @Test
  final void testFindY_ListOfDoubleListOfDoubleDouble() {
    for (int i = 0; i < X_TARGET.length; i++) {
      double actualY = LINEAR_LINEAR_INTERPOLATOR.findY(X_LIST, Y_LIST, X_TARGET[i]);
      assertEquals(Y_TARGET_LINX_LINY[i], actualY, TOL);
    }
  }

  @Test
  final void testFindY_XySequenceDouble() {
    for (int i = 0; i < X_TARGET.length; i++) {
      double actualY = LINEAR_LINEAR_INTERPOLATOR.findY(XY_SEQUENCE, X_TARGET[i]);
      assertEquals(Y_TARGET_LINX_LINY[i], actualY, TOL);
    }
  }

  @Test
  final void testFindXin_DoubleArrayDoubleArrayDoubleArray() {

    double[] xs = { 1, 2, 3 };
    double[] ys = { 0.2, 0.5, 0.8 };

    /* Missing coverage for findX functions (XFn) */

    double expected = 1.259921049894;
    double actual = LOG_LINEAR_INTERPOLATOR.findX(xs, ys, 0.3);
    assertEquals(expected, actual, TOL);

    expected = 1.442507049349;
    actual = LINEAR_LOG_INTERPOLATOR.findX(xs, ys, 0.3);
    assertEquals(expected, actual, TOL);

    expected = 1.358963821816;
    actual = LOG_LOG_INTERPOLATOR.findX(xs, ys, 0.3);
    assertEquals(expected, actual, TOL);
  }

  @Test
  final void testFindY_DoubleArrayDoubleArrayDoubleArray() {
    double[] actual = LINEAR_LINEAR_INTERPOLATOR.findY(X, Y, X_TARGET);
    assertArrayEquals(Y_TARGET_LINX_LINY, actual, TOL);

    actual = LOG_LINEAR_INTERPOLATOR.findY(X, Y, X_TARGET);
    assertArrayEquals(Y_TARGET_LOGX_LINY, actual, TOL);

    actual = LINEAR_LOG_INTERPOLATOR.findY(X, Y, X_TARGET);
    assertArrayEquals(Y_TARGET_LINX_LOGY, actual, TOL);

    actual = LOG_LOG_INTERPOLATOR.findY(X, Y, X_TARGET);
    assertArrayEquals(Y_TARGET_LOGX_LOGY, actual, TOL);
  }

  @Test
  final void testFindY_ListOfDoubleListOfDoubleDoubleArray() {
    double[] actual = LINEAR_LINEAR_INTERPOLATOR.findY(X_LIST, Y_LIST, X_TARGET);
    assertArrayEquals(Y_TARGET_LINX_LINY, actual, TOL);

    actual = LOG_LINEAR_INTERPOLATOR.findY(X_LIST, Y_LIST, X_TARGET);
    assertArrayEquals(Y_TARGET_LOGX_LINY, actual, TOL);

    actual = LINEAR_LOG_INTERPOLATOR.findY(X_LIST, Y_LIST, X_TARGET);
    assertArrayEquals(Y_TARGET_LINX_LOGY, actual, TOL);

    actual = LOG_LOG_INTERPOLATOR.findY(X_LIST, Y_LIST, X_TARGET);
    assertArrayEquals(Y_TARGET_LOGX_LOGY, actual, TOL);
  }

  @Test
  final void testFindY_XySequenceDoubleArray() {
    double[] actual = LINEAR_LINEAR_INTERPOLATOR.findY(XY_SEQUENCE, X_TARGET);
    assertArrayEquals(Y_TARGET_LINX_LINY, actual, TOL);

    actual = LOG_LINEAR_INTERPOLATOR.findY(XY_SEQUENCE, X_TARGET);
    assertArrayEquals(Y_TARGET_LOGX_LINY, actual, TOL);

    actual = LINEAR_LOG_INTERPOLATOR.findY(XY_SEQUENCE, X_TARGET);
    assertArrayEquals(Y_TARGET_LINX_LOGY, actual, TOL);

    actual = LOG_LOG_INTERPOLATOR.findY(XY_SEQUENCE, X_TARGET);
    assertArrayEquals(Y_TARGET_LOGX_LOGY, actual, TOL);
  }

  /* Errors and Edge Cases */
  @Test
  final void testErrorsAndEdgeCases() {

    double[] xs = { 2, 3, 4 };
    double[] ys = { 0.2, 0.5, 0.8 };
    Interpolator interp = Interpolator.builder().build();

    /* linear index */

    double yBelow = 0.1;
    double yAbove = 0.9;

    // out of range below
    double actual = interp.findX(xs, ys, yBelow);
    assertEquals(0.0, actual, 0.0);
    actual = interp.findX(
        Doubles.asList(xs),
        Doubles.asList(ys), yBelow);
    assertEquals(0.0, actual, 0.0);

    // out of range above
    actual = interp.findX(xs, ys, yAbove);
    assertEquals(0.0, actual, 0.0);
    actual = interp.findX(
        Doubles.asList(xs),
        Doubles.asList(ys), yAbove);
    assertEquals(0.0, actual, 0.0);

    /* binary index */

    // out of range below
    double xBelow = 1.0;
    actual = interp.findY(xs, ys, xBelow);
    assertEquals(-0.1, actual, TOL);

  }

}
