package gov.usgs.earthquake.nshmp.data;

import static org.junit.jupiter.api.Assertions.assertArrayEquals;
import static org.junit.jupiter.api.Assertions.assertDoesNotThrow;
import static org.junit.jupiter.api.Assertions.assertEquals;
import static org.junit.jupiter.api.Assertions.assertSame;
import static org.junit.jupiter.api.Assertions.assertThrows;

import java.util.Arrays;
import java.util.Optional;

import org.junit.jupiter.api.Test;

import com.google.common.primitives.Doubles;

class SequencesTests {

  @Test
  final void validateArraysTests() {
    // empty xs array
    assertThrows(IllegalArgumentException.class, () -> {
      Sequences.validateArrays(new double[0], new double[0]);
    });
    // xs and ys different sizes
    assertThrows(IllegalArgumentException.class, () -> {
      Sequences.validateArrays(new double[2], new double[3]);
    });
    // xs not monotonic
    assertThrows(IllegalArgumentException.class, () -> {
      Sequences.validateArrays(new double[] { 1, 0 }, new double[2]);
    });
    // xs not monotonic
    assertDoesNotThrow(() -> {
      Sequences.validateArrays(new double[] { 0, 1 }, new double[2]);
    });
    // single values makes it through monotonic test
    assertDoesNotThrow(() -> {
      Sequences.validateArrays(new double[1], new double[1]);
    });
  }

  @Test
  final void trimTests() {
    double[] xs1 = new double[1];
    double[] ys1 = new double[1];

    double[] xs3 = { 1, 2, 3 };
    double[] ys3zero = new double[3];
    double[] ys3ones = { 1, 1, 1 };

    // size and clear check
    ArrayXySequence xy1 = (ArrayXySequence) XySequence.create(xs1, ys1);
    assertSame(xy1, Sequences.trim(xy1));
    ArrayXySequence xy3a = (ArrayXySequence) XySequence.create(xs3, ys3zero);
    assertSame(xy3a, Sequences.trim(xy3a));

    // start and end y values > 0
    ArrayXySequence xy3b = (ArrayXySequence) XySequence.create(xs3, ys3ones);
    assertSame(xy3b, Sequences.trim(xy3b));

    double[] xs4toTrim = { 1, 2, 3, 4 };
    double[] ys4toTrim = { 1, 1, 1, 0 };
    double[] xs4expect = { 1, 2, 3 };
    double[] ys4expect = { 1, 1, 1 };

    // extra test that gets upper end y = 0
    ArrayXySequence xy4toTrim = (ArrayXySequence) XySequence.create(xs4toTrim, ys4toTrim);
    ArrayXySequence xy4expect = (ArrayXySequence) XySequence.create(xs4expect, ys4expect);
    assertEquals(xy4expect, Sequences.trim(xy4toTrim));

    double[] xs5toTrim = { 1, 2, 3, 4, 5 };
    double[] ys5toTrim = { 0, 1, 1, 1, 0 };
    double[] xs5expect = { 2, 3, 4 };
    double[] ys5expect = { 1, 1, 1 };

    // immutable trimmed
    ArrayXySequence xy5toTrim = (ArrayXySequence) XySequence.create(xs5toTrim, ys5toTrim);
    ArrayXySequence xy5expect = (ArrayXySequence) XySequence.create(xs5expect, ys5expect);
    assertEquals(xy5expect, Sequences.trim(xy5toTrim));

    // mutable trimmed
    MutableArrayXySequence xy5toTrimMutable = (MutableArrayXySequence) MutableXySequence.create(
        xs5toTrim,
        Optional.of(ys5toTrim));
    assertEquals(xy5expect, Sequences.trim(xy5toTrimMutable));
  }

  @Test
  final void nonZeroIndexTests() {
    double[] values = { 0.0, 0.0, 1.0, 1.0, 0.0, 0.0 };
    double[] zeros = { 0.0 };
    int first = 2;
    int last = 3;

    assertEquals(first, Sequences.firstNonZeroIndex(values), 0.0);
    assertEquals(-1, Sequences.firstNonZeroIndex(zeros), 0.0);
    assertEquals(last, Sequences.lastNonZeroIndex(values), 0.0);
    assertEquals(-1, Sequences.lastNonZeroIndex(zeros), 0.0);
  }

  @Test
  final void toCumulativeTests() {
    double[] xs = { 1, 2, 3, 4, 5 };
    double[] ys = { 1, 1, 1, 1, 1 };
    double[] ysExpect = { 5, 4, 3, 2, 1 };

    XySequence xyActual = Sequences.toCumulative(XySequence.create(xs, ys));
    XySequence xyExpect = XySequence.create(xs, ysExpect);
    assertEquals(xyExpect, xyActual);
  }

  @Test
  final void arrayBuilderTests() {

    double[] expected =
        { 5.05, 5.1499999999999995, 5.25, 5.35, 5.45, 5.55, 5.65, 5.75 };
    double[] actual = Sequences.arrayBuilder(5.05, 5.83, 0.1).centered().build();
    assertArrayEquals(expected, actual);

    expected =
        new double[] { 5.05, 5.15, 5.25, 5.35, 5.45, 5.55, 5.65, 5.75, 5.85 };
    actual = Sequences.arrayBuilder(5.05, 5.85, 0.1).scale(4).centered().build();
    assertArrayEquals(expected, actual);

    expected =
        new double[] { 6.55625, 6.66875, 6.78125, 6.89375, 7.00625, 7.11875, 7.23125, 7.34375 };
    actual = Sequences.arrayBuilder(6.55625, 7.34385, 0.1125).scale(5).centered().build();
    assertArrayEquals(expected, actual);

    expected =
        new double[] { 6.55625, 6.66875, 6.78125, 6.89375, 7.00625, 7.11875, 7.23125, 7.34375 };
    actual = Sequences.arrayBuilder(6.5, 7.4, 0.1125).scale(5).build();
    assertArrayEquals(expected, actual);

    expected = Arrays.copyOf(expected, expected.length);
    Doubles.reverse(expected);
    actual = Sequences.arrayBuilder(6.5, 7.4, 0.1125).descending().scale(5).build();
    assertArrayEquals(expected, actual);

    expected = new double[] { 6.575, 6.725 };
    actual = Sequences.arrayBuilder(6.5, 6.8, 0.15).scale(5).build();
    assertArrayEquals(expected, actual);

    double maxNoRound = 6.7999999; // 2 value array expected
    double[] expectNoRound = { 6.5, 6.65 };
    double[] actualNoRound = Sequences.arrayBuilder(6.5, maxNoRound, 0.15)
        .scale(5)
        .centered()
        .build();
    assertArrayEquals(expectNoRound, actualNoRound);

    double maxRound = 6.79999999; // 3 value array expected
    double[] expectRound = { 6.5, 6.65, 6.8 };
    double[] actualRound = Sequences.arrayBuilder(6.5, maxRound, 0.15)
        .scale(5)
        .centered()
        .build();
    assertArrayEquals(expectRound, actualRound);

  }

  @Test
  final void arrayBuilderExceptions() {

    /* finiteness */
    assertThrows(IllegalArgumentException.class, () -> {
      Sequences.arrayBuilder(Double.NaN, 6.5, 0.1).build();
    });
    assertThrows(IllegalArgumentException.class, () -> {
      Sequences.arrayBuilder(6.5, Double.NaN, 0.1).build();
    });
    assertThrows(IllegalArgumentException.class, () -> {
      Sequences.arrayBuilder(6.0, 6.5, Double.NaN).build();
    });

    /* max < min to 6 decimal precision */
    assertThrows(IllegalArgumentException.class, () -> {
      Sequences.arrayBuilder(6.5, 6.499999, 0.15).centered().build();
    });
    assertThrows(IllegalArgumentException.class, () -> {
      Sequences.arrayBuilder(6.5, 6.6, 0.15).build();
    });

    /* Δ <= 0 */
    assertThrows(IllegalArgumentException.class, () -> {
      Sequences.arrayBuilder(6.5, 6.6, 0.0).centered().build();
    });

  }

}
