package gov.usgs.earthquake.nshmp.data;

import static org.junit.jupiter.api.Assertions.assertArrayEquals;
import static org.junit.jupiter.api.Assertions.assertEquals;
import static org.junit.jupiter.api.Assertions.assertFalse;
import static org.junit.jupiter.api.Assertions.assertNotSame;
import static org.junit.jupiter.api.Assertions.assertSame;
import static org.junit.jupiter.api.Assertions.assertThrows;
import static org.junit.jupiter.api.Assertions.assertTrue;

import java.util.Arrays;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.junit.jupiter.api.Test;

import com.google.common.collect.ImmutableList;
import com.google.common.primitives.Doubles;

import gov.usgs.earthquake.nshmp.Text;

class XySequenceTests {

  /* Tests XySequence and array backed implementation. */

  private static double[] xs = new double[] { 0, 1, 2, 3 };
  private static double[] ys = new double[] { -1, 10.5, 5.25, 2.5 };
  private static XySequence xy = XySequence.create(xs, ys);

  @Test
  final void factoryConstructorTests() {
    // arrays
    assertArrayEquals(xs, ((ArrayXySequence) xy).xs, 0);
    assertArrayEquals(ys, ((ArrayXySequence) xy).ys, 0);
    // collections; guava Doubles copies array on toArray()
    List<Double> xList = Doubles.asList(xs);
    List<Double> yList = Doubles.asList(ys);
    XySequence xy2 = XySequence.create(xList, yList);
    assertArrayEquals(xs, ((ArrayXySequence) xy2).xs, 0);
    assertArrayEquals(ys, ((ArrayXySequence) xy2).ys, 0);

    XySequence xy3 = XySequence.copyOf(xy);
    assertSame(xy, xy3);

    // tests that a mutable sequence passed to the XySequence
    // constructor is copied
    MutableXySequence xy4 = MutableXySequence.copyOf(xy);
    XySequence xy5 = XySequence.copyOf(xy4);
    assertNotSame(xy4, xy5);

    assertThrows(IllegalArgumentException.class, () -> {
      XySequence.create(new double[0], new double[0]);
    });

    assertThrows(IllegalArgumentException.class, () -> {
      XySequence.create(xs, new double[0]);
    });

    assertThrows(IllegalArgumentException.class, () -> {
      double[] reversed = Arrays.copyOf(xs, xs.length);
      Doubles.reverse(reversed);
      XySequence.create(reversed, ys);
    });
  }

  private enum MapKey {
    K1;
  }

  @Test
  final void addToMapTest() {

    Map<MapKey, MutableXySequence> testMap = new HashMap<>();

    // test adding a mutable copy to map for non-existant key
    XySequence.addToMap(MapKey.K1, testMap, xy);
    assertNotSame(xy, testMap.get(MapKey.K1));
    assertTrue(testMap.get(MapKey.K1) instanceof MutableArrayXySequence);

    // test adding to existing key
    XySequence.addToMap(MapKey.K1, testMap, xy);
    double[] sum = DoubleData.add(Arrays.copyOf(ys, ys.length), ys);
    assertArrayEquals(sum, ((MutableArrayXySequence) testMap.get(MapKey.K1)).ys, 0);

    assertThrows(IllegalArgumentException.class, () -> {
      Map<MapKey, MutableXySequence> map = new HashMap<>();
      XySequence.addToMap(MapKey.K1, map, xy);
      XySequence xyNonDuplicateX = XySequence.create(new double[] { 1, 2, 3, 4 }, ys);
      XySequence.addToMap(MapKey.K1, map, xyNonDuplicateX);
    });
  }

  @Test
  final void pointSetTest() {
    assertThrows(UnsupportedOperationException.class, () -> {
      xy.stream().forEach(p -> p.set(0));
    });
  }

  @Test
  final void pointIndexTest() {
    int index = 0;
    for (XyPoint p : xy) {
      assertEquals(index++, p.index());
    }
  }

  @Test
  final void xyTest() {
    for (int index = 0; index < xy.size(); index++) {
      assertEquals(xs[index], xy.x(index), 0);
      assertEquals(ys[index], xy.y(index), 0);
    }

    assertThrows(IndexOutOfBoundsException.class, () -> {
      xy.x(10);
    });

    assertThrows(IndexOutOfBoundsException.class, () -> {
      xy.y(10);
    });
  }

  @Test
  final void xyValuesTest() {
    assertArrayEquals(xs, xy.xValues().toArray(), 0);
    assertArrayEquals(ys, xy.yValues().toArray(), 0);
  }

  @Test
  final void streamTest() {
    assertArrayEquals(xs, xy.stream().mapToDouble(XyPoint::x).toArray(), 0);
    assertArrayEquals(ys, xy.stream().mapToDouble(XyPoint::y).toArray(), 0);
  }

  @Test
  final void sizeTest() {
    assertEquals(xs.length, xy.size());
    assertEquals(ys.length, xy.size());
  }

  @Test
  final void minMaxTest() {
    XyPoint first = xy.min();
    assertEquals(xs[0], first.x(), 0);
    assertEquals(ys[0], first.y(), 0);
    XyPoint last = xy.max();
    assertEquals(xs[xs.length - 1], last.x(), 0);
    assertEquals(ys[ys.length - 1], last.y(), 0);
    // test correct class (immutable)
    assertEquals(first.getClass(), ArrayXySequence.Point.class);
    assertEquals(last.getClass(), ArrayXySequence.Point.class);
  }

  @Test
  final void isClearTest() {
    assertFalse(xy.isClear());
    XySequence xy = XySequence.create(
        new double[] { 1, 2, 3 },
        new double[3]);
    assertTrue(xy.isClear());
  }

  @Test
  final void trimTest() {
    double[] xsOriginal = { 0, 1, 2, 3, 4, 5 };
    double[] ysOriginal = { 0, 0, 1, 2, 0, 0 };
    double[] xsExpected = { 2, 3 };
    double[] ysExpected = { 1, 2 };
    XySequence xyExpected = XySequence.create(xsExpected, ysExpected);
    XySequence xyTrimmed = XySequence.create(xsOriginal, ysOriginal).trim();
    assertEquals(xyExpected, xyTrimmed);
    // test class type
    assertEquals(xyTrimmed.getClass(), ArrayXySequence.class);
  }

  @Test
  final void testEquals() {
    XySequence xyCopy = XySequence.copyOf(xy);
    // test ==
    assertTrue(xyCopy.equals(xy));
    // test instanceof
    assertFalse(xy.equals(new Object()));
    // test identical data
    XySequence xyDuplicate = XySequence.create(xs, ys);
    assertTrue(xy.equals(xyDuplicate));
    // test non-identical data
    XySequence xyNonDuplicateX = XySequence.create(new double[] { 1, 2, 3, 4 }, ys);
    assertFalse(xy.equals(xyNonDuplicateX));
    XySequence xyNonDuplicateY = XySequence.create(xs, new double[xs.length]);
    assertFalse(xy.equals(xyNonDuplicateY));
  }

  @Test
  final void testHashCode() {
    int hashActual = xy.hashCode();
    int hashExpected = Arrays.deepHashCode(new Object[] { xs, ys });
    assertEquals(hashExpected, hashActual);
  }

  @Test
  final void testValidateSequence() {
    // testing hash mismatch but matching values succeeds
    ArrayXySequence xyAltHashAndValues =
        (ArrayXySequence) XySequence.create(new double[] { 0, 1, 2, 3 }, ys);
    ArrayXySequence axy = (ArrayXySequence) xy;
    axy.validateSequence(xyAltHashAndValues);

    assertThrows(IllegalArgumentException.class, () -> {
      // casting required for internal methods
      ArrayXySequence xyAltHash =
          (ArrayXySequence) XySequence.create(new double[] { 1, 2, 3, 4 }, ys);
      ArrayXySequence axy2 = (ArrayXySequence) xy;
      axy2.validateSequence(xyAltHash);
    });
  }

  @Test
  final void testToString() {
    String expected = new StringBuilder()
        .append("ArrayXySequence:").append(Text.NEWLINE)
        .append(" [0.0, -1.0]").append(Text.NEWLINE)
        .append(" [1.0, 10.5]").append(Text.NEWLINE)
        .append(" [2.0, 5.25]").append(Text.NEWLINE)
        .append(" [3.0, 2.5]").toString();
    assertEquals(expected, xy.toString());
  }

  @Test
  final void testIteratorRemoveUOE() {
    assertThrows(UnsupportedOperationException.class, () -> {
      xy.iterator().remove();
    });
  }

  private static final XySequence xyBase;
  private static final XySequence xySameXs;
  private static final XySequence xyDifferentXs;
  private static final XySequence xyShorter;

  private static final List<XySequence> emptyCollection;
  private static final List<XySequence> singleSequence;
  private static final List<XySequence> identicalSequences;
  private static final List<XySequence> sameXs;
  private static final List<XySequence> differentXs;
  private static final List<XySequence> withShorterXs;

  private static final XySequence expectedSingleSequenceResults;
  private static final XySequence expectedIdenticalSequences;
  private static final XySequence expectedSameXs;
  private static final XySequence expectedDifferentXs;
  private static final XySequence expectedWithShorterXs;

  static {
    double[] xs1 = new double[] { 0, 1, 2, 3 };
    double[] xs2 = new double[] { 1, 2, 3, 4 };
    double[] xs3 = new double[] { 5, 6, 7 };

    double[] ys1 = new double[] { 10, 9, 8, 7 };
    double[] ys2 = new double[] { 5, 4, 3, 2 };
    double[] ys3 = new double[] { 10, 9, 8 };

    xyBase = XySequence.create(xs1, ys1);
    xySameXs = XySequence.create(xs1, ys2);
    xyDifferentXs = XySequence.create(xs2, ys2);
    xyShorter = XySequence.create(xs3, ys3);

    emptyCollection = new ImmutableList.Builder<XySequence>().build();

    singleSequence = new ImmutableList.Builder<XySequence>()
        .add(xyBase)
        .build();
    expectedSingleSequenceResults = xyBase;

    // add two copies of the same sequence
    identicalSequences = new ImmutableList.Builder<XySequence>()
        .add(xyBase)
        .add(xyBase)
        .build();
    double[] result = new double[xs1.length];
    Arrays.setAll(result, i -> ys1[i] + ys1[i]);
    expectedIdenticalSequences = XySequence.create(xs1, result);

    // sequences with the same x-values (similar)
    sameXs = new ImmutableList.Builder<XySequence>()
        .add(xyBase)
        .add(xySameXs)
        .build();
    Arrays.setAll(result, i -> ys1[i] + ys2[i]);
    expectedSameXs = XySequence.create(xs1, result);

    // sequences with different x-values
    differentXs = new ImmutableList.Builder<XySequence>()
        .add(xyBase)
        .add(xyDifferentXs)
        .build();
    // manually construct for now
    expectedDifferentXs = XySequence.create(
        new double[] { 0, 1, 2, 3, 4 },
        new double[] { 10, 14, 12, 10, 2 });

    // sequences with different x-values and different length
    withShorterXs = new ImmutableList.Builder<XySequence>()
        .add(xyBase)
        .add(xyDifferentXs)
        .add(xyShorter)
        .build();
    // manually construct for now
    expectedWithShorterXs = XySequence.create(
        new double[] { 0, 1, 2, 3, 4, 5, 6, 7 },
        new double[] { 10, 14, 12, 10, 2, 10, 9, 8 });
  }

  @Test
  final void areSimilarTests_emptyThrowsIAE() {
    assertThrows(IllegalArgumentException.class, () -> {
      XySequence.areSimilar(emptyCollection);
    });
  }

  @Test
  final void areSimilarTests() {
    assertTrue(XySequence.areSimilar(singleSequence));
    assertTrue(XySequence.areSimilar(identicalSequences));
    assertTrue(XySequence.areSimilar(sameXs));
    assertFalse(XySequence.areSimilar(differentXs));
    assertFalse(XySequence.areSimilar(withShorterXs));
  }

  @Test
  final void combineTests_emptyThrowsIAE() {
    assertThrows(IllegalArgumentException.class, () -> {
      XySequence.combine(emptyCollection);
    });
  }

  @Test
  final void combineTests_similarSequences() {
    assertEquals(expectedSingleSequenceResults, XySequence.combine(singleSequence));
    assertEquals(expectedIdenticalSequences, XySequence.combine(identicalSequences));
    assertEquals(expectedSameXs, XySequence.combine(sameXs));
  }

  @Test
  final void combineTests_dissimilarSequences() {
    assertEquals(expectedDifferentXs, XySequence.combine(differentXs));
    assertEquals(expectedWithShorterXs, XySequence.combine(withShorterXs));
  }

}
