package gov.usgs.earthquake.nshmp.gmm;

import static org.junit.jupiter.api.Assertions.assertEquals;

import java.io.IOException;
import java.net.URL;
import java.nio.charset.StandardCharsets;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import java.util.Set;
import java.util.stream.Collectors;
import java.util.stream.Stream;

import org.junit.jupiter.params.provider.Arguments;

import com.google.common.collect.FluentIterable;
import com.google.common.collect.Lists;
import com.google.common.io.Resources;
import com.google.common.primitives.Doubles;

import gov.usgs.earthquake.nshmp.Text;
import gov.usgs.earthquake.nshmp.Text.Delimiter;

class GmmTest {

  private static final double TOL = 1e-10;

  static void test(
      int index,
      Gmm gmm,
      Imt imt,
      double exMedian,
      double exSigma,
      String inputs) {
    List<GmmInput> inputsList = new ArrayList<>();

    try {
      inputsList = loadInputs(inputs);
    } catch (IOException ioe) {
      ioe.printStackTrace();
      System.exit(1);
    }

    GmmInput in = inputsList.get(index);
    GroundMotion sgm = GroundMotions.combine(gmm.instance(imt).calc(in));
    assertEquals(exMedian, Math.exp(sgm.mean()), TOL);
    assertEquals(exSigma, sgm.sigma(), TOL);
  }

  /* Use to generate Gmm result file */
  static void generateResults(
      Set<Gmm> gmms,
      Set<Imt> imts,
      String inputsFileName,
      String resultsFileName) throws IOException {

    List<GmmInput> inputs = loadInputs(inputsFileName);
    List<String> lines = new ArrayList<>();
    for (Gmm gmm : gmms) {
      for (Imt imt : imts) {
        GroundMotionModel gmModel = gmm.instance(imt);
        int modelIndex = 0;
        String id = gmm.name() + "-" + imt.name();
        for (GmmInput input : inputs) {
          GroundMotion sgm = GroundMotions.combine(gmModel.calc(input));
          String result = Text.join(
              Lists.newArrayList(modelIndex++ + "-" + id,
                  String.format("%.10f", Math.exp(sgm.mean())),
                  String.format("%.10f", sgm.sigma())),
              Delimiter.COMMA);
          lines.add(result);
        }
      }
    }
    Path out = Paths.get("tmp/gmm-tests/", resultsFileName);
    Files.write(out, lines, StandardCharsets.UTF_8);
  }

  static Stream<? extends Arguments> loadResults(String results, String inputs) throws IOException {
    URL url = Resources.getResource(GmmTest.class, results);

    return Resources.readLines(url, StandardCharsets.UTF_8).stream()
        .map(line -> GmmTest.resultsToObject(line, inputs))
        .map(Arguments::of);
  }

  /*
   * Loads GmmTest inputs from CSV file - Header line indicating parameter names
   * ('Mw,rjB,rRup,...') must start with '#' - Blank lines and lines containing
   * only whitespace are skipped - Lines whose first non-whitespace character is
   * '#' are comment lines and skipped
   */
  static List<GmmInput> loadInputs(String resource) throws IOException {
    URL url = Resources.getResource(GmmTest.class, resource);

    return Resources.readLines(url, StandardCharsets.UTF_8)
        .stream()
        .filter(GmmTest::filterBlankLines)
        .filter(GmmTest::filterComments)
        .map(GmmTest::argsToInput)
        .collect(Collectors.toList());
  }

  /* Predicate to skip lines whose first non-whitespace character is '#'. */
  private static boolean filterComments(String line) {
    // replaces inline `.filter(i -> !i.trim().startswith("#"))`
    return !line.trim().startsWith("#");
  }

  /* Predicate to skip blank lines and lines containing only whitespace. */
  private static boolean filterBlankLines(String line) {
    // replaces inline `.filter(i -> !i.trim().isEmpty())`
    return !line.trim().isEmpty();
  }

  private static Object[] resultsToObject(String line, String inputs) {
    Iterator<String> lineIt = Text.split(line, Delimiter.COMMA).iterator();
    Iterator<String> idIt = Text.split(lineIt.next(), Delimiter.DASH).iterator();
    return new Object[] {
        Integer.valueOf(idIt.next()), // inputs index
        Gmm.valueOf(idIt.next()), // Gmm
        Imt.valueOf(idIt.next()), // Imt
        Double.valueOf(lineIt.next()), // median
        Double.valueOf(lineIt.next()), // sigma
        inputs
    };
  }

  private static GmmInput argsToInput(String line) {
    Iterator<Double> it = FluentIterable
        .from(Text.split(line, Delimiter.COMMA))
        .transform(Doubles.stringConverter())
        .iterator();

    return GmmInput.builder()
        .mag(it.next())
        .distances(it.next(), it.next(), it.next())
        .dip(it.next())
        .width(it.next())
        .zTor(it.next())
        .zHyp(it.next())
        .rake(it.next())
        .vs30(it.next())
        .vsInf(it.next() > 0.0)
        .z1p0(it.next())
        .z2p5(it.next())
        .zSed(Double.NaN)
        .build();
  }

}
