package gov.usgs.earthquake.nshmp.gmm;

import static gov.usgs.earthquake.nshmp.gmm.Gmm.ASK_14;
import static gov.usgs.earthquake.nshmp.gmm.Gmm.BCHYDRO_12_INTERFACE;
import static gov.usgs.earthquake.nshmp.gmm.Gmm.BSSA_14;
import static gov.usgs.earthquake.nshmp.gmm.Gmm.CB_14;
import static gov.usgs.earthquake.nshmp.gmm.Gmm.CY_14;
import static gov.usgs.earthquake.nshmp.gmm.Imt.PGA;
import static gov.usgs.earthquake.nshmp.gmm.Imt.PGV;
import static gov.usgs.earthquake.nshmp.gmm.Imt.SA1P0;

import java.io.IOException;
import java.util.EnumSet;
import java.util.List;
import java.util.Set;
import java.util.stream.Stream;

import org.junit.jupiter.api.extension.ExtensionContext;
import org.junit.jupiter.params.ParameterizedTest;
import org.junit.jupiter.params.provider.Arguments;
import org.junit.jupiter.params.provider.ArgumentsProvider;
import org.junit.jupiter.params.provider.ArgumentsSource;

class UsgsPgvSupportTest implements ArgumentsProvider {

  private static String GMM_INPUTS = "/gmm/UsgsPgvSupport-inputs.csv";
  private static String GMM_RESULTS = "/gmm/UsgsPgvSupport-results.csv";

  @ParameterizedTest(name = "({index}) {0} {2} {1}")
  @ArgumentsSource(UsgsPgvSupportTest.class)
  void test(int index, Gmm gmm, Imt imt, double exMedian, double exSigma, String inputs) {
    GmmTest.test(index, gmm, imt, exMedian, exSigma, inputs);
  }

  @Override
  public Stream<? extends Arguments> provideArguments(ExtensionContext context) throws Exception {
    return GmmTest.loadResults(GMM_RESULTS, GMM_INPUTS);
  }

  /* Result generation sets */
  private static Set<Gmm> gmms = EnumSet.of(
      BCHYDRO_12_INTERFACE);

  private static Set<Imt> imts = EnumSet.of(
      PGV);

  public static void main(String[] args) throws IOException {
    GmmTest.generateResults(gmms, imts, GMM_INPUTS, GMM_RESULTS);

    /*
     * Generate broader test results (printed to console)
     */

    // Compute conditional PGV along with supplementary SA values for validation

    String gmm_inputs = GMM_INPUTS;

    List<Gmm> gmms = List.of(
        BCHYDRO_12_INTERFACE,
        ASK_14,
        BSSA_14,
        CB_14,
        CY_14);

    // Output, for each GMM and each GmmInput:
    // - Mw, rJB, rRup, Vs30
    // - PGA, sigmaPGA
    // - SA(T=1), sigmaSA(T=1)
    // - Native PGV, sigmaPGV (if supported)
    // - AB20 Conditional PGV model:
    // --- Tpgv
    // --- SA(Tpgv), sigmaSA(Tpgv)
    // --- PGVab20, sigmaPGVab20
    // - NH82 PGV:
    // --- PGVnh82, sigmaPGVnh82
    String header1 = ",,Inputs,,,,Native SA,,,,Native PGV,,AB_08 Conditional PGV,,,,,NH82 PGV,";
    String header2 =
        "Idx,GMM_NAME,Mw,rJB,rRup,vs30,PGA,sigmaPGA,SA(1.0),sigmaSA(1.0),PGV,sigmaPGV,Tpgv,SA(Tpgv),sigmaSA(Tpgv),PGVab20,sigmaPGVab20,PGVnh82,sigmaPGVnh82";
    String fmt =
        "%d,%s,%.2f,%.2f,%.2f,%.2f,%.8e,%.8e,%.8e,%.8e,%s,%s,%.8e,%.8e,%.8e,%.8e,%.8e,%.8e,%.8e\n";

    System.out.println(header1);
    System.out.println(header2);

    // todo: generate inputs given lists of target Mw, rJB, Vs30, using simple
    // map of Mw to zTor and zHyp to calculate rRup and width

    List<GmmInput> inputs = GmmTest.loadInputs(gmm_inputs);

    for (Gmm gmm : gmms) {

      int inputIdx = 0;

      for (GmmInput in : inputs) {

        /*
         * Compute native PGA and SA1P0
         */
        GroundMotion sgmPga = GroundMotions.combine(gmm.instance(PGA).calc(in));
        GroundMotion sgmSA1p0 = GroundMotions.combine(gmm.instance(SA1P0).calc(in));

        /*
         * Compute native PGV, if supported
         */
        double nativePgv = -1;
        double nativeSigmaPgv = -1;
        String nativePgvStr;
        String nativeSigmaPgvStr;
        if (gmm.supportedImts().contains(PGV)) {
          // BCH supports PGV now (via UsgsPgvSupport), so none of these GMMs
          // will bypass this
          GroundMotion sgmPgv = GroundMotions.combine(gmm.instance(PGV).calc(in));
          nativePgv = Math.exp(sgmPgv.mean());
          nativeSigmaPgv = sgmPgv.sigma();
          nativePgvStr = String.format("%.8e", nativePgv);
          nativeSigmaPgvStr = String.format("%.8e", nativeSigmaPgv);
        } else {
          // output N/A for Pgv
          nativePgvStr = "N/A";
          nativeSigmaPgvStr = "N/A";
        }

        /*
         * Compute AB20 Conditional PGV parameters and values
         */

        // compute Tpgv
        double tPgv = UsgsPgvSupport.computeTPgv(in.Mw);

        // compute SA(Tpgv)
        GroundMotion sgmTPgv = UsgsPgvSupport.calcSaGroundMotion(gmm, in);

        // compute AB20 PGV
        GroundMotion sgmPgvAB20 = UsgsPgvSupport.calcAB20Pgv(gmm, in);

        /*
         * Compute NH82 PGV
         */
        GroundMotion sgmPgvNH82 = UsgsPgvSupport.calcPgvNH(gmm, in);

        /*
         * Print results
         */
        // "IDX-GMM_NAME,M,Rrup,Vs30,PGA,sigmaPGA,SA1P0,sigmaSA1P0,PGV,sigmaPGV,Tpgv,SA(Tpgv),sigmaSA(Tpgv),PGVab20,sigmaPGVab20,PGVnh82,sigmaPGVnh82";
        System.out.printf(fmt, inputIdx++, gmm.name(), in.Mw, in.rJB, in.rRup, in.vs30,
            Math.exp(sgmPga.mean()), sgmPga.sigma(),
            Math.exp(sgmSA1p0.mean()), sgmSA1p0.sigma(),
            nativePgvStr, nativeSigmaPgvStr,
            tPgv, Math.exp(sgmTPgv.mean()), sgmTPgv.sigma(),
            Math.exp(sgmPgvAB20.mean()), sgmPgvAB20.sigma(),
            Math.exp(sgmPgvNH82.mean()), sgmPgvNH82.sigma());

      }
    }

  }

}
