package gov.usgs.earthquake.nshmp.tree;

import static org.junit.jupiter.api.Assertions.assertEquals;
import static org.junit.jupiter.api.Assertions.assertThrows;

import java.util.List;

import org.junit.jupiter.api.Test;

import gov.usgs.earthquake.nshmp.Text;

class LogicTreeTests {

  private static enum EnumKey {
    ID_0,
    ID_1,
    ID_2;
  }

  @Test
  final void treeTests() {

    /* Tests building and common methods for different tree/branch types. */

    String[] id = new String[] { "id-0", "id-1", "id-2" };
    String[] value = new String[] { "v-0", "v-1", "v-2" };
    double[] weight = new double[] { 0.2, 0.7, 0.1 };

    /* RegularTree */
    LogicTree.Builder<String> regularTreeBuilder =
        LogicTree.builder("basic-tree");
    LogicTree<String> tree = regularTreeBuilder
        .addBranch(id[0], value[0], weight[0])
        .addBranch(id[1], value[1], weight[1])
        .addBranch(id[2], value[2], weight[2])
        .build();
    assertEquals(3, tree.size());
    assertEquals("basic-tree", tree.name());

    Branch<String> branch = tree.get(1);
    assertEquals(id[1], branch.id());
    assertEquals(value[1], branch.value());
    assertEquals(weight[1], branch.weight());
    assertEquals(branchString(id[1], weight[1], value[1]), branch.toString());
    assertEquals(tree.get(0), tree.iterator().next()); // iterator

    /* StringValueTree */
    LogicTree.StringValueBuilder stringValueTreeBuilder =
        LogicTree.stringValueBuilder("string-value-tree");
    tree = stringValueTreeBuilder
        .addBranch(id[0], weight[0])
        .addBranch(id[1], weight[1])
        .addBranch(id[2], weight[2])
        .build();

    branch = tree.get(2);
    assertEquals(id[2], branch.id());
    assertEquals(id[2], branch.value());
    assertEquals(0.1, branch.weight());
    assertEquals(branchString(id[2], weight[2], id[2]), branch.toString());

    /* EnumTree */
    LogicTree.EnumBuilder<EnumKey, String> enumTreeBuilder =
        LogicTree.enumBuilder("enum-tree");
    tree = enumTreeBuilder
        .addBranch(EnumKey.ID_0, value[0], weight[0])
        .addBranch(EnumKey.ID_1, value[1], weight[1])
        .addBranch(EnumKey.ID_2, value[2], weight[2])
        .build();

    branch = tree.get(2);
    assertEquals(EnumKey.ID_2.name(), branch.id());
    assertEquals(value[2], branch.value());
    assertEquals(0.1, branch.weight());
    assertEquals(branchString(EnumKey.ID_2.name(), weight[2], value[2]), branch.toString());

    /* EnumValueTree */
    LogicTree.EnumValueBuilder<EnumKey> enumValueTreeBuilder =
        LogicTree.enumValueBuilder("enum-value-tree");
    LogicTree<EnumKey> enumValueTree = enumValueTreeBuilder
        .addBranch(EnumKey.ID_0, weight[0])
        .addBranch(EnumKey.ID_1, weight[1])
        .addBranch(EnumKey.ID_2, weight[2])
        .build();

    Branch<EnumKey> enumBranch = enumValueTree.get(2);
    assertEquals(EnumKey.ID_2.name(), enumBranch.id());
    assertEquals(EnumKey.ID_2, enumBranch.value());
    assertEquals(0.1, enumBranch.weight());
    assertEquals(
        branchString(EnumKey.ID_2.name(), weight[2], EnumKey.ID_2.name()),
        enumBranch.toString());

    /* LogicGroup */
    LogicGroup.Builder<String> groupBuilder = LogicGroup.builder();
    LogicGroup<String> group = groupBuilder
        .addBranch(id[0], value[0], weight[0])
        .addBranch(id[1], value[1], weight[1])
        .build();

    assertEquals("logic-group", group.name());
    branch = group.get(0);
    assertEquals(id[0], branch.id());
    assertEquals(value[0], branch.value());
    assertEquals(weight[0], branch.weight());
    assertEquals(branchString(id[0], weight[0], value[0]), branch.toString());
    assertEquals(2, group.size());
    assertEquals(group.get(0), group.iterator().next()); // iterator

  }

  private static String branchString(String id, double weight, String value) {
    return id + " [" + weight + "] " + value;
  }

  final static String TEST_TREE = "test-tree";

  @Test
  final void treeToStringTest() {

    String[] id = new String[] { "id-0", "id-1", "id-2" };
    String[] value = new String[] { "v-0", "v-1", "v-2" };
    double[] weight = new double[] { 0.4, 0.1, 0.5 };

    LogicTree<String> tree = LogicTree.<String> builder(TEST_TREE)
        .addBranch(id[0], value[0], weight[0])
        .addBranch(id[1], value[1], weight[1])
        .addBranch(id[2], value[2], weight[2])
        .build();
    String treeString = new StringBuilder()
        .append(TEST_TREE).append(" ┬ ").append(branchString(id[0], weight[0], value[0]))
        .append(Text.NEWLINE)
        .append("          ├ ").append(branchString(id[1], weight[1], value[1]))
        .append(Text.NEWLINE)
        .append("          └ ").append(branchString(id[2], weight[2], value[2]))
        .toString();
    assertEquals(treeString, tree.toString());

    LogicTree<String> singletonTree = LogicTree.singleton(TEST_TREE, id[0], value[0]);
    String singletonTreeString = new StringBuilder()
        .append(TEST_TREE).append(" ─ ").append(branchString(id[0], 1.0, value[0]))
        .toString();
    assertEquals(singletonTreeString, singletonTree.toString());

    /* Logic Group */
    LogicGroup<String> group = LogicGroup.<String> builder()
        .addBranch(id[0], value[0], 1.0)
        .addBranch(id[1], value[1], 1.0)
        .build();
    String groupString = new StringBuilder()
        .append("logic-group").append(" ┬ ").append(branchString(id[0], 1.0, value[0]))
        .append(Text.NEWLINE)
        .append("            └ ").append(branchString(id[1], 1.0, value[1]))
        .toString();
    assertEquals(groupString, group.toString());
  }

  @Test
  final void builderExceptionTests() {

    /* Empty tree: IllegalState, in RegularTree constructor */
    assertThrows(IllegalStateException.class, () -> {
      LogicTree.builder(TEST_TREE).build();
    });

    /* Weight sum 1 != 1.0, in RegularTree constructor */
    assertThrows(IllegalArgumentException.class, () -> {
      LogicTree.builder(TEST_TREE)
          .addBranch("id", "value", 0.1)
          .build();
    });

    /* Empty branch id */
    assertThrows(IllegalArgumentException.class, () -> {
      LogicTree.builder(TEST_TREE)
          .addBranch(" ", "value", 1.0);
    });
    assertThrows(IllegalArgumentException.class, () -> {
      LogicTree.stringValueBuilder(TEST_TREE)
          .addBranch(" ", 1.0);
    });

    /* Duplicate branch id. */
    assertThrows(IllegalArgumentException.class, () -> {
      LogicTree.builder(TEST_TREE)
          .addBranch("id", "value1", 0.4)
          .addBranch("id", "value2", 0.6);
    });

    /* Used builder */
    assertThrows(IllegalStateException.class, () -> {
      LogicTree.Builder<String> builder = LogicTree.<String> builder(TEST_TREE)
          .addBranch("id", "value", 1.0);
      builder.build();
      builder.build();
    });
  }

  @Test
  final void groupExceptionTests() {

    /* no sampling permitted */
    assertThrows(UnsupportedOperationException.class, () -> {
      LogicGroup<String> group = LogicGroup.<String> builder()
          .addBranch("b1", "v1", 0.2)
          .build();
      group.sample(0.5);
    });
    assertThrows(UnsupportedOperationException.class, () -> {
      LogicGroup<String> group = LogicGroup.<String> builder()
          .addBranch("b1", "v1", 0.2)
          .build();
      group.sample(new double[] { 0.1, 0.5 });
    });

    /* Used builder */
    assertThrows(IllegalStateException.class, () -> {
      LogicGroup.Builder<String> builder = LogicGroup.<String> builder()
          .addBranch("id", "value", 1.0);
      builder.build();
      builder.build();
    });

    /* Empty group */
    assertThrows(IllegalStateException.class, () -> {
      LogicGroup.builder().build();
    });

  }

  @Test
  final void sampleTest() {

    String[] id = new String[] { "id-0", "id-1", "id-2", "id-3" };
    String[] value = new String[] { "v-0", "v-1", "v-2", "v-3" };
    double[] weight = new double[] { 0.4, 0.3, 0.2, 0.1 };
    double[] probabilities = new double[] { 0.2, 0.4, 0.41, 0.7, 0.71, 0.9, 0.91, 1.0 };

    LogicTree.Builder<String> treeBuilder = LogicTree.builder(TEST_TREE);
    LogicTree<String> tree = treeBuilder
        .addBranch(id[0], value[0], weight[0])
        .addBranch(id[1], value[1], weight[1])
        .addBranch(id[2], value[2], weight[2])
        .addBranch(id[3], value[3], weight[3])
        .build();

    List<Branch<String>> expected = List.of(
        tree.get(0),
        tree.get(1),
        tree.get(1),
        tree.get(2),
        tree.get(2),
        tree.get(3),
        tree.get(3),
        tree.get(3));

    List<Branch<String>> samples = tree.sample(probabilities);
    assertEquals(expected, samples);

    /* Test P<0 and P>1 */
    assertEquals(tree.get(0), tree.sample(-0.5));
    assertEquals(tree.get(3), tree.sample(1.5));

    /*
     * Creating the cumulative weight array introduces double precision math
     * errors such that the test weights herein become [0.4, 0.7,
     * 0.8999999999999999, 0.9999999999999999] if we did not round them after
     * cumulating. This behavior is tested by sampling at the actual weight
     * values. For example, if we sample at 0.8999999999999999 and had not
     * rounded then we would get back branch 3 instead of 2.
     */
    assertEquals(tree.get(2), tree.sample(0.8999999999999999));
  }

  private static final String SINGLETON_VALUE = "singleton-value";
  private static final String SINGLETON_ID = "singleton-id";

  @Test
  final void singletonTest() {
    // LogicTree<String> tree = LogicTree.singleton(SINGLETON_VALUE);
    // assertTrue(tree.size() == 1);
    // assertEquals(tree.branches().get(0).value(), SINGLETON_VALUE);
    // assertEquals(tree.branches().get(0).weight(), 1.0);
    LogicTree<String> tree = LogicTree.singleton(
        TEST_TREE,
        SINGLETON_ID,
        SINGLETON_VALUE);
    assertEquals(tree.get(0).id(), SINGLETON_ID);
  }

  @Test
  final void branchesImmutableTest() {
    /* Ensure branches() is unmodifiable */
    assertThrows(UnsupportedOperationException.class, () -> {
      LogicTree<String> tree = LogicTree.singleton(
          TEST_TREE,
          SINGLETON_ID,
          SINGLETON_VALUE);
      tree.remove(0);
    });
  }

}
